/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-1999, William Chia-Wei Cheng.
 *
 * Permission limited to the use, copy, display, distribute without
 * charging for a fee, and produce derivative works of "tgif" and
 * its documentation for not-for-profit purpose is hereby granted by
 * the Author, provided that the above copyright notice appears in
 * all copies made of "tgif" and that both the copyright notice
 * and this permission notice appear in supporting documentation,
 * and that the name of the Author not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.  All other rights (including, but not limited to, the
 * right to sell "tgif", the right to sell or distribute derivative
 * works of "tgif", the right to distribute "tgif" for a fee, and
 * the right to include "tgif" or derivative works of "tgif" in a
 * for-sale product) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * @(#)$Header: /mm/src/tgif/v4/RCS/file.c,v 4.58 1999/05/28 18:59:12 william Exp $
 */

#define _INCLUDE_FROM_FILE_C_

#include "tgifdefs.h"
#include "expfdefs.h"
#include "cmdids.h"

#include "align.e"
#include "arc.e"
#include "attr.e"
#include "auxtext.e"
#include "box.e"
#include "button.e"
#include "choice.e"
#include "cmd.e"
#include "color.e"
#include "cutpaste.e"
#include "cursor.e"
#include "dialog.e"
#include "drawing.e"
#include "dup.e"
#include "eps.e"
#include "exec.e"
#include "file.e"
#include "font.e"
#include "grid.e"
#include "group.e"
#include "import.e"
#include "mainloop.e"
#include "mainmenu.e"
#include "mark.e"
#include "menu.e"
#include "menuinfo.e"
#include "miniline.e"
#include "move.e"
#include "msg.e"
#include "names.e"
#include "navigate.e"
#include "obj.e"
#include "oval.e"
#include "page.e"
#include "pattern.e"
#include "poly.e"
#include "polygon.e"
#include "prtgif.e"
#include "ps.e"
#include "raster.e"
#include "rcbox.e"
#include "rect.e"
#include "remote.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"
#include "shape.e"
#include "special.e"
#include "stk.e"
#include "stretch.e"
#include "text.e"
#include "util.e"
#include "version.e"
#include "wb.e"
#include "xbitmap.e"
#include "xpixmap.e"
#include "xprtfltr.e"

#ifdef _METRIC_PIX_PER_INCH
#define METRIC_PIX_PER_INCH 127
#endif /* _METRIC_PIX_PER_INCH */

#define CUR_VERSION 37

int	PRTGIF=FALSE;
int	prTgifFoundColorInfo=FALSE;
char	curFileName[MAXPATHLENGTH+1];
int	curFileDefined=FALSE;
int	fileVersion=INVALID;
int	curFileWriteVersion=CUR_VERSION;
int	importingFile=FALSE;
int	serializingFile=FALSE;
int	deserializingFile=FALSE;
int	psDotsPerInch=72;
float	printMag=(float)100.0;
int	saveTmpOnReturn=TRUE;
int	warpToWinCenter=TRUE;
float	tiledPageScaling=(float)0.9;

char	*psXOffStr[MAXPAGESTYLES] = {"0","0"};
float	psXOff[MAXPAGESTYLES] =     { 0,  0 };
char	* * psYOffStr=NULL;
float	* psYOff=NULL;
float	* psPageWidthInInch=NULL;
float	* psPageHeightInInch=NULL;

char	printCommand[MAXSTRING+1];
char	outputDir[MAXPATHLENGTH+1];

char	* savedComments=NULL;
int	savedCommentsLen=0;
int	saveCommentsInSaveNew=TRUE;
int	usePsAdobeString=FALSE;
char	adobeString[80], epsfString[80];

int	readingPageNum=0;
int	loadedCurPageNum=0;

int	writeFileFailed=FALSE;
int	foundGoodStateObject=FALSE;

int	cmdLineHasPageNum=FALSE;
int	cmdLinePageNum=(-1);
char	cmdLinePageNumStr[80];

int	cmdLineOneFilePerPage=FALSE;
int	cmdLineA4=FALSE;
int	cmdLineColor=FALSE;
int	cmdLineRequestedColor=FALSE;

int	cmdLineDumpURL=FALSE;
int	cmdLineDumpURLWithHeader=FALSE;
int	cmdLineDumpURLHeaderOnly=FALSE;
int	cmdLineDumpURLShowStatus=FALSE;

int	cmdLineDosEpsFilter=FALSE;
int	cmdLinePreviewOnly=FALSE;

int	tmpFileMode=PSFILE_MOD;
char	cmdLineTmpFileModeStr[MAXSTRING];

int	cmdLineOpenDisplay=FALSE;
int	cmdLineStdOut=FALSE;
int	cmdLineWhereToPrint=INVALID;
int	cmdLineTiffEPSI=INVALID;

char	cmdLineFileToExec[MAXSTRING];

int	generateTiffEPSI=FALSE;
int	showPageInEPS=TRUE;

int	gnQuit=FALSE;

int	myFileBgPixel=INVALID, myFileFgPixel=INVALID;
char	*myFileBgColorStr=NULL, *myFileFgColorStr=NULL;
XColor	myFileBgColor;
int	myFileBgPixmapW=(-1), myFileBgPixmapH=(-1);
Pixmap	myFileBgPixmap=None;
char	*myFileBgGifURL=NULL;

int	dumpOneFilePerPage=FALSE;

int	bufferAsFileForWhiteBoard=FALSE;

int	flushColormapOnOpen=FALSE;

struct BBRec	*gpExportClipBBox=NULL;

int		gnNumFilePSFontAliases=0;
KeyValInfo	*gaFilePSFontAliases=NULL;

static int	psRegMarksInTiledPageMode=INVALID;
static float	psRegMarksGray=(float)0.95;

static int importingPageNum=(-1);
static char importingPageName[MAXSTRING+1];

struct DocFontRec {
   char name[80];
   int len;
   struct DocFontRec *next;
};

static struct DocFontRec *firstDocFont=NULL;

void UpdateDocumentFonts(ps_font_name)
   char *ps_font_name;
{
   int len=strlen(ps_font_name);
   struct DocFontRec *df_ptr;

   for (df_ptr=firstDocFont; df_ptr != NULL; df_ptr=df_ptr->next) {
      if (df_ptr->len == len && strcmp(df_ptr->name, ps_font_name) == 0) {
         return;
      }
   }
   df_ptr = (struct DocFontRec *)malloc(sizeof(struct DocFontRec));
   if (df_ptr == NULL) FailAllocMessage();
   memset(df_ptr, 0, sizeof(struct DocFontRec));
   df_ptr->len = len;
   strcpy(df_ptr->name, ps_font_name);
   df_ptr->next = firstDocFont;
   firstDocFont = df_ptr;
}

static int dontUseShortHandPS=INVALID;
static int dontCondense=INVALID;
static int stripSimpleComments=INVALID;

void SetCmdLineDontCondense()
{
   dontUseShortHandPS = TRUE;
   dontCondense = TRUE;
   stripSimpleComments = FALSE;
}

void SetCmdLineCondensedPS()
{
   dontUseShortHandPS = FALSE;
   dontCondense = FALSE;
   stripSimpleComments = TRUE;
}

static char gszResidual[80];
static int gnResidualLen=0;

static
void DumpResidual(write_fp)
   FILE *write_fp;
{
   if (gnResidualLen > 0) {
      if (fputs(gszResidual, write_fp) == EOF) writeFileFailed = TRUE;
      if (fputs("\n", write_fp) == EOF) writeFileFailed = TRUE;
      *gszResidual = '\0';
      gnResidualLen = 0;
   }
}

static
void AddToResidual(write_fp, buf)
   FILE *write_fp;
   char *buf;
{
   int len=strlen(buf);

   if (len <= 0) return;

   if (len+gnResidualLen >= 77) {
      DumpResidual(write_fp);
      if (len >= 77) {
         if (fputs(buf, write_fp) == EOF) writeFileFailed = TRUE;
         if (fputs("\n", write_fp) == EOF) writeFileFailed = TRUE;
         return;
      }
   }
   if (gnResidualLen == 0) {
      sprintf(&gszResidual[gnResidualLen], "%s", buf);
      gnResidualLen += len;
   } else {
      sprintf(&gszResidual[gnResidualLen], " %s", buf);
      gnResidualLen += len+1;
   }
}

static
char *GetPostScriptToken(read_fname, pszStart, ppszEnd)
   char *read_fname, *pszStart, **ppszEnd;
{
   char *psz=NULL;

   *ppszEnd = NULL;
   while (*pszStart == ' ' || *pszStart == '\t') pszStart++;
   if (*pszStart == '\0') return NULL;
   if (*pszStart == '%') return NULL;
   if (*pszStart == '(') {
      for (psz=pszStart; *psz != ')' && *psz != '\0'; psz++) {
         if (*psz == '\\') {
            psz++;
            if (*psz >= '0' && *psz <= '7') {
               psz++;
               psz++;
            }
         }
      }
      if (*psz == ')') {
         *ppszEnd = (&psz[1]);
      } else {
         /* no problem - Adobe Illustrator 6.0 does this a lot */
      }
   } else {
      for (psz=pszStart;
            *psz!='\0' && *psz!=' ' && *psz!='\t' && *psz!='(' && *psz!='%';
            psz++) {
      }
      if (*psz == '%') {
         *psz = '\0';
      }
      if (*psz != '\0') {
         *ppszEnd = psz;
      }
   }
   return pszStart;
}

static
void CondenseAndCopyPostScriptFile(read_fp, write_fp, read_fname, condense)
   FILE *read_fp, *write_fp;
   char *read_fname;
   int condense;
{
   if (condense) {
      char *buf;
      int previewing=FALSE, begin_preview_len=strlen("%%BeginPreview:");

      if (stripSimpleComments == INVALID) stripSimpleComments = FALSE;
      *gszResidual = '\0';
      gnResidualLen = 0;
      while ((buf=UtilGetALine(read_fp)) != NULL) {
         if (*buf == '%') {
            if (strncmp(buf, "%%BeginPreview:", begin_preview_len) == 0) {
               previewing = TRUE;
            } else if (strncmp(buf, "%%EndPreview", begin_preview_len-3) == 0) {
               previewing = FALSE;
            }
            if (!stripSimpleComments || buf[1] == '%' || buf[1] == '!' ||
                  previewing) {
               DumpResidual(write_fp);
               AddToResidual(write_fp, buf);
               DumpResidual(write_fp);
            }
         } else {
            char *psz=NULL, *pszStart=buf, *pszEnd=NULL;

            while ((psz=GetPostScriptToken(read_fname, pszStart, &pszEnd)) !=
                  NULL) {
               char saved_ch='\0';

               if (pszEnd != NULL) {
                  saved_ch = (*pszEnd);
                  *pszEnd = '\0';
               }
               AddToResidual(write_fp, psz);
               if (pszEnd == NULL) {
                  break;
               }
               *pszEnd = saved_ch;
               pszStart = pszEnd;
               pszEnd = NULL;
            }
         }
         UtilFree(buf);
      }
      DumpResidual(write_fp);
   } else {
      char tmp_str[MAXSTRING<<1];

      while (fgets(tmp_str, MAXSTRING, read_fp) != NULL) {
         if (fputs(tmp_str, write_fp) == EOF) {
            writeFileFailed = TRUE;
         }
      }
   }
}

static
int CondensePostScriptFile(read_fname, buf_sz)
   char *read_fname;
   int buf_sz;
{
   char write_fname[MAXSTRING+1];
   FILE *read_fp=NULL, *write_fp=NULL;

   sprintf(write_fname, "%sTgifXXXXXX", TMP_DIR);
   mktemp(write_fname);
   unlink(write_fname);

   if ((read_fp=fopen(read_fname, "r")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for reading, print aborted.",
            read_fname);
      if (PRTGIF) {
         fprintf(stderr, "%s\n", gszMsgBox);
      } else {
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
      return FALSE;
   }
   if ((write_fp=fopen(write_fname, "w")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for writing, print aborted.",
            write_fname);
      if (PRTGIF) {
         fprintf(stderr, "%s\n", gszMsgBox);
      } else {
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
      fclose(read_fp);
      return FALSE;
   }
   CondenseAndCopyPostScriptFile(read_fp, write_fp, read_fname, TRUE);
   fclose(read_fp);
   fclose(write_fp);
   unlink(read_fname);
   UtilStrCpyN(read_fname, buf_sz, write_fname);

   return TRUE;
}

static
int CopyAFile(file1, file2, condense)
   char *file1, *file2;
   int condense;
{
   char tmp_str[MAXSTRING+1], *rest;
   FILE *fp1, *fp2;

   if ((fp1=fopen(file1, "r")) == NULL) {
      sprintf(tmp_str, "Cannot open '%s' for read.", file1);
      if (PRTGIF) {
         fprintf(stderr, "%s\n", tmp_str);
      } else {
         MsgBox(tmp_str, TOOL_NAME, INFO_MB);
      }
      return FALSE;
   }
   if (file2 == NULL) {
      /* PRTGIF && cmdLineStdOut */
      fp2 = stdout;
   } else {
      int short_name=FALSE;

      if ((short_name=IsPrefix(bootDir, file2, &rest))) ++rest;
      if ((fp2=fopen(file2, "w")) == NULL) {
         if (PRTGIF) {
            fprintf(stderr, "Cannot open '%s' for write.\n", file2);
         } else {
            if (short_name) {
               sprintf(tmp_str, "Cannot open '%s' for write.", rest);
            } else {
               sprintf(tmp_str, "Cannot open '%s' for write.", file2);
            }
            MsgBox(tmp_str, TOOL_NAME, INFO_MB);
         }
         fclose(fp1);
         return FALSE;
      }
   }
   writeFileFailed = FALSE;
   CondenseAndCopyPostScriptFile(fp1, fp2, file1, (condense && !dontCondense));
   fclose(fp1);
   if (file2 != NULL) fclose(fp2);
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      if (file2 == NULL) {
         sprintf(tmp_str, "Fail to write file.\n\nFile system may be full.");
      } else {
         sprintf(tmp_str, "Fail to write to '%s'.\n\nFile system may be full.",
               file2);
      }
      if (PRTGIF) {
         fprintf(stderr, "%s\n", tmp_str);
      } else {
         MsgBox(tmp_str, TOOL_NAME, INFO_MB);
      }
      return FALSE;
   }
   return TRUE;
}

int FindProgramInPath(cmd, msg)
   char *cmd, *msg;
{
   char *psz=strchr(cmd, ' '), szPath[MAXPATHLENGTH];

   if (psz != NULL) *psz = '\0';
   UtilStrCpyN(szPath, sizeof(szPath), cmd);
   if (psz != NULL) *psz = ' ';

   if (*szPath == '\0') {
      /* no program specified, assume it's okay */
      return TRUE;
   } else if (*szPath == DIR_SEP) {
      if (UtilPathExists(szPath)) return TRUE;
   } else {
      char *env=NULL;

      psz = getenv("PATH");
      if (psz == NULL) {
         /* cannot get the PATH environment, assume it's okay */
         return TRUE;
      }
      env = UtilStrDup(psz);
      if (env == NULL) FailAllocMessage();
      for (psz=strtok(env, ":"); psz != NULL; psz=strtok(NULL, ":")) {
         sprintf(gszMsgBox, "%s%c%s", psz, DIR_SEP, szPath);
         if (UtilPathExists(gszMsgBox)) {
            UtilFree(env);
            return TRUE;
         }
      }
      UtilFree(env);
   }
   if (msg == NULL || *msg == '\0') {
      sprintf(gszMsgBox, "Cannot find '%s'%s.\n\n%s '%s' anyway?",
            szPath, (*szPath==DIR_SEP ? "" : " in the PATH"),
            "Would you like to execute", cmd);
   } else {
      sprintf(gszMsgBox, "Cannot find '%s'%s.  %s\n\n%s '%s' anyway?",
            szPath, (*szPath==DIR_SEP ? "" : " in the PATH"), msg,
            "Would you like to execute", cmd);
   }
   return (MsgBox(gszMsgBox, TOOL_NAME, YNC_MB)==MB_ID_YES);
}

int ExecuteCmd(cmd)
   char *cmd;
{
   char tmp_str[MAXSTRING+1];
   FILE *fp;

   sprintf(gszMsgBox, "Executing '%s'...", cmd);
   SetStringStatus(gszMsgBox);
   if (!PRTGIF) XSync(mainDisplay, False);
   if ((fp=(FILE*)popen(cmd, "r")) == NULL) return FALSE;

   while (fgets(tmp_str, MAXSTRING, fp) != NULL) {
      if (PRTGIF) {
         fprintf(stderr, "%s", tmp_str);
      } else {
         Msg(tmp_str);
      }
   }
   pclose(fp);
   SetStringStatus("...Done");
   return TRUE;
}

void CleanUpComments()
{
   if (savedComments != NULL) {
      free(savedComments);
      savedComments = NULL;
      savedCommentsLen = 0;
   }
}

void ClearFileInfo(clean_up_comments)
   int clean_up_comments;
{
   curFileName[0] = '\0';
   curFileDefined = FALSE;
   if (!curDirIsLocal) strcpy(curDir, curLocalDir);
   curDirIsLocal = TRUE;
   *curLocalDir = '\0';
   *curSymDir = '\0';

   if (clean_up_comments) {
      CleanUpComments();
   }
}

int OkayToCreateFile(FileName)
   char *FileName;
{
   FILE *fp;

   if ((fp=fopen(FileName, "r")) == NULL) return TRUE;
   fclose(fp);
   sprintf(gszMsgBox, "File '%s' exists, okay to overwrite? [ync](y)",
         FileName);
   switch (MsgBox(gszMsgBox, TOOL_NAME, YNC_MB)) {
   case MB_ID_YES: break;
   case MB_ID_NO: return FALSE;
   case MB_ID_CANCEL: return FALSE;
   }
   unlink(FileName);
   return TRUE;
}

void SaveObj(FP, ObjPtr, Level)
   FILE *FP;
   struct ObjRec *ObjPtr;
   int Level;
{
   switch (ObjPtr->type) {
   case OBJ_POLY: SavePolyObj(FP, ObjPtr); break;
   case OBJ_BOX: SaveBoxObj(FP, ObjPtr); break;
   case OBJ_OVAL: SaveOvalObj(FP, ObjPtr); break;
   case OBJ_TEXT: SaveTextObj(FP, ObjPtr); break;
   case OBJ_POLYGON: SavePolygonObj(FP, ObjPtr); break;
   case OBJ_ARC: SaveArcObj(FP, ObjPtr); break;
   case OBJ_RCBOX: SaveRCBoxObj(FP, ObjPtr); break;
   case OBJ_XBM: SaveXBmObj(FP, ObjPtr); break;
   case OBJ_XPM: SaveXPmObj(FP, ObjPtr); break;
   case OBJ_GROUP: SaveGroupObj(FP, ObjPtr, Level); break;
   case OBJ_SYM: SaveCompObj(FP, ObjPtr, Level); break;
   case OBJ_ICON: SaveIconObj(FP, ObjPtr, Level); break;
   case OBJ_PIN: SavePinObj(FP, ObjPtr, Level); break;
   }
}

int FileNameHasExtension(file_name, file_type, pn_gzipped, pn_no_name)
   char *file_name;
   int file_type, *pn_gzipped, *pn_no_name;
{
   static char stszObjFileExt[20], stszGzObjFileExt[20];
   static char stszSymFileExt[20];
   static char stszPinFileExt[20];
   static int initialized=FALSE;
   char *psz=NULL;

   if (!initialized) {
      sprintf(stszObjFileExt, ".%s", OBJ_FILE_EXT);
      sprintf(stszGzObjFileExt, ".%s.gz", OBJ_FILE_EXT);
      sprintf(stszSymFileExt, ".%s", SYM_FILE_EXT);
      sprintf(stszPinFileExt, ".%s", PIN_FILE_EXT);
      initialized = TRUE;
   }
   if (pn_gzipped != NULL) *pn_gzipped = FALSE;
   switch (file_type) {
   case OBJ_FILE_TYPE:
      if ((psz=strstr(file_name, ".obj")) != NULL &&
            strcmp(psz, ".obj") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, ".obj.gz")) != NULL &&
            strcmp(psz, ".obj.gz") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         if (pn_gzipped != NULL) *pn_gzipped = TRUE;
         return TRUE;
      } else if ((psz=strstr(file_name, ".tgo")) != NULL &&
            strcmp(psz, ".tgo") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, ".tgo.gz")) != NULL &&
            strcmp(psz, ".tgo.gz") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         if (pn_gzipped != NULL) *pn_gzipped = TRUE;
         return TRUE;
      } else if ((psz=strstr(file_name, stszObjFileExt)) != NULL &&
            strcmp(psz, stszObjFileExt) == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, stszGzObjFileExt)) != NULL &&
            strcmp(psz, stszGzObjFileExt) == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         if (pn_gzipped != NULL) *pn_gzipped = TRUE;
         return TRUE;
      }
      break;
   case SYM_FILE_TYPE:
      if ((psz=strstr(file_name, ".sym")) != NULL &&
            strcmp(psz, ".sym") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, ".tgs")) != NULL &&
            strcmp(psz, ".tgs") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, stszSymFileExt)) != NULL &&
            strcmp(psz, stszSymFileExt) == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      }
      break;
   case PIN_FILE_TYPE:
      if ((psz=strstr(file_name, ".pin")) != NULL &&
            strcmp(psz, ".pin") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, ".tgp")) != NULL &&
            strcmp(psz, ".tgp") == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      } else if ((psz=strstr(file_name, stszPinFileExt)) != NULL &&
            strcmp(psz, stszPinFileExt) == 0) {
         if (pn_no_name != NULL) *pn_no_name = (psz == file_name);
         return TRUE;
      }
      break;
   }
   if (pn_no_name != NULL) *pn_no_name = (*file_name == '\0');

   return FALSE;
}

void RemoveFileNameExtension(file_name)
   char *file_name;
{
   static char stszObjFileExt[20], stszGzObjFileExt[20];
   static char stszSymFileExt[20];
   static char stszPinFileExt[20];
   static int initialized=FALSE;
   char *psz=NULL;
   int len=strlen(file_name);

   if (!initialized) {
      sprintf(stszObjFileExt, ".%s", OBJ_FILE_EXT);
      sprintf(stszGzObjFileExt, ".%s.gz", OBJ_FILE_EXT);
      sprintf(stszSymFileExt, ".%s", SYM_FILE_EXT);
      sprintf(stszPinFileExt, ".%s", PIN_FILE_EXT);
      initialized = TRUE;
   }
   if ((psz=strstr(file_name, ".obj")) != NULL &&
         strcmp(psz, ".obj") == 0) {
      file_name[len-strlen(".obj")] = '\0';
   } else if ((psz=strstr(file_name, ".obj.gz")) != NULL &&
         strcmp(psz, ".obj.gz") == 0) {
      file_name[len-strlen(".obj.gz")] = '\0';
   } else if ((psz=strstr(file_name, ".tgo")) != NULL &&
         strcmp(psz, ".tgo") == 0) {
      file_name[len-strlen(".tgo")] = '\0';
   } else if ((psz=strstr(file_name, ".tgo.gz")) != NULL &&
         strcmp(psz, ".tgo.gz") == 0) {
      file_name[len-strlen(".tgo.gz")] = '\0';
   } else if ((psz=strstr(file_name, stszObjFileExt)) != NULL &&
         strcmp(psz, stszObjFileExt) == 0) {
      file_name[len-strlen(stszObjFileExt)] = '\0';
   } else if ((psz=strstr(file_name, stszGzObjFileExt)) != NULL &&
         strcmp(psz, stszGzObjFileExt) == 0) {
      file_name[len-strlen(stszGzObjFileExt)] = '\0';
   } else if ((psz=strstr(file_name, ".sym")) != NULL &&
         strcmp(psz, ".sym") == 0) {
      file_name[len-strlen(".sym")] = '\0';
   } else if ((psz=strstr(file_name, ".tgs")) != NULL &&
         strcmp(psz, ".tgs") == 0) {
      file_name[len-strlen(".tgs")] = '\0';
   } else if ((psz=strstr(file_name, stszSymFileExt)) != NULL &&
         strcmp(psz, stszSymFileExt) == 0) {
      file_name[len-strlen(stszSymFileExt)] = '\0';
   } else if ((psz=strstr(file_name, ".pin")) != NULL &&
         strcmp(psz, ".pin") == 0) {
      file_name[len-strlen(".pin")] = '\0';
   } else if ((psz=strstr(file_name, ".tgp")) != NULL &&
         strcmp(psz, ".tgp") == 0) {
      file_name[len-strlen(".tgp")] = '\0';
   } else if ((psz=strstr(file_name, stszPinFileExt)) != NULL &&
         strcmp(psz, stszPinFileExt) == 0) {
      file_name[len-strlen(stszPinFileExt)] = '\0';
   }
}

void Save(FP, BotObjPtr, Level, PageNumber)
   FILE *FP;
   struct ObjRec *BotObjPtr;
   int Level, PageNumber;
{
   int watch_cursor=watchCursorOnMainWindow;
   struct ObjRec *obj_ptr=NULL;

   if (BotObjPtr != NULL && !copyInDrawTextMode) {
      SetCurChoice(NOTHING);

      if (!watch_cursor) {
         SetWatchCursor(drawWindow);
         SetWatchCursor(mainWindow);
      }
   }
   if (Level == 0 && PageNumber == 1) {
      char font_str[MAXSTRING];

      ResetXPmErrorMessage();
      GetPSFontStr(curFont, curStyle, font_str);
      SetFullVersionString();
      if (fprintf(FP, "%%TGIF %s\n", fullVersionString) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "state(%1d,%1d,%.3f,", pageStyle, CUR_VERSION,
            printMag) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,", drawOrigX, drawOrigY, zoomScale) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,", xyEnglishGrid, snapOn, colorIndex) ==
            EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,", horiAlign, vertAlign, lineWidth) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,%1d,", curSpline, lineStyle, objFill,
            penPat) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,'%s',%1d,%1d,", textJust,
            /* font_str starts with the '/' character */
            &font_str[1], curStyle, GetCurSzUnit()) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,", 0, curDash, gridSystem) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,", xyMetricGrid, textVSpace, zoomedIn) ==
            EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,", gridShown, moveMode, ROTATE0) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,", rcbRadius, useGray) == EOF) {
         writeFileFailed = TRUE;
      }
      if (fprintf(FP, "%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d).\n",
            pageLayoutMode,
            (pageLayoutMode == PAGE_STACK) ? curPageNum : paperCol,
            (pageLayoutMode == PAGE_STACK) ? lastPageNum : paperRow,
            pageLineShownInTileMode, colorDump,
            round(((float)onePageWidth)*printMag/100.0),
            round(((float)onePageHeight)*printMag/100.0),
            stretchableText, textRotation, rotationIncrement,
            transPat) == EOF) {
         writeFileFailed = TRUE;
      }
      if ((BotObjPtr != NULL || tgifObj != NULL) && !copyInDrawTextMode &&
            !serializingFile) {
         if (fprintf(FP, "%%\n") == EOF) {
            writeFileFailed = TRUE;
         }
         if (fprintf(FP, "%% @%s%s\n", "(#)$H", "eader$") == EOF) {
            writeFileFailed = TRUE;
         }
         if (fprintf(FP, "%% %s\n", "%W%") == EOF) {
            writeFileFailed = TRUE;
         }
         if (fprintf(FP, "%%\n") == EOF) {
            writeFileFailed = TRUE;
         }
         if (savedComments != NULL) {
            if (fputs(savedComments, FP) == EOF) {
               writeFileFailed = TRUE;
            }
         }
         if (tgifObj->lattr != NULL) {
            if (fprintf(FP, "file_attr(") == EOF) writeFileFailed = TRUE;
            SaveAttrs(FP, tgifObj->lattr);
            if (fprintf(FP, ").\n") == EOF) writeFileFailed = TRUE;
         }
         GetUnitSpec(gszMsgBox);
         if (fprintf(FP, "unit(\"") == EOF) writeFileFailed = TRUE;
         SaveString(FP, gszMsgBox);
         if (fprintf(FP, "\").\n") == EOF) writeFileFailed = TRUE;
         if (shapeShadowInResource ||
               shapeShadowDx != 0 || shapeShadowDy != 0) {
            if (fprintf(FP, "shapeshadow(%1d,%1d).\n", shapeShadowDx,
                  shapeShadowDy) == EOF) {
               writeFileFailed = TRUE;
            }
         }
         SaveColors(FP);
         if (slideShowInfoValid) {
            if (fprintf(FP, "slideshow_info('%s',%1d,%1d).\n",
                  (slideShowBorderColor==NULL ? "" : slideShowBorderColor),
                  slideShowXOffset, slideShowYOffset) == EOF) {
               writeFileFailed = TRUE;
            }
         }
         if (fprintf(FP, "script_frac(\"%s\").\n", scriptFractionStr) == EOF) {
            writeFileFailed = TRUE;
         }
         if (fprintf(FP, "fg_bg_colors('%s','%s').\n",
               colorMenuItems[colorIndex], (*defaultBgColorStr=='\0' ?
               myBgColorStr : defaultBgColorStr)) == EOF) {
            /*
             * The above should use colorMenuItmes[bgColorIndex]
             *       and not myBgColorStr.
             */
            writeFileFailed = TRUE;
         }
      }
   }
   if (Level == 0 && BotObjPtr != NULL && !copyInDrawTextMode) {
      if (fprintf(FP, "page(%1d,\"%s\",%1d,'%s').\n", PageNumber,
            ((pageLayoutMode==PAGE_TILE || curPage->name==NULL) ? "" :
            curPage->name), curPage->layer_on,
            ((pageLayoutMode==PAGE_TILE || curPage->page_file_name==NULL) ? "" :
            curPage->page_file_name)) == EOF) {
         writeFileFailed = TRUE;
      }
   }

   for (obj_ptr=BotObjPtr; obj_ptr != NULL; obj_ptr=obj_ptr->prev) {
      SaveObj(FP, obj_ptr, Level);
      if (obj_ptr->prev == NULL) {
         if (Level == 0) {
            if (fprintf(FP, ".\n") == EOF) writeFileFailed = TRUE;
         } else {
            if (fprintf(FP, "\n") == EOF) writeFileFailed = TRUE;
         }
      } else {
         if (Level == 0) {
            if (fprintf(FP, ".\n") == EOF) writeFileFailed = TRUE;
         } else {
            if (fprintf(FP, ",\n") == EOF) writeFileFailed = TRUE;
         }
      }
   }
   if (BotObjPtr != NULL && !copyInDrawTextMode) {
      if (!watch_cursor) {
         SetDefaultCursor(mainWindow);
         ShowCursor();
      }
   }
}

int SaveTmpFile(NewFileName)
   char *NewFileName;
   /* return TRUE if file successfully saved */
{
   char new_file_name[MAXPATHLENGTH+1], *rest=NULL;
   FILE *fp=NULL;
   int count=0, file_type=INVALID, short_name, watch_cursor=FALSE;
   struct PageRec *saved_cur_page;
   struct ObjRec *obj_ptr;
   struct ObjRec *obj_ptr1=NULL, *obj_ptr2=NULL;
   struct AttrRec *attr_ptr;

   strcpy(new_file_name, NewFileName);

   saved_cur_page = curPage;
   for (curPage=firstPage; curPage != NULL; curPage=curPage->next) {
      for (obj_ptr=curPage->top; obj_ptr != NULL; obj_ptr=obj_ptr->next) {
         if (obj_ptr->type == OBJ_SYM) {
            if (obj_ptr1 == NULL) {
               obj_ptr1 = obj_ptr;
            } else if (obj_ptr2 == NULL) {
               obj_ptr2 = obj_ptr;
            }
            count++;
         }
      }
   }
   curPage = saved_cur_page;

   switch (count) {
   case 0:
      sprintf(new_file_name, "%s.%s", NewFileName, OBJ_FILE_EXT);
      file_type = OBJ_FILE_TYPE;
      break;
   case 1:
      if (lastPageNum != 1) {
         MsgBox("A symbol file can only have one page.\n\nSave aborted!",
               TOOL_NAME, INFO_MB);
         return INVALID;
      }
      if ((attr_ptr=FindAttrWithName(obj_ptr1, "type=", NULL)) != NULL &&
            strcmp(attr_ptr->attr_value.s, "tgif_pin") == 0) {
         sprintf(new_file_name, "%s.%s", NewFileName, PIN_FILE_EXT);
         file_type = PIN_FILE_TYPE;
      } else {
         sprintf(new_file_name, "%s.%s", NewFileName, SYM_FILE_EXT);
         file_type = SYM_FILE_TYPE;
      }
      break;
   case 2:
      if (lastPageNum != 1) {
         MsgBox("A pin file can only have one page.\n\nSave aborted!",
               TOOL_NAME, INFO_MB);
         return INVALID;
      }
      sprintf(new_file_name, "%s.%s", NewFileName, PIN_FILE_EXT);
      file_type = PIN_FILE_TYPE;
      break;
   default:
      MsgBox("Too many symbols!\n\nSymbol file not saved.",
            TOOL_NAME, INFO_MB);
      return INVALID;
   }

   unlink(new_file_name);

   if ((short_name=IsPrefix(bootDir, new_file_name, &rest))) ++rest;
   if ((fp=fopen(new_file_name, "w")) == NULL) {
      if (short_name) {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", rest);
      } else {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", new_file_name);
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return INVALID;
   }

   if (short_name) {
      sprintf(gszMsgBox, "Saving temporary file '%s' ...", rest);
   } else {
      sprintf(gszMsgBox, "Saving temporary file '%s' ...", new_file_name);
   }
   Msg(gszMsgBox);

   watch_cursor = watchCursorOnMainWindow;
   if (!watch_cursor) {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
   }
   writeFileFailed = FALSE;
   MakeQuiescent();
   saved_cur_page = curPage;
   for (curPage=firstPage, count=1; curPage != NULL;
         curPage=curPage->next, count++) {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save(fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose(fp);
   if (!watch_cursor) {
      SetDefaultCursor(mainWindow);
      ShowCursor(); 
   }
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            new_file_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   } else {
      if (short_name) {
         sprintf(gszMsgBox, "Temporary file '%s' saved.", rest);
      } else {
         sprintf(gszMsgBox, "Temporary file '%s' saved.", new_file_name);
      }
      Msg(gszMsgBox);
   }

   if (tmpFileMode != 0 && chmod(new_file_name, tmpFileMode)) {
      if (short_name) {
         sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", rest, tmpFileMode);
      } else {
         sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", new_file_name,
               tmpFileMode);
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   }
   return file_type;
}

void SaveNewFile(SaveSelectedOnly, pszNewFile)
   int SaveSelectedOnly;
   char *pszNewFile;
{
   char new_file_name[MAXPATHLENGTH+1], gzipped_fname[MAXPATHLENGTH+1];
   char new_full_name[MAXPATHLENGTH+1], tmp_str[MAXPATHLENGTH+1], *rest=NULL;
   char saved_cur_dir[MAXPATHLENGTH+1], saved_cur_file_name[MAXPATHLENGTH+1];
   char obj_ext_str[MAXSTRING+1], gz_obj_ext_str[MAXSTRING+1];
   char sym_ext_str[MAXSTRING+1], pin_ext_str[MAXSTRING+1];
   int saved_cur_file_defined=FALSE, no_name=FALSE, gzipped=FALSE, file_type=0;
   int count=0, len, short_name=FALSE, ok=TRUE, watch_cursor=FALSE;
   struct ObjRec *obj_ptr=NULL, *saved_top_obj=NULL, *saved_bot_obj=NULL;
   struct ObjRec *obj_ptr1=NULL, *obj_ptr2=NULL;
   struct SelRec *top_sel_ptr=NULL, *bot_sel_ptr=NULL;
   struct SelRec *sel_ptr=NULL, *next_sel=NULL;
   struct PageRec *saved_cur_page=NULL;
   struct AttrRec *attr_ptr=NULL;
   FILE *fp=NULL;

   file_type = OBJ_FILE_TYPE;
   *gzipped_fname = '\0';

   if (SaveSelectedOnly && pszNewFile == NULL && topSel == NULL) {
      MsgBox("No objects selected!\n\nNothing saved!", TOOL_NAME, INFO_MB);
      return;
   }
   if (SaveSelectedOnly && pszNewFile != NULL) {
      strcpy(new_file_name, pszNewFile);
   } else {
      if (curDirIsLocal) {
         sprintf(gszMsgBox, "( working directory: %s )", curDir);
      } else {
         sprintf(gszMsgBox, "( working directory: %s )", curLocalDir);
      }
      Dialog("Please enter new file name: ( <CR>: accept, <ESC>: cancel )",
            gszMsgBox, new_file_name);
   }
   UtilTrimBlanks(new_file_name);
   if (*new_file_name == '\0') return;
   len = strlen(new_file_name);

   if (SaveSelectedOnly) {
      for (sel_ptr=topSel; sel_ptr != NULL; sel_ptr=sel_ptr->next) {
         if (sel_ptr->obj->type == OBJ_SYM) {
            if (obj_ptr1 == NULL) {
               obj_ptr1 = sel_ptr->obj;
            } else if (obj_ptr2 == NULL) {
               obj_ptr2 = sel_ptr->obj;
            }
            count++;
         }
      }
   } else {
      saved_cur_page = curPage;
      for (curPage=firstPage; curPage != NULL; curPage=curPage->next) {
         for (obj_ptr=curPage->top; obj_ptr != NULL; obj_ptr=obj_ptr->next) {
            if (obj_ptr->type == OBJ_SYM) {
               if (obj_ptr1 == NULL) {
                  obj_ptr1 = obj_ptr;
               } else if (obj_ptr2 == NULL) {
                  obj_ptr2 = obj_ptr;
               }
               count++;
            }
         }
      }
      curPage = saved_cur_page;
   }
   sprintf(obj_ext_str, ".%s", OBJ_FILE_EXT);
   sprintf(gz_obj_ext_str, ".%s.gz", OBJ_FILE_EXT);
   sprintf(sym_ext_str, ".%s", SYM_FILE_EXT);
   sprintf(pin_ext_str, ".%s", PIN_FILE_EXT);

   if (count > 2) {
      MsgBox("Too many symbols!\n\nFile not saved.", TOOL_NAME, INFO_MB);
      return;
   }
   if (count > 1 && lastPageNum != 1) {
      sprintf(gszMsgBox, "A %s file can only have one page.\n\nSave aborted!",
            (count == 1 ? "symbol" : "pin"));
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   switch (count) {
   case 0:
      if (FileNameHasExtension(new_file_name, OBJ_FILE_TYPE, &gzipped,
            &no_name)) {
         /* don't modify */
      } else if (FileNameHasExtension(new_file_name, SYM_FILE_TYPE, NULL,
            &no_name)) {
         MsgBox("Cannot save as a symbol file.\n\nNo symbol defined.",
               TOOL_NAME, INFO_MB);
         return;
      } else if (FileNameHasExtension(new_file_name, PIN_FILE_TYPE, NULL,
            &no_name)) {
         MsgBox("Cannot save as a pin file.\n\nNo symbol defined.",
               TOOL_NAME, INFO_MB);
         return;
      } else {
         strcat(new_file_name, obj_ext_str);
      }
      file_type = OBJ_FILE_TYPE;
      break;
   case 1:
      if ((attr_ptr=FindAttrWithName(obj_ptr1, "type=", NULL)) != NULL &&
            strcmp(attr_ptr->attr_value.s, "tgif_pin") == 0) {
         if (FileNameHasExtension(new_file_name, OBJ_FILE_TYPE, NULL,
               &no_name)) {
            MsgBox("Cannot save as an object file; one symbol defined.",
                  TOOL_NAME, INFO_MB);
            return;
         } else if (FileNameHasExtension(new_file_name, SYM_FILE_TYPE, NULL,
               &no_name)) {
            sprintf(gszMsgBox, "%s.\n\n%s.",
                  "Cannot save as a symbol file",
                  "Please save the file as a pin file");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            return;
         } else if (FileNameHasExtension(new_file_name, PIN_FILE_TYPE, NULL,
               &no_name)) {
            /* don't modify */
         } else {
            strcat(new_file_name, pin_ext_str);
         }
         file_type = PIN_FILE_TYPE;
      } else {
         if (FileNameHasExtension(new_file_name, OBJ_FILE_TYPE, NULL,
               &no_name)) {
            MsgBox("Cannot save as an object file; one symbol defined.",
                  TOOL_NAME, INFO_MB);
            return;
         } else if (FileNameHasExtension(new_file_name, SYM_FILE_TYPE, NULL,
               &no_name)) {
            /* don't modify */
         } else if (FileNameHasExtension(new_file_name, PIN_FILE_TYPE, NULL,
               &no_name)) {
            MsgBox("Cannot save as a pin file; only one symbol is defined.",
                  TOOL_NAME, INFO_MB);
            return;
         } else {
            strcat(new_file_name, sym_ext_str);
         }
         file_type = SYM_FILE_TYPE;
      }
      break;
   case 2:
      if (FileNameHasExtension(new_file_name, OBJ_FILE_TYPE, NULL, &no_name)) {
         MsgBox("Cannot save as an object file; two symbols defined.",
               TOOL_NAME, INFO_MB);
         return;
      } else if (FileNameHasExtension(new_file_name, SYM_FILE_TYPE, NULL,
            &no_name)) {
         MsgBox("Cannot save as a symbol file; two symbols defined.",
               TOOL_NAME, INFO_MB);
         return;
      } else if (FileNameHasExtension(new_file_name, PIN_FILE_TYPE, NULL,
            &no_name)) {
         /* don't modify */
      } else {
         strcat(new_file_name, pin_ext_str);
      }
      if (obj_ptr2->fattr != NULL) {
         obj_ptr = obj_ptr2;
         obj_ptr2 = obj_ptr1;
         obj_ptr1 = obj_ptr;
      }
      if ((attr_ptr=FindAttrWithName(obj_ptr1, "type=", NULL)) == NULL) {
         MsgBox("Cannot find an attribute name 'type'.\n\nPin file not saved.",
               TOOL_NAME, INFO_MB);
         return;
      }
      if (strcmp(attr_ptr->attr_value.s, "tgif_pin") != 0) {
         MsgBox("Symbol type is not 'tgif_pin'.\n\nPin file not saved.",
               TOOL_NAME, INFO_MB);
         return;
      }
      file_type = PIN_FILE_TYPE;
      break;
   default: return;
   }
   if (no_name) {
      MsgBox("No file name specified.\n\nFile not saved.",
            TOOL_NAME, INFO_MB);
      return;
   }
   if (*new_file_name == DIR_SEP) {
      strcpy(new_full_name, new_file_name);
   } else if (curDirIsLocal) {
      sprintf(new_full_name, "%s%c%s", curDir, DIR_SEP, new_file_name);
   } else {
      sprintf(new_full_name, "%s%c%s", curLocalDir, DIR_SEP, new_file_name);
   }
   if (file_type == OBJ_FILE_TYPE && gzipped) {
      char tmp_fname[MAXPATHLENGTH];

      sprintf(tmp_fname, "%sTgifXXXXXX", TMP_DIR);
      mktemp(tmp_fname);
      unlink(tmp_fname);
      strcpy(gzipped_fname, new_full_name);
      strcpy(new_full_name, tmp_fname);
   }
   if (!OkayToCreateFile((*gzipped_fname=='\0') ? new_full_name :
         gzipped_fname)) {
      if (*gzipped_fname != '\0') unlink(new_full_name);
      return;
   }
   if (*gzipped_fname == '\0') {
      if ((short_name=IsPrefix(bootDir, new_full_name, &rest))) ++rest;
   } else {
      if ((short_name=IsPrefix(bootDir, gzipped_fname, &rest))) ++rest;
   }
   if ((fp=fopen(new_full_name, "w")) == NULL) {
      if (short_name) {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", rest);
      } else {
         if (*gzipped_fname == '\0') {
            sprintf(gszMsgBox, "Cannot open '%s', file not saved.",
                  new_full_name);
         } else {
            sprintf(gszMsgBox, "Cannot open '%s', file not saved.",
                  gzipped_fname);
         }
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      if (*gzipped_fname != '\0') unlink(new_full_name);
      return;
   }
   if (!SaveSelectedOnly) BeforeNavigate();

   if (SaveSelectedOnly) {
      PushPageInfo();
      if (pageLayoutMode == PAGE_STACK) {
         paperCol = paperRow = curPageNum = lastPageNum = 1;
      }
      if (curDirIsLocal) {
         strcpy(saved_cur_dir, curDir);
      } else {
         strcpy(saved_cur_dir, curLocalDir);
      }
      strcpy(saved_cur_file_name, curFileName);
      saved_cur_file_defined = curFileDefined;
      saved_top_obj = topObj;
      saved_bot_obj = botObj;

      if (topSel == NULL) {
         topObj = botObj = NULL;
      } else {
         JustDupSelObj(&top_sel_ptr, &bot_sel_ptr);
         topObj = top_sel_ptr->obj;
         botObj = bot_sel_ptr->obj;
      }
      firstPage = lastPage = curPage =
            (struct PageRec *)malloc(sizeof(struct PageRec));
      if (firstPage == NULL) FailAllocMessage();
      memset(firstPage, 0, sizeof(struct PageRec));
      firstPage->layer_on = TRUE;
      firstPage->top = topObj;
      firstPage->bot = botObj;
      firstPage->next = firstPage->prev = NULL;
      if (pageLayoutMode == PAGE_STACK) {
         firstPage->draw_orig_x = drawOrigX;
         firstPage->draw_orig_y = drawOrigY;
         firstPage->zoom_scale = zoomScale;
         firstPage->zoomed_in = zoomedIn;
         curPageNum = lastPageNum = 1;
      }
      for (sel_ptr=topSel, obj_ptr=topObj; obj_ptr!=NULL;
            sel_ptr=sel_ptr->next, obj_ptr=obj_ptr->next) {
         CopyObjId(sel_ptr->obj, obj_ptr);
         CopyObjLocks(sel_ptr->obj, obj_ptr);
      }
   }

   if (curDirIsLocal) {
      strcpy(tmp_str, curDir);
   } else {
      strcpy(tmp_str, curLocalDir);
   }
   if (*gzipped_fname == '\0') {
      SetCurDir(new_full_name);
   } else {
      SetCurDir(gzipped_fname);
   }
   curFileDefined = TRUE;

   switch (count) {
   case 0:
      *curSymDir = '\0';
      if ((strcmp(tmp_str, (curDirIsLocal ? curDir : curLocalDir)) != 0) ||
            (!NameInCurDir(curFileName))) {
         UpdateDirInfo();
      }
      break;
   default:
      strcpy(curSymDir, (curDirIsLocal ? curDir : curLocalDir));
      if (!DirInSymPath(curDirIsLocal ? curDir : curLocalDir)) {
         UpdateSymInfo();
      }
      break;
   }
   if (short_name) {
      sprintf(gszMsgBox, "Saving '%s' ...", rest);
   } else {
      if (*gzipped_fname == '\0') {
         sprintf(gszMsgBox, "Saving '%s' ...", new_full_name);
      } else {
         sprintf(gszMsgBox, "Saving '%s' ...", gzipped_fname);
      }
   }
   Msg(gszMsgBox);

   if (!saveCommentsInSaveNew) CleanUpComments();
   watch_cursor = watchCursorOnMainWindow;
   if (!watch_cursor) {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
   }
   writeFileFailed = FALSE;
   saved_cur_page = curPage;
   for (curPage=firstPage, count=1; curPage != NULL;
         curPage=curPage->next, count++) {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save(fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose(fp);
   if (!watch_cursor) {
      SetDefaultCursor(mainWindow);
      ShowCursor(); 
   }
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      FailToWriteFileMessage(new_full_name);
      ok = FALSE;
   } else {
      if (*gzipped_fname != '\0') {
         if (!GzipFile(new_full_name, gzipped_fname)) {
            sprintf(gszMsgBox, "%s '%s' to '%s'.\n\n%s '%s'.  %s.",
                  "Problems encountered while (g)zipping", new_full_name,
                  gzipped_fname, "The unzipped file is", new_full_name,
                  "Please save a copy of it before proceeding");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            ok = FALSE;
         }
         unlink(new_full_name);
         strcpy(new_full_name, gzipped_fname);
      }
      if (ok) {
         if (short_name) {
            sprintf(gszMsgBox, "File '%s' saved.", rest);
         } else {
            sprintf(gszMsgBox, "File '%s' saved.", new_full_name);
         }
         Msg(gszMsgBox);
      } else {
         if (short_name) {
            sprintf(gszMsgBox, "File '%s' is NOT saved.", rest);
         } else {
            sprintf(gszMsgBox, "File '%s' is NOT saved.", new_full_name);
         }
         Msg(gszMsgBox);
      }
   }
   if (SaveSelectedOnly) {
      CleanUpPage();
      for (sel_ptr=top_sel_ptr; sel_ptr != NULL; sel_ptr=next_sel) {
         next_sel = sel_ptr->next;
         free(sel_ptr);
      }

      topObj = saved_top_obj;
      botObj = saved_bot_obj;
      if (curDirIsLocal) {
         strcpy(curDir, saved_cur_dir);
      } else {
         strcpy(curLocalDir, saved_cur_dir);
      }
      strcpy(curFileName, saved_cur_file_name);
      curFileDefined = saved_cur_file_defined;
      PopPageInfo();
   } else if (ok) {
      SetFileModified(FALSE);
      RedrawTitleWindow();
   }
   if (!SaveSelectedOnly) CommitNavigate();
}

void SaveSymInLibrary()
{
   char new_file_name[MAXPATHLENGTH+1];
   char new_full_name[MAXPATHLENGTH+1], dir_name[MAXPATHLENGTH+1];
   char saved_dir[MAXPATHLENGTH+1], saved_file[MAXPATHLENGTH+1];
   char saved_sym_dir[MAXPATHLENGTH+1], *rest=NULL;
   char sym_ext_str[MAXSTRING+1], pin_ext_str[MAXSTRING+1], *c_ptr;
   FILE *fp;
   int count=0, short_name, watch_cursor=FALSE;
   int saved_cur_file_defined, file_type=INVALID;
   struct ObjRec *obj_ptr;
   struct PageRec *saved_cur_page;
   struct ObjRec *obj_ptr1=NULL, *obj_ptr2=NULL;
   struct AttrRec *attr_ptr;

   saved_cur_page = curPage;
   for (curPage=firstPage; curPage != NULL; curPage=curPage->next) {
      for (obj_ptr=curPage->top; obj_ptr != NULL; obj_ptr=obj_ptr->next) {
         if (obj_ptr->type == OBJ_SYM) {
            if (obj_ptr1 == NULL) {
               obj_ptr1 = obj_ptr;
            } else if (obj_ptr2 == NULL) {
               obj_ptr2 = obj_ptr;
            }
            count++;
         }
      }
   }
   curPage = saved_cur_page;

   switch (count) {
   case 0:
      MsgBox("No symbol found!\n\nSymbol file not saved.",
            TOOL_NAME, INFO_MB);
      return;
   case 1:
      if (lastPageNum != 1) {
         MsgBox("A symbol file can only have one page.\n\nSave aborted!",
               TOOL_NAME, INFO_MB);
         return;
      }
      if ((attr_ptr=FindAttrWithName(obj_ptr1, "type=", NULL)) != NULL &&
            strcmp(attr_ptr->attr_value.s, "tgif_pin") == 0) {
         file_type = PIN_FILE_TYPE;
      } else {
         file_type = SYM_FILE_TYPE;
      }
      break;
   case 2:
      if (lastPageNum != 1) {
         MsgBox("A pin file can only have one page.\n\nSave aborted!",
               TOOL_NAME, INFO_MB);
         return;
      }
      if (obj_ptr2->fattr != NULL) {
         obj_ptr = obj_ptr2;
         obj_ptr2 = obj_ptr1;
         obj_ptr1 = obj_ptr;
      }
      if ((attr_ptr=FindAttrWithName(obj_ptr1, "type=", NULL)) == NULL) {
         MsgBox("Cannot find an attribute name 'type'.\n\nPin file not saved.",
               TOOL_NAME, INFO_MB);
         return;
      }
      if (strcmp(attr_ptr->attr_value.s, "tgif_pin") != 0) {
         MsgBox("Symbol type is not 'tgif_pin'.\n\nPin file not saved.",
               TOOL_NAME, INFO_MB);
         return;
      }
      file_type = PIN_FILE_TYPE;
      break;
   default:
      MsgBox("Too many symbols!\n\nSymbol file not saved.",
            TOOL_NAME, INFO_MB);
      return;
   }
   MakeQuiescent();

   sprintf(sym_ext_str, ".%s", SYM_FILE_EXT);
   sprintf(pin_ext_str, ".%s", PIN_FILE_EXT);

   if (*curFileName == '\0') {
      int no_name=FALSE;

      Dialog("Please enter new file name:", "( <CR>: accept, <ESC>: cancel )",
            new_file_name);
      UtilTrimBlanks(new_file_name);
      if (*new_file_name == '\0') return;

      if (FileNameHasExtension(new_file_name, SYM_FILE_TYPE, NULL, &no_name)) {
         /* do nothing */
      } else if (FileNameHasExtension(new_file_name, PIN_FILE_TYPE, NULL,
            &no_name)) {
         /* do nothing */
      } else if (count == 1) {
         if (file_type == SYM_FILE_TYPE) {
            strcat(new_file_name, sym_ext_str);
         } else {
            strcat(new_file_name, pin_ext_str);
         }
      } else {
         strcat(new_file_name, pin_ext_str);
      }
      if (no_name) {
         MsgBox("No file name specified.\n\nFile not saved.",
               TOOL_NAME, INFO_MB);
         return;
      }
   } else {
      c_ptr = UtilStrRChr(curFileName, DIR_SEP);
      strcpy(new_file_name, (c_ptr==NULL) ? curFileName : ++c_ptr);
   }
   if (SelectSymDir(dir_name) == INVALID) {
      Msg("");
      return;
   }
   if (strcmp(dir_name, ".") == 0) {
      sprintf(new_full_name, "%s%c%s", (curDirIsLocal ? curDir : curLocalDir),
            DIR_SEP, new_file_name);
   } else {
      sprintf(new_full_name, "%s%c%s", dir_name, DIR_SEP, new_file_name);
   }

   if (!OkayToCreateFile(new_full_name)) return;

   if ((short_name=IsPrefix(bootDir, new_full_name, &rest))) ++rest;
   if ((fp=fopen(new_full_name, "w")) == NULL) {
      if (short_name) {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", rest);
      } else {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", new_full_name);
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }

   strcpy(saved_dir, (curDirIsLocal ? curDir : curLocalDir));
   strcpy(saved_file, curFileName);
   strcpy(saved_sym_dir, curSymDir);
   saved_cur_file_defined = curFileDefined;
   SetCurDir(new_full_name);
   curFileDefined = TRUE;

   strcpy(curSymDir, (curDirIsLocal ? curDir : curLocalDir));
   if (!DirInSymPath(curDirIsLocal ? curDir : curLocalDir)) UpdateSymInfo();

   if (short_name) {
      sprintf(gszMsgBox, "Saving '%s' ...", rest);
   } else {
      sprintf(gszMsgBox, "Saving '%s' ...", new_full_name);
   }
   Msg(gszMsgBox);

   if (!saveCommentsInSaveNew) CleanUpComments();
   watch_cursor = watchCursorOnMainWindow;
   if (!watch_cursor) {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
   }
   writeFileFailed = FALSE;
   saved_cur_page = curPage;
   for (curPage=firstPage, count=1; curPage != NULL;
         curPage=curPage->next, count++) {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save(fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose(fp);
   if (!watch_cursor) {
      SetDefaultCursor(mainWindow);
      ShowCursor(); 
   }
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            new_full_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   } else {
      if (short_name) {
         sprintf(gszMsgBox, "File '%s' saved.", rest);
      } else {
         sprintf(gszMsgBox, "File '%s' saved.", new_full_name);
      }
      Msg(gszMsgBox);
   }

   if (curDirIsLocal) {
      strcpy(curDir, saved_dir);
   } else {
      strcpy(curLocalDir, saved_dir);
   }
   strcpy(curFileName, saved_file);
   strcpy(curSymDir, saved_sym_dir);
   curFileDefined = saved_cur_file_defined;

   RedrawTitleWindow();
}

void SaveFile()
{
   int count=0, short_name, file_type=INVALID, no_name=FALSE;
   struct ObjRec *obj_ptr=NULL;
   FILE *fp=NULL;
   char ext[MAXPATHLENGTH+1], gzipped_fname[MAXPATHLENGTH+1];
   char full_name[MAXPATHLENGTH+1], *rest=NULL, *psz=NULL;
   struct PageRec *saved_cur_page=NULL;
   struct ObjRec *obj_ptr1=NULL, *obj_ptr2=NULL;
   struct AttrRec *attr_ptr=NULL;
   int watch_cursor=FALSE, gzipped=FALSE;

   *gzipped_fname = '\0';
   if (!curFileDefined || !curDirIsLocal) {
      SaveNewFile(FALSE, NULL);
      return;
   }
   if ((psz=UtilStrRChr(curFileName, '.')) == NULL) {
      TgAssert(FALSE, "Cannot find '.' in curFileName in SaveFile()", NULL);
      return;
   }
   if (strcmp(psz, ".gz") == 0) {
      char *dot_ptr=psz;

      gzipped = TRUE;
      *dot_ptr = '\0';
      if ((psz=UtilStrRChr(curFileName, '.')) == NULL) {
         TgAssert(FALSE, "Cannot find '.' in curFileName in SaveFile()", NULL);
         *dot_ptr = '.';
         return;
      }
      UtilStrCpyN(ext, sizeof(ext), psz);
      *dot_ptr = '.';
   } else {
      UtilStrCpyN(ext, sizeof(ext), psz);
   }
   saved_cur_page = curPage;
   for (curPage=firstPage; curPage != NULL; curPage=curPage->next) {
      for (obj_ptr=curPage->top; obj_ptr != NULL; obj_ptr=obj_ptr->next) {
         if (obj_ptr->type == OBJ_SYM) {
            if (obj_ptr1 == NULL) {
               obj_ptr1 = obj_ptr;
            } else if (obj_ptr2 == NULL) {
               obj_ptr2 = obj_ptr;
            }
            count++;
         }
      }
   }
   curPage = saved_cur_page;

   switch (count) {
   case 0:
      if (FileNameHasExtension(ext, SYM_FILE_TYPE, NULL, &no_name)) {
         sprintf(gszMsgBox, "%s.\n\n%s.",
               "No symbol defined in a symbol file", "Symbol file not saved");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return;
      } else if (FileNameHasExtension(ext, PIN_FILE_TYPE, NULL, &no_name)) {
         sprintf(gszMsgBox, "%s.\n\n%s.",
               "No symbol defined in a pin file", "Pin file not saved");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return;
      }
      file_type = OBJ_FILE_TYPE;
      break;
   case 1:
      if (FileNameHasExtension(ext, OBJ_FILE_TYPE, NULL, &no_name)) {
         sprintf(gszMsgBox, "%s.\n\n%s.",
               "One symbol defined in OBJECT file", "Object file not saved");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return;
      }
      if ((attr_ptr=FindAttrWithName(obj_ptr1, "type=", NULL)) != NULL &&
            strcmp(attr_ptr->attr_value.s, "tgif_pin") == 0) {
         if (FileNameHasExtension(ext, SYM_FILE_TYPE, NULL, &no_name)) {
            sprintf(gszMsgBox, "%s.\n\n%s.",
                  "The symbol is of type tgif_pin",
                  "It should be saved as a pin file");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            return;
         }
         file_type = PIN_FILE_TYPE;
      } else {
         if (FileNameHasExtension(ext, PIN_FILE_TYPE, NULL, &no_name)) {
            sprintf(gszMsgBox, "%s.\n\n%s.",
                  "Only one symbol defined in a pin file",
                  "Pin file not saved");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            return;
         }
         file_type = SYM_FILE_TYPE;
      }
      break;
   case 2:
      if (FileNameHasExtension(ext, OBJ_FILE_TYPE, NULL, &no_name)) {
         sprintf(gszMsgBox, "%s.\n\n%s.",
               "One symbol defined in OBJECT file", "Object file not saved");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return;
      } else if (FileNameHasExtension(ext, SYM_FILE_TYPE, NULL, &no_name)) {
         sprintf(gszMsgBox, "%s.\n\n%s.",
               "Too many symbols defined in a symbol file",
               "Symbol file not saved");
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         return;
      }
      file_type = PIN_FILE_TYPE;
      break;
   default:
      sprintf(gszMsgBox, "%s!\n\n%s.",
            "Too many symbols in a file", "File not saved");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   if (file_type == SYM_FILE_TYPE || file_type == PIN_FILE_TYPE) {
      sprintf(full_name, "%s%c%s", curSymDir, DIR_SEP, curFileName);
   } else {
      sprintf(full_name, "%s%c%s", curDir, DIR_SEP, curFileName);
   }
   if (file_type == OBJ_FILE_TYPE && gzipped) {
      char tmp_fname[MAXPATHLENGTH];

      sprintf(tmp_fname, "%sTgifXXXXXX", TMP_DIR);
      mktemp(tmp_fname);
      unlink(tmp_fname);
      strcpy(gzipped_fname, full_name);
      strcpy(full_name, tmp_fname);
   }
   if (*gzipped_fname == '\0') {
      if ((short_name=IsPrefix(bootDir, full_name, &rest))) ++rest;
   } else {
      if ((short_name=IsPrefix(bootDir, gzipped_fname, &rest))) ++rest;
   }
   if ((fp=fopen(full_name, "w")) == NULL) {
      if (short_name) {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", rest);
      } else {
         sprintf(gszMsgBox, "Cannot open '%s', file not saved.", full_name);
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      if (*gzipped_fname != '\0') unlink(full_name);
      return;
   }
   if (short_name) {
      sprintf(gszMsgBox, "Saving '%s' ...", rest);
   } else {
      if (*gzipped_fname == '\0') {
         sprintf(gszMsgBox, "Saving '%s' ...", full_name);
      } else {
         sprintf(gszMsgBox, "Saving '%s' ...", gzipped_fname);
      }
   }
   Msg(gszMsgBox);

   watch_cursor = watchCursorOnMainWindow;
   if (!watch_cursor) {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
   }
   writeFileFailed = FALSE;
   MakeQuiescent();
   saved_cur_page = curPage;
   for (curPage=firstPage, count=1; curPage != NULL;
         curPage=curPage->next, count++) {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save(fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;

   fclose(fp);

   if (!watch_cursor) {
      SetDefaultCursor(mainWindow);
      ShowCursor();
   }
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            full_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   } else {
      if (*gzipped_fname != '\0') {
         if (!GzipFile(full_name, gzipped_fname)) {
            sprintf(gszMsgBox, "%s '%s' to '%s'.\n\n%s '%s'.  %s.",
                  "Problems encountered while (g)zipping", full_name,
                  gzipped_fname, "The unzipped file is", full_name,
                  "Please save a copy of it before proceeding");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            unlink(full_name);
            return;
         }
         unlink(full_name);
         strcpy(full_name, gzipped_fname);
      }
      if (short_name) {
         sprintf(gszMsgBox, "File '%s' saved.", rest);
      } else {
         sprintf(gszMsgBox, "File '%s' saved.", full_name);
      }
      Msg(gszMsgBox);
      SetFileModified(FALSE);
      if (!NameInCurDir(curFileName)) UpdateDirInfo();
   }
}

char *ParseStr(Str, C, Left, LeftSz)
   char *Str, *Left;
   int C, LeftSz;
{
   register char *s=Str, *l=Left;
   register int len=0;
   int max_len=LeftSz-1;
   char the_char=(char)C;

   while (*s != '\0' && *s != the_char) {
      if (len < max_len) {
         *l++ = *s++;
         len++;
      } else {
         break;
      }
   }

   if (*s == the_char) s++;
   *l = '\0';

   while (len >= 2 && *Left == '\'' && *(--l) == '\'') {
      char *c_ptr, *c_ptr1;

      *l-- = '\0';
      len -= 2;
      c_ptr = &Left[1];
      c_ptr1 = Left;
      while (*c_ptr != '\0') *c_ptr1++ = *c_ptr++;
      *c_ptr1 = '\0';
   }
   return s;
}

char *FindChar(C, Str)
   int C;
   char *Str;
   /* returns the address of the character right after C of the string Str */
{
   register char *s=Str, the_char=(char)C;

   while (*s != '\0' && *s != the_char) s++;

   if (*s == the_char) s++;
   return s;
}

void SaveCreatorID(FP, obj_ptr, psz_prefix)
   FILE *FP;
   struct ObjRec *obj_ptr;
   char *psz_prefix;
{
   char buf[MAXSTRING];

   if (obj_ptr->creator_full_id == NULL) {
      sprintf(buf, "%1d/%s", obj_ptr->id, gszLocalPID);
   } else {
      strcpy(buf, obj_ptr->creator_full_id);
   }
   if (fprintf(FP, "\n%s\"", psz_prefix) == EOF) writeFileFailed = TRUE;
   SaveString(FP, buf);
   if (fprintf(FP, "\",") == EOF) writeFileFailed = TRUE;
}

int ReadCreatorID(FP, ObjPtr)
   FILE *FP;
   struct ObjRec **ObjPtr;
{
   int ok=TRUE;
   char *c_ptr=NULL, *line=NULL, full_id[MAXSTRING];

   if ((line=UtilGetALine(FP)) == NULL) {
      sprintf(gszMsgBox, "%s, %d:  EOF in ReadCreatorID().  Read aborted!",
            scanFileName, scanLineNum);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return FALSE;
   }
   scanLineNum++;
   c_ptr = FindChar((int)'"', line);
   c_ptr = ParseStr(c_ptr, (int)'"', full_id, sizeof(full_id));
   if (c_ptr == NULL) {
      sprintf(gszMsgBox, "%s, %d:  Invalid creator_full_id in %s.  %s!",
            scanFileName, scanLineNum, "ReadCreatorID()", "Read aborted");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      ok = FALSE;
   }
   (*ObjPtr)->creator_full_id = UtilStrDup(full_id);
   if ((*ObjPtr)->creator_full_id == NULL) FailAllocMessage();
   free(line);
   return ok;
}

static int gnCannotFindColorMsg=FALSE;

static
int ReadState(Inbuf)
   char *Inbuf;
{
   char *s, font_str[MAXSTRING], sb_font_str[MAXSTRING];
   int page_style, forced_use_gray=FALSE, compat_dpi, font_sz=0, sz_unit=0;
   int page_arg1=0, page_arg2=0, rotate=ROTATE0;
   int one_page_width=0, one_page_height=0;

   *font_str = *sb_font_str = '\0';
   s = FindChar((int)'(', Inbuf);
   if (sscanf(s, "%d", &page_style) != 1) return FALSE;
   s = FindChar((int)',', s);
   if (*s == '\0') {
      fileVersion = INVALID;
   } else if (sscanf(s, "%d", &fileVersion) != 1) {
      return FALSE;
   }

   if (fileVersion > CUR_VERSION) return FALSE;

   gnCannotFindColorMsg = FALSE;

   if (!importingFile) {
      if (fileVersion <= 13) {
         switch (page_style) {
         case PORTRAIT: printMag = (float)100.0; break;
         case LANDSCAPE: printMag = (float)100.0; break;
         case HIGHPORT: printMag = (float)50.0; page_style = PORTRAIT; break;
         case HIGHLAND: printMag = (float)50.0; page_style = LANDSCAPE; break;
         case SLIDEPORT: printMag = (float)200.0; page_style = PORTRAIT; break;
         case SLIDELAND: printMag = (float)200.0; page_style = LANDSCAPE; break;
         default:
            sprintf(gszMsgBox, "Unrecognizable page style '%1d'.",
                  page_style);
            TwoLineMsg(gszMsgBox, "Portrait style assumed.");
            page_style = PORTRAIT;
            printMag = (float)100.0;
            break;
         }
      } else {
         if (page_style != PORTRAIT && page_style != LANDSCAPE) {
            sprintf(gszMsgBox, "Unrecognizable page style '%1d'.", page_style);
            TwoLineMsg(gszMsgBox, "Portrait style assumed.");
            page_style = PORTRAIT;
         }
         s = FindChar((int)',', s);
         sscanf(s, "%f", &printMag);
      }
      pageStyle = page_style;
   }

   if (PRTGIF && useGray) forced_use_gray = TRUE;
   if (importingFile) return TRUE;

   page_arg1 = page_arg2 = 1;
   one_page_width = onePageWidth;
   one_page_height = onePageHeight;

   if (fileVersion >= 2) {
      compat_dpi = FONT_DPI_75;
      curDash = 0;
      gridSystem = ENGLISH_GRID;
      xyMetricGrid = DEFAULT_METRIC_GRID;
      textVSpace = 0;
      zoomedIn = FALSE;
      rcbRadius = DEF_RCB_RADIUS;
      pageLayoutMode = PAGE_STACK;
      paperCol = paperRow = 1;
      curPageNum = lastPageNum = 1;
      pageLineShownInTileMode = TRUE;

      if (usePaperSizeStoredInFile) ResetOnePageSize();

      s = FindChar((int)',', s);

      InitScan(s, "\t\n, ");

      if (fileVersion <= 3) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");

         if (lineWidth == LINE_CURVED) {
            lineWidth = 0;
            curSpline = LT_SPLINE;
         } else {
            curSpline = LT_STRAIGHT;
         }
      } else if (fileVersion <= 7) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
      } else if (fileVersion <= 8) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
      } else if (fileVersion <= 11) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
      } else if (fileVersion <= 12) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
      } else if (fileVersion <= 18) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
      } else if (fileVersion <= 19) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
      } else if (fileVersion <= 21) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
      } else if (fileVersion <= 26) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
         GETINT("state", rotate,        "Text Rotation");
         GETINT("state", rcbRadius,     "RCBox Radius");
      } else if (fileVersion <= 27) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
         GETINT("state", rotate,        "Text Rotation");
         GETINT("state", rcbRadius,     "RCBox Radius");
         GETINT("state", useGray,       "Use Gray Scale");
      } else if (fileVersion <= 28) {
         /* Matsuda's Version */
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
         GETINT("state", rotate,        "Text Rotation");
         GETINT("state", rcbRadius,     "RCBox Radius");
         GETINT("state", useGray,       "Use Gray Scale");
         GETINT("state", paperCol,      "Num Columns");
         GETINT("state", paperRow,      "Num Rows");
         GETINT("state", curPageNum,    "Current Page Number");
         GETINT("state", lastPageNum,   "Last Page Number");
      } else if (fileVersion <= 29) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETINT("state", curFont,       "Font Name");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
         GETINT("state", rotate,        "Text Rotation");
         GETINT("state", rcbRadius,     "RCBox Radius");
         GETINT("state", useGray,       "Use Gray Scale");
         GETINT("state", pageLayoutMode,"Page Layout Mode");
         GETINT("state", page_arg1,     "Page Layout Subarg 1");
         GETINT("state", page_arg2,     "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
      } else if (fileVersion <= 30) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETSTR("state", font_str,      "Font Name String");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
         GETINT("state", rotate,        "Text Rotation");
         GETINT("state", rcbRadius,     "RCBox Radius");
         GETINT("state", useGray,       "Use Gray Scale");
         GETINT("state", pageLayoutMode,"Page Layout Mode");
         GETINT("state", page_arg1,     "Page Layout Subarg 1");
         GETINT("state", page_arg2,     "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
      } else if (fileVersion <= 31) {
         GETINT("state", drawOrigX,     "X Draw Origin");
         GETINT("state", drawOrigY,     "Y Draw Origin");
         GETINT("state", zoomScale,     "Zoom scale");
         GETINT("state", xyEnglishGrid, "English Grid");
         GETINT("state", snapOn,        "Grid");
         GETINT("state", colorIndex,    "Color");
         GETINT("state", horiAlign,     "Horizontal Align");
         GETINT("state", vertAlign,     "Vertical Align");
         GETINT("state", lineWidth,     "Line Width");
         GETINT("state", curSpline,     "Spline");
         GETINT("state", lineStyle,     "Line Style");
         GETINT("state", objFill,       "Fill Pattern");
         GETINT("state", penPat,        "Pen Pattern");
         GETINT("state", textJust,      "Text Justify");
         GETSTR("state", font_str,      "Font Name String");
         GETINT("state", curStyle,      "Font Style");
         GETINT("state", font_sz,       "Font Size");
         GETINT("state", compat_dpi,    "Font DPI");
         GETINT("state", curDash,       "Dash Style");
         GETINT("state", gridSystem,    "Grid System");
         GETINT("state", xyMetricGrid,  "Metric Grid");
         GETINT("state", textVSpace,    "Text Vertical Spacing");
         GETINT("state", zoomedIn,      "Zoomed In");
         GETINT("state", gridShown,     "Grid Shown");
         GETINT("state", moveMode,      "Move Mode");
         GETINT("state", rotate,        "Text Rotation");
         GETINT("state", rcbRadius,     "RCBox Radius");
         GETINT("state", useGray,       "Use Gray Scale");
         GETINT("state", pageLayoutMode,"Page Layout Mode");
         GETINT("state", page_arg1,     "Page Layout Subarg 1");
         GETINT("state", page_arg2,     "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
         GETINT("state", colorDump,     "Print In Color");
      } else if (fileVersion <= 32) {
         GETINT("state", drawOrigX,      "X Draw Origin");
         GETINT("state", drawOrigY,      "Y Draw Origin");
         GETINT("state", zoomScale,      "Zoom scale");
         GETINT("state", xyEnglishGrid,  "English Grid");
         GETINT("state", snapOn,         "Grid");
         GETINT("state", colorIndex,     "Color");
         GETINT("state", horiAlign,      "Horizontal Align");
         GETINT("state", vertAlign,      "Vertical Align");
         GETINT("state", lineWidth,      "Line Width");
         GETINT("state", curSpline,      "Spline");
         GETINT("state", lineStyle,      "Line Style");
         GETINT("state", objFill,        "Fill Pattern");
         GETINT("state", penPat,         "Pen Pattern");
         GETINT("state", textJust,       "Text Justify");
         GETSTR("state", font_str,       "Font Name String");
         GETINT("state", curStyle,       "Font Style");
         GETINT("state", font_sz,        "Font Size");
         GETINT("state", compat_dpi,     "Font DPI");
         GETINT("state", curDash,        "Dash Style");
         GETINT("state", gridSystem,     "Grid System");
         GETINT("state", xyMetricGrid,   "Metric Grid");
         GETINT("state", textVSpace,     "Text Vertical Spacing");
         GETINT("state", zoomedIn,       "Zoomed In");
         GETINT("state", gridShown,      "Grid Shown");
         GETINT("state", moveMode,       "Move Mode");
         GETINT("state", rotate,         "Text Rotation");
         GETINT("state", rcbRadius,      "RCBox Radius");
         GETINT("state", useGray,        "Use Gray Scale");
         GETINT("state", pageLayoutMode, "Page Layout Mode");
         GETINT("state", page_arg1,      "Page Layout Subarg 1");
         GETINT("state", page_arg2,      "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
         GETINT("state", colorDump,      "Print In Color");
         GETINT("state", one_page_width, "One Page Width");
         GETINT("state", one_page_height,"One Page Height");
      } else if (fileVersion <= 34) {
         GETINT("state", drawOrigX,      "X Draw Origin");
         GETINT("state", drawOrigY,      "Y Draw Origin");
         GETINT("state", zoomScale,      "Zoom scale");
         GETINT("state", xyEnglishGrid,  "English Grid");
         GETINT("state", snapOn,         "Grid");
         GETINT("state", colorIndex,     "Color");
         GETINT("state", horiAlign,      "Horizontal Align");
         GETINT("state", vertAlign,      "Vertical Align");
         GETINT("state", lineWidth,      "Line Width");
         GETINT("state", curSpline,      "Spline");
         GETINT("state", lineStyle,      "Line Style");
         GETINT("state", objFill,        "Fill Pattern");
         GETINT("state", penPat,         "Pen Pattern");
         GETINT("state", textJust,       "Text Justify");
         GETSTR("state", font_str,       "Font Name String");
         GETINT("state", curStyle,       "Font Style");
         GETINT("state", font_sz,        "Font Size");
         GETINT("state", compat_dpi,     "Font DPI");
         GETINT("state", curDash,        "Dash Style");
         GETINT("state", gridSystem,     "Grid System");
         GETINT("state", xyMetricGrid,   "Metric Grid");
         GETINT("state", textVSpace,     "Text Vertical Spacing");
         GETINT("state", zoomedIn,       "Zoomed In");
         GETINT("state", gridShown,      "Grid Shown");
         GETINT("state", moveMode,       "Move Mode");
         GETINT("state", rotate,         "Text Rotation");
         GETINT("state", rcbRadius,      "RCBox Radius");
         GETINT("state", useGray,        "Use Gray Scale");
         GETINT("state", pageLayoutMode, "Page Layout Mode");
         GETINT("state", page_arg1,      "Page Layout Subarg 1");
         GETINT("state", page_arg2,      "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
         GETINT("state", colorDump,      "Print In Color");
         GETINT("state", one_page_width, "One Page Width");
         GETINT("state", one_page_height,"One Page Height");
         GETINT("state", stretchableText,"Stretchable Text");
         GETINT("state", textRotation,   "Text Rotation");
         GETINT("state", rotationIncrement,"Rotation Increment");
      } else if (fileVersion <= 35) {
         GETINT("state", drawOrigX,      "X Draw Origin");
         GETINT("state", drawOrigY,      "Y Draw Origin");
         GETINT("state", zoomScale,      "Zoom scale");
         GETINT("state", xyEnglishGrid,  "English Grid");
         GETINT("state", snapOn,         "Grid");
         GETINT("state", colorIndex,     "Color");
         GETINT("state", horiAlign,      "Horizontal Align");
         GETINT("state", vertAlign,      "Vertical Align");
         GETINT("state", lineWidth,      "Line Width");
         GETINT("state", curSpline,      "Spline");
         GETINT("state", lineStyle,      "Line Style");
         GETINT("state", objFill,        "Fill Pattern");
         GETINT("state", penPat,         "Pen Pattern");
         GETINT("state", textJust,       "Text Justify");
         GETSTR("state", font_str,       "Font Name String");
         GETINT("state", curStyle,       "Font Style");
         GETINT("state", font_sz,        "Font Size");
         GETINT("state", compat_dpi,     "Font DPI");
         GETINT("state", curDash,        "Dash Style");
         GETINT("state", gridSystem,     "Grid System");
         GETINT("state", xyMetricGrid,   "Metric Grid");
         GETINT("state", textVSpace,     "Text Vertical Spacing");
         GETINT("state", zoomedIn,       "Zoomed In");
         GETINT("state", gridShown,      "Grid Shown");
         GETINT("state", moveMode,       "Move Mode");
         GETINT("state", rotate,         "Text Rotation");
         GETINT("state", rcbRadius,      "RCBox Radius");
         GETINT("state", useGray,        "Use Gray Scale");
         GETINT("state", pageLayoutMode, "Page Layout Mode");
         GETINT("state", page_arg1,      "Page Layout Subarg 1");
         GETINT("state", page_arg2,      "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
         GETINT("state", colorDump,      "Print In Color");
         GETINT("state", one_page_width, "One Page Width");
         GETINT("state", one_page_height,"One Page Height");
         GETINT("state", stretchableText,"Stretchable Text");
         GETINT("state", textRotation,   "Text Rotation");
         GETINT("state", rotationIncrement,"Rotation Increment");
         GETINT("state", transPat,       "Pattern Transparency");
      } else {
         GETINT("state", drawOrigX,      "X Draw Origin");
         GETINT("state", drawOrigY,      "Y Draw Origin");
         GETINT("state", zoomScale,      "Zoom scale");
         GETINT("state", xyEnglishGrid,  "English Grid");
         GETINT("state", snapOn,         "Grid");
         GETINT("state", colorIndex,     "Color");
         GETINT("state", horiAlign,      "Horizontal Align");
         GETINT("state", vertAlign,      "Vertical Align");
         GETINT("state", lineWidth,      "Line Width");
         GETINT("state", curSpline,      "Spline");
         GETINT("state", lineStyle,      "Line Style");
         GETINT("state", objFill,        "Fill Pattern");
         GETINT("state", penPat,         "Pen Pattern");
         GETINT("state", textJust,       "Text Justify");
         GETSTR("state", font_str,       "Font Name String");
         GETINT("state", curStyle,       "Font Style");
         GETINT("state", sz_unit,        "Font Size Unit");
         GETINT("state", compat_dpi,     "Font DPI");
         GETINT("state", curDash,        "Dash Style");
         GETINT("state", gridSystem,     "Grid System");
         GETINT("state", xyMetricGrid,   "Metric Grid");
         GETINT("state", textVSpace,     "Text Vertical Spacing");
         GETINT("state", zoomedIn,       "Zoomed In");
         GETINT("state", gridShown,      "Grid Shown");
         GETINT("state", moveMode,       "Move Mode");
         GETINT("state", rotate,         "Text Rotation");
         GETINT("state", rcbRadius,      "RCBox Radius");
         GETINT("state", useGray,        "Use Gray Scale");
         GETINT("state", pageLayoutMode, "Page Layout Mode");
         GETINT("state", page_arg1,      "Page Layout Subarg 1");
         GETINT("state", page_arg2,      "Page Layout Subarg 2");
         GETINT("state", pageLineShownInTileMode,"Page Lines Shown");
         GETINT("state", colorDump,      "Print In Color");
         GETINT("state", one_page_width, "One Page Width");
         GETINT("state", one_page_height,"One Page Height");
         GETINT("state", stretchableText,"Stretchable Text");
         GETINT("state", textRotation,   "Text Rotation");
         GETINT("state", rotationIncrement,"Rotation Increment");
         GETINT("state", transPat,       "Pattern Transparency");
      }
      if (fileVersion <= 28) readingPageNum++;
      if (fileVersion <= 29) {
         VerifyCompatibleFontIndex(&curFont);
         font_sz = GetCompatibleSize(compat_dpi, font_sz);
         curSzUnit = FontSizeToSzUnit(font_sz);
         if (PRTGIF) {
            PrTgifInitBaseFonts();
         }
      } else {
         char *psz=NULL;
         int len=strlen(font_str);

         if (fileVersion <= 35) {
            curSzUnit = FontSizeToSzUnit(font_sz);
         } else {
            curSzUnit = sz_unit;
         }
         if (len > 1 && *font_str == '\'' && font_str[len-1] == '\'') {
            char *psz1=(&font_str[1]);

            font_str[len-1] = '\0';
            psz = font_str;
            while (*psz1 != '\0') *psz++ = *psz1++;
            *psz = '\0';
         }
         if ((psz=strchr(font_str, '%')) != NULL) {
            *psz++ = '\0';
            /* Ex., ignore last '%' if font_str eq "Courier%". */
            if (*psz != '\0') {
               strcpy(sb_font_str, font_str);
               strcpy(font_str, psz);
            }
         }
         curFont = GetFontIndex(font_str, curStyle, TRUE);
         if (curFont == INVALID && !PRTGIF) {
            sprintf(gszMsgBox, "Cannot find screen font for '%s'.  %s.",
                  font_str, "Times font is used instead");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            curFont = FONT_TIM;
         }
      }
      switch (pageLayoutMode) {
      case PAGE_STACK:
         curPageNum = page_arg1;
         lastPageNum = page_arg2;
         paperCol = paperRow = 1;
         pageLineShownInTileMode = TRUE;
         break;
      case PAGE_TILE:
         paperCol = page_arg1;
         paperRow = page_arg2;
         curPageNum = lastPageNum = 1;
         break;
      }
      loadedCurPageNum = curPageNum;
      InitPage();
      curPageNum = loadedCurPageNum;

      if (fileVersion <= 13) {
         switch (gridSystem) {
         case ENGLISH_GRID:
            drawOrigX += HALF_INCH;
            drawOrigY += HALF_INCH;
            break;
         case METRIC_GRID:
            drawOrigX += 2.5*ONE_CM;
            drawOrigY += 2.5*ONE_CM;
            break;
         }
      }
      if ((usePaperSizeStoredInFile || PRTGIF) && fileVersion >= 32) {
         if (pageStyle == LANDSCAPE) {
            onePageWidth = one_page_height;
            onePageHeight = one_page_width;
         } else {
            onePageWidth = one_page_width;
            onePageHeight = one_page_height;
         }
         SetPSPageWidthHeight();
         if (PRTGIF) {
            UpdPageStyle(pageStyle);
         }
      }
      if (PRTGIF && !cmdLineOpenDisplay) {
         if (forced_use_gray) useGray = TRUE;
         return TRUE;
      }
      if (rotate != 0) {
         switch (rotate) {
         case ROTATE0: textRotation = 0; break;
         case ROTATE90: textRotation = (90<<6); break;
         case ROTATE180: textRotation = (180<<6); break;
         case ROTATE270: textRotation = (270<<6); break;
         }
         rotationIncrement = (90<<6);
         rotate = ROTATE0;
      }
      if (colorIndex >= maxColors) {
         fprintf(stderr, "%s #%1d, use '%s' %s.\n",
               "In reading state, cannot find color", colorIndex,
               colorMenuItems[defaultColorIndex], "as the current color");
         colorIndex = defaultColorIndex;
         gnCannotFindColorMsg = TRUE;
      }
      SetCanvasFont();
   }
   UpdPageStyle(pageStyle);
   if (PRTGIF && !cmdLineOpenDisplay) return TRUE;

   if (lineWidth >= maxLineWidths) {
      fprintf(stderr, "%s '%1d' is out of range!  Set to 0.\n",
            "File's linewidth index", lineWidth);
      lineWidth = 0;
   }

   UpdDrawWinWH();
   RedrawScrollBars();
   ShowPage();
   ShowPageLayout();
   UpdDrawWinBBox();

   SetDefaultDrawWinClipRecs();

   DrawPaperBoundary(drawWindow);
   RedrawGridLines(drawWindow);
   RedrawPageLines(drawWindow);
   RedrawRulers();
   RedrawChoiceWindow();

   return TRUE;
}

static
void ReadObjAttrs(MinFileVersion, FP, ObjPtr)
   int MinFileVersion;
   FILE *FP;
   struct ObjRec **ObjPtr;
{
   struct AttrRec *top_attr=NULL, *bot_attr=NULL, *attr_ptr;

   if (fileVersion <= MinFileVersion) return;

   while (ReadAttr(FP, &attr_ptr)) {
      attr_ptr->owner = *ObjPtr;
      attr_ptr->prev = NULL;
      attr_ptr->next = top_attr;
      if (top_attr == NULL) {
         bot_attr = attr_ptr;
      } else {
         top_attr->prev = attr_ptr;
      }
      top_attr = attr_ptr;
   }
   if (bot_attr != NULL) bot_attr->next = NULL;
   if (*ObjPtr == NULL) {
      DelAllAttrs(top_attr);
   } else {
      (*ObjPtr)->fattr = top_attr;
      (*ObjPtr)->lattr = bot_attr;
   }
}

static
int ReadPageObj(Inbuf, ppsz_page_name)
   char *Inbuf, **ppsz_page_name;
{
   int page_num=0;
   char *s=NULL, *c_ptr=NULL;

   if (ppsz_page_name != NULL) *ppsz_page_name = NULL;
   s = FindChar((int)'(', Inbuf);
   if (sscanf(s, "%d", &page_num) != 1) {
      (void)sprintf(gszMsgBox, "%s, %d:  Missing %s in %s.",
            scanFileName, scanLineNum, "page_num", "page");
      if (PRTGIF) {
         fprintf(stderr, "%s\n", gszMsgBox);
      } else {
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
      return FALSE;
   }
   if (!importingFile) {
      char saved_ch='\0';

      for (curPage=firstPage; page_num != 1 && curPage != NULL;
            curPage=curPage->next, page_num--) {
      }
      if (curPage == NULL) {
         sprintf(gszMsgBox, "Malformed input file (%s %1d).  Abort!",
               "apparently reading page", page_num);
         if (PRTGIF) {
            fprintf(stderr, "%s\n", gszMsgBox);
         } else {
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
         return FALSE;
      }
      s = FindChar((int)',', s);
      c_ptr = FindChar((int)'"', s);
      s = ReadString(c_ptr);
      saved_ch = *(--s);
      *s = '\0';
      if (*c_ptr != '\0') {
         curPage->name = UtilStrDup(c_ptr);
         if (curPage->name == NULL) FailAllocMessage();
      }
      *s++ = saved_ch;
      InitScan(s, "\t\n, )");

      topObj = curPage->top;
      botObj = curPage->bot;

      curPage->layer_on = TRUE;
      if (fileVersion <= 32) {
      } else {
         int layer_on=TRUE;
         char page_file_name[MAXSTRING];

         if (GETINT("page", layer_on, "color layer on") == INVALID) {
            return FALSE;
         }
         curPage->layer_on = layer_on;
         /*
          * This is a hack!  If the file version becomes >= 38, this code
          *       must be removed!
          */
         if (GETSTR("page", page_file_name, "page_file_name") == INVALID) {
            /* should not get here because we should have read in: "." */
            return FALSE;
         } else if (strcmp(page_file_name, ".") == 0) {
            *page_file_name = '\0';
         } else if (*page_file_name == '\'') {
            UtilRemoveQuotes(page_file_name);
         }
         if (*page_file_name != '\0') {
            char *psz=strchr(page_file_name, ' ');

            if (psz != NULL) {
               char truncated_name[MAXSTRING];

               *psz = *truncated_name = '\0';
               UtilStrCpyN(truncated_name, sizeof(truncated_name),
                     page_file_name);
               *psz = ' ';
               sprintf(gszMsgBox, "%s: '%s'.  %s '%s'.",
                     "Invalid page file name", page_file_name,
                     "Truncated to", truncated_name);
               if (PRTGIF) {
                  fprintf(stderr, "%s\n", gszMsgBox);
               } else {
                  MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
               }
               strcpy(page_file_name, truncated_name);
            }
            curPage->page_file_name = UtilStrDup(page_file_name);
            if (curPage->page_file_name == NULL) FailAllocMessage();
         }
      }
   } else {
      s = FindChar((int)',', s);
      c_ptr = FindChar((int)'"', s);
      s = ReadString(c_ptr);
      *(--s) = '\0';
      if (*c_ptr != '\0' && ppsz_page_name != NULL) {
         *ppsz_page_name = UtilStrDup(c_ptr);
         if (*ppsz_page_name == NULL) FailAllocMessage();
      }
   }
   readingPageNum++;
   return TRUE;
}

static
int ReadUnitObj(Inbuf)
   char *Inbuf;
{
   if (!importingFile) {
      char *s=FindChar((int)'(', Inbuf), *c_ptr;

      c_ptr = FindChar((int)'"', s);
      s = ReadString(c_ptr);
      *(--s) = '\0';
      if (SetUnit(c_ptr)) {
         return TRUE;
      } else {
         return FALSE;
      }
   }
   return TRUE;
}

static
int ReadScriptFracObj(Inbuf)
   char *Inbuf;
{
   if (!importingFile) {
      char *s=FindChar((int)'(', Inbuf), *c_ptr;

      c_ptr = FindChar((int)'"', s);
      s = ReadString(c_ptr);
      *(--s) = '\0';
      if (SetScriptFractionValue(c_ptr)) {
         return TRUE;
      } else {
         return FALSE;
      }
   }
   return TRUE;
}

#define GETSSVALUE(val,name) ScanValue("%d", &(val), name, "shapeshadow")

static
int ReadShapeShadowObj(Inbuf)
   char *Inbuf;
{
   if (!importingFile) {
      char *s=FindChar((int)'(', Inbuf);
      int dx=0, dy=0;

      InitScan(s, "\t\n, ");

      if (GETSSVALUE(dx, "dx") == INVALID ||
          GETSSVALUE(dy, "dy") == INVALID) {
         return FALSE;
      }
      shapeShadowDx = dx;
      shapeShadowDy = dy;
   }
   return TRUE;
}

static
int ReadFgBgColors(Inbuf)
   char *Inbuf;
{
   if (!importingFile && !PRTGIF) {
      int new_alloc=FALSE, bg_color_index=INVALID;
      char *s=FindChar((int)'(', Inbuf), fg_color_str[40], bg_color_str[40];

      s = ParseStr(s, (int)',', fg_color_str, sizeof(fg_color_str));
      s = ParseStr(s, (int)')', bg_color_str, sizeof(bg_color_str));
      UtilTrimBlanks(fg_color_str);
      UtilTrimBlanks(bg_color_str);
      allocColorFailed = FALSE;
      colorIndex = QuickFindColorIndex(NULL, fg_color_str, &new_alloc,
            TRUE);
      if (gnCannotFindColorMsg && !allocColorFailed) {
         fprintf(stderr, "    (%s '%s'.)\n",
               "Current color is corrected to be", colorMenuItems[colorIndex]);
      }
      bg_color_index = QuickFindColorIndex(NULL, bg_color_str, &new_alloc,
            FALSE);
      if (bg_color_index == INVALID) {
         fprintf(stderr, "%s: '%s'.  (%s '%s' instead.)\n",
               "Fail to allocate the background color", bg_color_str,
               "Use the current default background color", defaultBgColorStr);
      } else {
         strcpy(defaultBgColorStr, bg_color_str);
         defaultBgColorIndex = bg_color_index;
      }
   }
   return TRUE;
}

static
int FreeBufAndReturn(buf, rc)
   char *buf;
   int rc;
{
   if (buf != NULL) free(buf);
   return rc;
}

int ReadObj(FP, ObjPtr)
   FILE *FP;
   struct ObjRec **ObjPtr;
{
   char *line, obj_name[80];
   int read_state_ok;

   *ObjPtr = NULL;
   while ((line=UtilGetALine(FP)) != NULL) {
      scanLineNum++;
      if (*line == ']') return FreeBufAndReturn(line, FALSE);

      if (*line == '%') {
         if (!importingFile && line[1]=='%') {
            int line_len=strlen(line);

            if (savedComments == NULL) {
               if ((savedComments=(char*)malloc((line_len+2)*sizeof(char)))
                     == NULL) {
                  FailAllocMessage();
               }
               *savedComments = '\0';
            } else {
               if ((savedComments=(char*)realloc(savedComments,
                     savedCommentsLen+line_len+2)) == NULL) {
                  FailAllocMessage();
               }
               savedComments[savedCommentsLen] = '\0';
            }
            strcat(savedComments, line);
            savedCommentsLen += line_len;
            savedComments[savedCommentsLen++] = '\n';
            savedComments[savedCommentsLen] = '\0';
         }
         if (strncmp(line, "%TGWB end - ", strlen("%TGWB end - ")) == 0) {
            return FreeBufAndReturn(line, FALSE);
         }
         free(line);
         continue;
      }

      if (ParseStr(line, (int)'(', obj_name, sizeof(obj_name)) == NULL) {
      } else if (strcmp(obj_name, "poly") == 0) {
         ReadPolyObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(INVALID, FP, ObjPtr);
         if (RetractedArrowAttr(*ObjPtr) ||
               AutoRetractedArrowAttr(*ObjPtr, TRUE)) {
            /* fake the undoingOrRedoing so that no */
            /*          actual auto-adjusting is done */
            undoingOrRedoing = TRUE;
            AdjObjSplineVs(*ObjPtr);
            undoingOrRedoing = FALSE;
         }
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "box") == 0) {
         ReadBoxObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "oval") == 0) {
         ReadOvalObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "text") == 0) {
         ReadTextObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "polygon") == 0) {
         ReadPolygonObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "arc") == 0) {
         ReadArcObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "rcbox") == 0) {
         ReadRCBoxObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "xbm") == 0) {
         ReadXBmObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "xpm") == 0) {
         ReadXPmObj(FP, line, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "group") == 0) {
         ReadGroupObj(FP, OBJ_GROUP, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(INVALID, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "sym") == 0) {
         ReadGroupObj(FP, OBJ_SYM, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(INVALID, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "icon") == 0) {
         ReadGroupObj(FP, OBJ_ICON, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(INVALID, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "pin") == 0) {
         ReadGroupObj(FP, OBJ_PIN, ObjPtr);
         if (*ObjPtr == NULL) return FreeBufAndReturn(line, FALSE);
         if (deserializingFile) ReadCreatorID(FP, ObjPtr);
         ReadObjAttrs(INVALID, FP, ObjPtr);
         AdjObjBBox(*ObjPtr);
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "page") == 0) {
         int ok;
         char *page_name=NULL;

         *ObjPtr = NULL;
         if (importingFile && !importingIconFile) {
            if (importingPageNum == (-1) && *importingPageName == '\0') {
               if (readingPageNum == 1) {
                  /* as if read to the end of file */
                  if (!PRTGIF && !pastingFile) {
                     MsgBox("Only page 1 is imported from a multipage file.",
                           TOOL_NAME, INFO_MB);
                  }
                  return FreeBufAndReturn(line, FALSE);
               }
            } else {
               if (readingPageNum == importingPageNum) {
                  /* as if read to the end of file */
                  if (!PRTGIF && !pastingFile) {
                     sprintf(gszMsgBox, "%s %1d %s.",
                           "Only page", importingPageNum,
                           "is imported from a multipage file");
                     MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
                  }
                  return FreeBufAndReturn(line, FALSE);
               }
            }
         }
         ok = ReadPageObj(line, &page_name);
         if (importingFile && !importingIconFile &&
               importingPageNum == (-1) && *importingPageName != '\0' &&
               page_name != NULL && strcmp(page_name,importingPageName)==0) {
            *importingPageName = '\0';
            importingPageNum = readingPageNum;
         }
         if (page_name != NULL) free(page_name);
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "state") == 0) {
         if ((read_state_ok=ReadState(line)) == TRUE) {
            foundGoodStateObject = TRUE;
         }
         *ObjPtr = NULL;
         return FreeBufAndReturn(line, (read_state_ok) ? TRUE : INVALID);
      } else if (strcmp(obj_name, "file_attr") == 0) {
         if (importingFile && !importingIconFile) {
            struct AttrRec *saved_first_attr, *saved_last_attr;

            saved_first_attr = tgifObj->fattr;
            saved_last_attr = tgifObj->lattr;
            tgifObj->fattr = NULL;
            tgifObj->lattr = NULL;

            ReadObjAttrs(START_HAVING_ATTRS-1, FP, &tgifObj);
            DelAllAttrs(tgifObj->fattr);

            tgifObj->fattr = saved_first_attr;
            tgifObj->lattr = saved_last_attr;
         } else {
            ReadObjAttrs(START_HAVING_ATTRS-1, FP, &tgifObj);
         }
         return FreeBufAndReturn(line, TRUE);
      } else if (strcmp(obj_name, "unit") == 0) {
         int ok=ReadUnitObj(line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "shapeshadow") == 0) {
         int ok=ReadShapeShadowObj(line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "color_info") == 0) {
         int ok=ReadColors(FP, line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "cmd") == 0) {
         int ok=ReadCmd(FP, line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "before_image") == 0) {
         int ok=ReadBeforeImage(FP, line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "after_positions") == 0) {
         int ok=ReadAfterPositions(FP, line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "after_image") == 0) {
         int ok=ReadAfterImage(FP, line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "slideshow_info") == 0) {
         int ok=ReadSlideShowInfo(line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "script_frac") == 0) {
         int ok=ReadScriptFracObj(line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      } else if (strcmp(obj_name, "fg_bg_colors") == 0) {
         int ok=ReadFgBgColors(line);

         *ObjPtr = NULL;
         return FreeBufAndReturn(line, ok);
      }
      free(line);
   }
   return FALSE;
}

void AdjForOldVersion(obj_ptr)
   struct ObjRec *obj_ptr;
{
   if (fileVersion <= 13) {
      switch (gridSystem) {
      case ENGLISH_GRID:
         MoveObj(obj_ptr, (int)(HALF_INCH), (int)(HALF_INCH));
         break;
      case METRIC_GRID:
         MoveObj(obj_ptr, (int)(2.5*ONE_CM), (int)(2.5*ONE_CM));
         break;
      }
   }
}

int DownloadRemoteFile(file_name, ppsz_content_type, ppsz_page_spec,
      pn_is_html, return_tmp_fname, psz_final_url, cb_final_url)
   char *file_name, **ppsz_content_type, **ppsz_page_spec, *return_tmp_fname,
         *psz_final_url;
   int *pn_is_html, cb_final_url;
   /* file_name is assumed to be remote */
{
   int ok=TRUE;
   char remote_fname[MAXPATHLENGTH+1], *tmp_remote_fname=NULL;

   if (!FileIsRemote(file_name)) return FALSE;

   *remote_fname = '\0';
   if (ppsz_page_spec != NULL) *ppsz_page_spec = NULL;
   if (pn_is_html != NULL) *pn_is_html = FALSE;
   if (!FormNewFileName(curDir, file_name, NULL, remote_fname,
         ppsz_page_spec)) {
      sprintf(gszMsgBox, "Invalid remote file name '%s'.", file_name);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      ok = FALSE;
   } else {
      int remote_buf_sz=0;
      char *remote_buf=NULL;

      SaveStatusStrings();
      ok = LoadRemoteFileInMem(remote_fname, &remote_buf, ppsz_content_type,
            &remote_buf_sz, pn_is_html, navigateRefresh, psz_final_url,
            cb_final_url);
      RestoreStatusStrings();
      if (ok && remote_buf != NULL) {
         if ((tmp_remote_fname=WriteRemoteFileIntoTemp(remote_buf,
               remote_buf_sz, NULL)) == NULL) {
            ok = FALSE;
         }
      }
      if (remote_buf != NULL) FreeRemoteBuf(remote_buf);
   }
   if (tmp_remote_fname != NULL) {
      strcpy(return_tmp_fname, tmp_remote_fname);
      FreeRemoteBuf(tmp_remote_fname);
   }
   return ok;
}

int ImportGivenFile(file_name, group_and_lock)
   char *file_name;
   int group_and_lock;
   /* returns TRUE if ok */
   /* returns FALSE if file_name looks ok, only have temporary problems */
   /* returns BAD if file_name is bad */
{
   struct ObjRec *obj_ptr, *saved_top_obj, *saved_bot_obj;
   char *rest, remote_fname[MAXPATHLENGTH+1], *remote_buf=NULL;
   char tmp_filename[MAXPATHLENGTH+1], *tmp_remote_fname=NULL, *page_spec=NULL;
   char gzipped_fname[MAXPATHLENGTH+1];
   int short_name=FALSE, read_status=0, remote_buf_sz=0;
   int tmp_linenum, file_is_remote=FALSE, interrupted;
   FILE *fp=NULL;
   XEvent ev;

   *gzipped_fname = '\0';
   if (FileIsRemote(file_name)) {
      int rc=TRUE;

      if (!FormNewFileName(curDir, file_name, NULL, remote_fname, &page_spec)) {
         sprintf(gszMsgBox, "Invalid remote file name '%s'.", file_name);
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         rc = BAD;
      } else {
         int is_html=FALSE;;

         SaveStatusStrings();
         rc = LoadRemoteFileInMem(remote_fname, &remote_buf, NULL,
               &remote_buf_sz, &is_html, TRUE, NULL, 0);
         RestoreStatusStrings();
         if (rc && remote_buf != NULL) {
            if ((tmp_remote_fname=WriteRemoteFileIntoTemp(remote_buf,
                  remote_buf_sz, NULL)) != NULL) {
               file_is_remote = TRUE;
            } else {
               rc = FALSE;
            }
         }
      }
      if (!file_is_remote) {
         if (remote_buf != NULL) FreeRemoteBuf(remote_buf);
         return rc;
      }
   } else {
      int gzipped=FALSE;

      if (FileNameHasExtension(file_name, OBJ_FILE_TYPE, &gzipped, NULL) &&
            gzipped) {
         char *tmp_fname=NULL;

         if ((tmp_fname=GunzipFileIntoTemp(file_name)) == NULL) {
            return BAD;
         } else {
            strcpy(gzipped_fname, file_name);
            strcpy(file_name, tmp_fname);
            free(tmp_fname);
         }
      }
   }
   XSync(mainDisplay, False);
   if (XCheckMaskEvent(mainDisplay, ExposureMask, &ev)) {
      ExposeEventHandler(&ev, TRUE);
   }
   if (file_is_remote) {
      if ((fp=fopen(tmp_remote_fname, "r")) == NULL) {
         sprintf(gszMsgBox, "Cannot read tmp file '%s'.", tmp_remote_fname);
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         unlink(tmp_remote_fname);
         FreeRemoteBuf(remote_buf);
         FreeRemoteBuf(tmp_remote_fname);
         /* temporary problem */
         return FALSE;
      }
   } else {
      if (*gzipped_fname == '\0') {
         if ((short_name=IsPrefix(bootDir, file_name, &rest))) ++rest;
      } else {
         if ((short_name=IsPrefix(bootDir, gzipped_fname, &rest))) ++rest;
      }
      if ((fp=fopen(file_name, "r")) == NULL) {
         if (short_name) {
            sprintf(gszMsgBox, "Cannot import '%s'.", rest);
         } else {
            if (*gzipped_fname == '\0') {
               sprintf(gszMsgBox, "Cannot import '%s'.", file_name);
            } else {
               sprintf(gszMsgBox, "Cannot import '%s'.", gzipped_fname);
            }
         }
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         if (*gzipped_fname != '\0') unlink(file_name);
         /* temporary problem -- may be the file does not exist yet */
         return FALSE;
      }
   }
   strcpy(tmp_filename, scanFileName);
   tmp_linenum = scanLineNum;
   if (file_is_remote) {
      strcpy(scanFileName, tmp_remote_fname);
   } else {
      strcpy(scanFileName, (short_name ? rest : file_name));
   }
   scanLineNum = 0;

   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   curPage->top = curPage->bot = topObj = botObj = NULL;

   if (file_is_remote) {
      sprintf(gszMsgBox, "Importing '%s' ...", remote_fname);
   } else {
      if (short_name) {
         sprintf(gszMsgBox, "Importing '%s' ...", rest);
      } else {
         if (*gzipped_fname == '\0') {
            sprintf(gszMsgBox, "Importing '%s' ...", file_name);
         } else {
            sprintf(gszMsgBox, "Importing '%s' ...", gzipped_fname);
         }
      }
   }
   Msg(gszMsgBox);

   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   numRedrawBBox = 0;
   readingPageNum = 0;
   ShowInterrupt(1);
   interrupted = FALSE;
   foundGoodStateObject = FALSE;
   importingPageNum = (-1);
   *importingPageName = '\0';
   if (page_spec != NULL) {
      if (*page_spec == '#') {
         importingPageNum = atoi(&page_spec[1]);
         if (importingPageNum < 1) {
            importingPageNum = (-1);
            sprintf(gszMsgBox, "Invalid page number '%s' specified.",
                  page_spec);
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
      } else {
         strcpy(importingPageName, page_spec);
      }
   }
   if (importingPageNum == (-1) && *importingPageName == '\0') {
      while (!interrupted && (read_status=ReadObj(fp, &obj_ptr)) == TRUE) {
         if (obj_ptr != NULL) {
            obj_ptr->tmp_parent = NULL;
            AdjForOldVersion(obj_ptr);
            UnlockAnObj(obj_ptr);
            AddObj(NULL, topObj, obj_ptr);
            if (!interrupted &&
                  (PointInBBox(obj_ptr->x, obj_ptr->y, drawWinBBox) ||
                  BBoxIntersect(obj_ptr->bbox, drawWinBBox))) {
               if (!DrawObj(drawWindow, obj_ptr)) interrupted = TRUE;
               if (CheckInterrupt()) interrupted = TRUE;
            }
         }
      }
   } else if (importingPageNum == (-1)) {
      while (!interrupted && (read_status=ReadObj(fp, &obj_ptr)) == TRUE) {
         if (obj_ptr != NULL) {
            obj_ptr->tmp_parent = NULL;
            if (importingPageNum == (-1)) {
               FreeObj(obj_ptr);
               if (!interrupted && CheckInterrupt()) interrupted = TRUE;
            } else {
               AdjForOldVersion(obj_ptr);
               UnlockAnObj(obj_ptr);
               AddObj(NULL, topObj, obj_ptr);
               if (!interrupted &&
                     (PointInBBox(obj_ptr->x, obj_ptr->y, drawWinBBox) ||
                     BBoxIntersect(obj_ptr->bbox, drawWinBBox))) {
                  if (!DrawObj(drawWindow, obj_ptr)) interrupted = TRUE;
                  if (CheckInterrupt()) interrupted = TRUE;
               }
            }
         }
      }
   } else {
      while (!interrupted && (read_status=ReadObj(fp, &obj_ptr)) == TRUE) {
         if (obj_ptr != NULL) {
            obj_ptr->tmp_parent = NULL;
            if (importingPageNum != readingPageNum) {
               FreeObj(obj_ptr);
               if (!interrupted && CheckInterrupt()) interrupted = TRUE;
            } else {
               AdjForOldVersion(obj_ptr);
               UnlockAnObj(obj_ptr);
               AddObj(NULL, topObj, obj_ptr);
               if (!interrupted &&
                     (PointInBBox(obj_ptr->x, obj_ptr->y, drawWinBBox) ||
                     BBoxIntersect(obj_ptr->bbox, drawWinBBox))) {
                  if (!DrawObj(drawWindow, obj_ptr)) interrupted = TRUE;
                  if (CheckInterrupt()) interrupted = TRUE;
               }
            }
         }
      }
   }
   if (!PRTGIF && colorLayers && needToRedrawColorWindow) {
      RedrawColorWindow();
   }
   if (interrupted) {
      MsgBox("User interrupt.\n\nDrawing aborted.", TOOL_NAME, INFO_MB);
   }
   HideInterrupt();

   if (fp != NULL) fclose(fp);
   if (*gzipped_fname != '\0') {
      unlink(file_name);
      strcpy(file_name, gzipped_fname);
   }
   strcpy(scanFileName, tmp_filename);
   scanLineNum = tmp_linenum;

   if (read_status == INVALID) {
      if (fileVersion > CUR_VERSION) {
         sprintf(gszMsgBox, "%s (=%1d) %s.\n\n%s!\n\n%s %s.  %s <URL:%s>.",
               "File version", fileVersion, "too large",
               "Import aborted", "You may need a more recent version of",
               TOOL_NAME, "Please check", homePageURL);
      } else {
         sprintf(gszMsgBox, "%s.\n\n%s!", "File corrupted", "Import aborted");
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      SetDefaultCursor(mainWindow);
      ShowCursor();
      if (file_is_remote) {
         unlink(tmp_remote_fname);
         FreeRemoteBuf(remote_buf);
         FreeRemoteBuf(tmp_remote_fname);
      }
      /* temporary problem -- may be the file will be fixed later */
      return FALSE;
   }
   if (file_is_remote) {
      if (!foundGoodStateObject) PasteString(remote_buf);

      unlink(tmp_remote_fname);
      FreeRemoteBuf(remote_buf);
      FreeRemoteBuf(tmp_remote_fname);
   }

   if (topObj != NULL) SetFileModified(TRUE);
   justDupped = FALSE;

   if (group_and_lock && topObj != NULL) {
      if (topObj != botObj || topObj->type==OBJ_POLY ||
            topObj->type==OBJ_POLYGON) {
         struct ObjRec *top_obj=topObj, *bot_obj=botObj;

         SelAllObj(FALSE);
         curPage->top = curPage->bot = topObj = botObj = NULL;
         CreateGroupObj(top_obj, bot_obj);
         RemoveAllSel();
      }
      topObj->locked = TRUE;
   }
   SelAllObj(FALSE);

   if (botObj != NULL) {
      botObj->next = saved_top_obj;
   } else {
      topObj = saved_top_obj;
   }
   if (saved_top_obj != NULL) {
      saved_top_obj->prev = botObj;
      botObj = saved_bot_obj;
   }
   curPage->top = topObj;
   curPage->bot = botObj;

   if (!(file_is_remote && !foundGoodStateObject) && topSel != NULL) {
      PrepareToRecord(CMD_NEW, NULL, NULL, 0);
      RecordCmd(CMD_NEW, NULL, topSel, botSel, numObjSelected);
   }
   HighLightForward();

   if (!importFromLibrary && !file_is_remote) {
      SetCurImportDir(file_name);
   }
   if (file_is_remote) {
      sprintf(gszMsgBox, "'%s' imported.", remote_fname);
   } else {
      if (short_name) {
         sprintf(gszMsgBox, "'%s' imported.", rest);
      } else {
         sprintf(gszMsgBox, "'%s' imported.", file_name);
      }
   }
   Msg(gszMsgBox);
   SetDefaultCursor(mainWindow);
   ShowCursor();
   if (page_spec != NULL) free(page_spec);
   return TRUE;
}

void ImportFile()
{
   char file_name[MAXPATHLENGTH+1];

   MakeQuiescent();

   importingFile = TRUE;
   if (importFromLibrary) {
      char name[MAXSTRING+1], path[MAXSTRING+1];

      if (SelectFromLibrary("Please select an object file to IMPORT...",
            OBJ_FILE_EXT, name, path) == INVALID) {
         importingFile = FALSE;
         return;
      }
      sprintf(file_name, "%s%c%s", path, DIR_SEP, name);
   } else if (SelectFileNameToImport(
         "Please select an object file to IMPORT...",
         OBJ_FILE_EXT, file_name) == INVALID) {
      importingFile = FALSE;
      return;
   }
   ImportGivenFile(file_name, FALSE);
   importingFile = FALSE;
}

int ClearBgColorInfo(nRedraw)
   int nRedraw;
{
   myFileBgPixel = myFileFgPixel = INVALID;
   if (myFileFgColorStr != NULL) {
      free(myFileFgColorStr);
      myFileFgColorStr = NULL;
   }
   if (myFileBgColorStr != NULL) {
      free(myFileBgColorStr);
      myFileBgColorStr = NULL;
      XSetWindowBackground(mainDisplay, drawWindow, myBgPixel);
      if (nRedraw) {
         ClearAndRedrawDrawWindow();
      }
      RecalcXorPixels();
      return TRUE;
   }
   return FALSE;
}

int ClearBgPixmapInfo(nRedraw)
   int nRedraw;
{
   int changed=FALSE;

   if (myFileBgGifURL != NULL) {
      free(myFileBgGifURL);
      myFileBgGifURL = NULL;
      changed = TRUE;
   }
   myFileBgPixmapW = myFileBgPixmapH = (-1);
   if (myFileBgPixmap != None) {
      XFreePixmap(mainDisplay, myFileBgPixmap);
      myFileBgPixmap = None;
      XSetWindowBackgroundPixmap(mainDisplay, drawWindow, None);
      changed = TRUE;
   }
   if (changed && nRedraw) {
      ClearAndRedrawDrawWindow();
   }
   return changed;
}

int SetFileBgColor()
   /* assumes myFileBgColorStr is set */
{
   int new_alloc=FALSE;
   int index=QuickFindColorIndex(NULL, myFileBgColorStr, &new_alloc, TRUE);

   if (index != INVALID) {
      myFileBgPixel = colorPixels[index];
      myFileBgColor.pixel = colorPixels[index];
      myFileBgColor.red = tgifColors[index].red;
      myFileBgColor.green = tgifColors[index].green;
      myFileBgColor.blue = tgifColors[index].blue;
      XSetWindowBackground(mainDisplay, drawWindow, colorPixels[index]);
      RecalcXorPixels();
      return TRUE;
   }
   sprintf(gszMsgBox,
         "Cannot allocate bg_color '%s', default bgcolor used.",
         myFileBgColorStr);
   Msg(gszMsgBox);
   return FALSE;
}

int SetFileFgColor()
   /* assumes myFileFgColorStr is set */
{
   int new_alloc=FALSE;
   int index=QuickFindColorIndex(NULL, myFileFgColorStr, &new_alloc, TRUE);

   if (index != INVALID) {
      myFileFgPixel = colorPixels[index];
      return TRUE;
   }
   sprintf(gszMsgBox,
         "Cannot allocate fg_color '%s', default fgcolor used.",
         myFileFgColorStr);
   Msg(gszMsgBox);
   return FALSE;
}

int SetFileBgPixmap()
{
   int ok=TRUE;
   char tmp_xpm_fname[MAXPATHLENGTH+1];

   if (myFileBgGifURL == NULL) return FALSE;

   *tmp_xpm_fname = '\0';
   if (FileIsRemote(myFileBgGifURL)) {
      char tmp_gif_fname[MAXPATHLENGTH+1], *content_type=NULL, *page_spec=NULL;
      int is_html=FALSE;

      *tmp_gif_fname = '\0';
      if (DownloadRemoteFile(myFileBgGifURL, &content_type, &page_spec,
            &is_html, tmp_gif_fname, NULL, 0)) {
         if (page_spec != NULL || is_html) {
            sprintf(gszMsgBox, "Invalid remote file name '%s'.",
                  myFileBgGifURL);
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            ok = FALSE;
         } else {
            SetWatchCursor(drawWindow);
            SetWatchCursor(mainWindow);
            SaveStatusStrings();
            ok = ConvertGifToXpm(tmp_gif_fname, tmp_xpm_fname);
            RestoreStatusStrings();
            SetDefaultCursor(mainWindow);
            ShowCursor();
         }
      } else {
         ok = FALSE;
      }
      if (content_type != NULL) FreeRemoteBuf(content_type);
      if (page_spec != NULL) UtilFree(page_spec);
      if (*tmp_gif_fname != '\0') unlink(tmp_gif_fname);
   } else {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
      SaveStatusStrings();
      ok = ConvertGifToXpm(myFileBgGifURL, tmp_xpm_fname);
      RestoreStatusStrings();
      SetDefaultCursor(mainWindow);
      ShowCursor();
   }
   if (ok) {
      int image_w, image_h, w, h, ncolors=0;
      int chars_per_pixel=0, first_pixel_is_bg=FALSE, *pixels=NULL;
      char *color_char=NULL, **color_str=NULL, *xpm_data=NULL;
      Pixmap pixmap=None, bitmap=None;
      XImage *image=NULL, *bitmap_image=NULL;

      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
      if (!MyReadPixmapFile(tmp_xpm_fname, &image_w, &image_h, &w, &h, &pixmap,
            &image, &bitmap, &bitmap_image, &ncolors, &chars_per_pixel,
            &first_pixel_is_bg, &color_char, &color_str, &pixels,
            &xpm_data) == BitmapSuccess) {
         ok = FALSE;
      }
      SetDefaultCursor(mainWindow);
      ShowCursor();
      unlink(tmp_xpm_fname);
      if (bitmap != None) XFreePixmap(mainDisplay, bitmap);
      if (image != NULL) XDestroyImage(image);
      if (bitmap_image != NULL) XDestroyImage(bitmap_image);
      if (color_char != NULL) free(color_char);
      if (color_str != NULL) {
         int i;

         for (i=0; i < ncolors; i++) free(color_str[i]);
         free(color_str);
      }
      if (ok) {
         myFileBgPixmapW = image_w;
         myFileBgPixmapH = image_h;
         myFileBgPixmap = pixmap;
         XSetWindowBackgroundPixmap(mainDisplay, drawWindow, myFileBgPixmap);
      }
   }
   if (!ok) {
      sprintf(gszMsgBox, "Cannot set background pixmap to '%s'.",
            myFileBgGifURL);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   }
   return ok;
}

void CheckFileAttrsInLoad()
{
   struct AttrRec *attr_ptr;
   int changed=FALSE;

   if ((attr_ptr=FindFileAttrWithName("bg_color=")) != NULL) {
      struct AttrRec *fg_attr_ptr;
      int ok=TRUE;

      if ((fg_attr_ptr=FindFileAttrWithName("fg_color=")) != NULL) {
         if (myFileFgColorStr == NULL ||
               strcmp(fg_attr_ptr->attr_value.s, myFileFgColorStr) != 0) {
            myFileFgColorStr = UtilStrDup(fg_attr_ptr->attr_value.s);
            if (myFileFgColorStr == NULL) FailAllocMessage();
            if (!SetFileFgColor()) {
               ClearBgColorInfo(TRUE);
               ok = FALSE;
            }
         }
      }
      if (ok && (myFileBgColorStr == NULL ||
            strcmp(attr_ptr->attr_value.s, myFileBgColorStr) != 0)) {
         myFileBgColorStr = UtilStrDup(attr_ptr->attr_value.s);
         if (myFileBgColorStr == NULL) FailAllocMessage();
         if (SetFileBgColor()) {
            ClearAndRedrawDrawWindow();
         } else {
            ClearBgColorInfo(TRUE);
         }
      }
      return;
   } else {
      changed = ClearBgColorInfo(FALSE);
   }
   /*
    * Can't really get this to work because of scrolling and zooming.
    */
/*
   if ((attr_ptr=FindFileAttrWithName("bg_gif=")) != NULL) {
      if (myFileBgGifURL == NULL ||
            strcmp(attr_ptr->attr_value.s, myFileBgGifURL) != 0) {
         if (myFileBgPixmap != None) XFreePixmap(mainDisplay, myFileBgPixmap);
         myFileBgPixmap = None;
         myFileBgPixmapW = myFileBgPixmapH = (-1);
         if (myFileBgGifURL != NULL) free(myFileBgGifURL);
         myFileBgGifURL = UtilStrDup(attr_ptr->attr_value.s);
         if (myFileBgGifURL == NULL) FailAllocMessage();
         if (SetFileBgPixmap()) {
            ClearAndRedrawDrawWindow();
         } else {
            ClearBgPixmapInfo(TRUE);
         }
      }
      return;
   } else {
      if (!changed) {
         changed = ClearBgPixmapInfo(FALSE);
      } else {
         ClearBgPixmapInfo(FALSE);
      }
   }
 */
   if (changed) {
      if (myFileBgPixmap == None) {
         XSetWindowBackground(mainDisplay, drawWindow, myBgPixel);
      }
      ClearAndRedrawDrawWindow();
   }
}

int LoadFile(FullName, ObjFile, GzippedObjFile)
   char *FullName;
   int ObjFile; /* equals TRUE if the file is an OBJ file */
                /* equals FALSE if the file is a SYM or PIN file */
                /* equals -1 if the file is an temporary OBJ file */
{
   struct ObjRec *obj_ptr=NULL;
   char file_name[MAXPATHLENGTH+1], tmp_filename[MAXPATHLENGTH+1];
   char saved_cur_dir[MAXPATHLENGTH+1], *rest=NULL;
   char gzipped_fname[MAXPATHLENGTH+1];
   int read_status, short_name, tmp_linenum, interrupted;
   FILE *fp=NULL;
   XEvent ev;

   *gzipped_fname = '\0';
   if (ObjFile != FALSE) {
      strcpy(saved_cur_dir, curDir);
   } else {
      strcpy(saved_cur_dir, curSymDir);
   }
   if (GzippedObjFile) {
      char *tmp_fname=NULL;

      if ((tmp_fname=GunzipFileIntoTemp(FullName)) == NULL) {
         return FALSE;
      } else {
         strcpy(gzipped_fname, FullName);
         strcpy(file_name, tmp_fname);
         free(tmp_fname);
      }
   } else {
      strcpy(file_name, FullName);
   }
   if (*gzipped_fname == '\0') {
      if ((short_name=IsPrefix(bootDir, file_name, &rest))) ++rest;
   } else {
      if ((short_name=IsPrefix(bootDir, gzipped_fname, &rest))) ++rest;
   }
   if ((fp=fopen(file_name, "r")) == NULL) {
      if (short_name) {
         sprintf(gszMsgBox, "Cannot open '%s'.", rest);
      } else {
         if (*gzipped_fname == '\0') {
            sprintf(gszMsgBox, "Cannot open '%s'.", file_name);
         } else {
            sprintf(gszMsgBox, "Cannot open '%s'.", gzipped_fname);
         }
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      if (*gzipped_fname != '\0') unlink(file_name);
      return FALSE;
   }
   if (!navigatingBackAndForth) BeforeNavigate();

   ClearFileInfo(TRUE);
   if (usePaperSizeStoredInFile) ResetOnePageSize();

   strcpy(tmp_filename, scanFileName);
   tmp_linenum = scanLineNum;
   strcpy(scanFileName, (short_name ? rest : file_name));
   scanLineNum = 0;

   TieLooseEnds();
   CleanUpDrawingWindow();
   SetFileModified(FALSE);

   XSync(mainDisplay, False);
   while (XCheckWindowEvent(mainDisplay, drawWindow, ExposureMask, &ev)) ;

   SaveStatusStrings();

   if (short_name) {
      sprintf(gszMsgBox, "Loading '%s' ...", rest);
   } else {
      if (*gzipped_fname == '\0') {
         sprintf(gszMsgBox, "Loading '%s' ...", file_name);
      } else {
         sprintf(gszMsgBox, "Loading '%s' ...", gzipped_fname);
      }
   }
   SetStringStatus(gszMsgBox);
   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   if (flushColormapOnOpen) {
      if (FlushColormap()) {
         Msg("Colormap flushed.");
      }
   }
   XClearWindow(mainDisplay, drawWindow);
   somethingHighLighted = FALSE;

   numRedrawBBox = 0;
   ShowInterrupt(1);
   interrupted = FALSE;
   readingPageNum = 0;
   loadedCurPageNum = 0;
   foundGoodStateObject = FALSE;
   while ((read_status=ReadObj(fp, &obj_ptr)) == TRUE) {
      if (obj_ptr != NULL) {
         obj_ptr->tmp_parent = NULL;
         AdjForOldVersion(obj_ptr);
         AddObj(NULL, topObj, obj_ptr);
         if (!interrupted && readingPageNum == loadedCurPageNum &&
               (PointInBBox(obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect(obj_ptr->bbox, drawWinBBox))) {
            if (!DrawObj(drawWindow, obj_ptr)) interrupted = TRUE;
            if (CheckInterrupt()) interrupted = TRUE;
         }
      }
   }
   fclose(fp);
   if (*gzipped_fname != '\0') unlink(file_name);

   if (!PRTGIF && colorLayers && needToRedrawColorWindow) {
      RedrawColorWindow();
   }
   if (interrupted) {
      MsgBox("User interrupt.\n\nDrawing aborted.", TOOL_NAME, INFO_MB);
   }
   HideInterrupt();

   strcpy(scanFileName, tmp_filename);
   scanLineNum = tmp_linenum;

   if (read_status == INVALID) {
      if (fileVersion > CUR_VERSION) {
         sprintf(gszMsgBox, "%s (=%1d) %s.\n\n%s!\n\n%s %s.  %s <URL:%s>.",
               "File version", fileVersion, "too large",
               "Open aborted", "You may need a more recent version of",
               TOOL_NAME, "Please check", homePageURL);
      } else {
         sprintf(gszMsgBox, "%s.\n\n%s!", "File corrupted", "Open aborted");
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      DrawPaperBoundary(drawWindow);
      RedrawGridLines(drawWindow);
      SetDefaultCursor(mainWindow);
      ShowCursor();

      RestoreStatusStrings();

      return FALSE;
   }
   if (ObjFile == TRUE) {
      if (*gzipped_fname == '\0') {
         SetCurDir(file_name);
      } else {
         SetCurDir(gzipped_fname);
      }
      *curSymDir = '\0';
   } else if (ObjFile == FALSE) {
      SetCurSymDir(file_name);
   }
   curFileDefined = TRUE;

   CheckFileAttrsInLoad();

   if (loadedCurPageNum <= 0 || curPage == NULL) {
      DelAllPages();
      loadedCurPageNum = curPageNum = lastPageNum = 1;
      InitPage();
      DrawPaperBoundary(drawWindow);
      RedrawGridLines(drawWindow);
      RedrawPageLines(drawWindow);
      RedrawChoiceWindow();
   }
   GotoPageNum(loadedCurPageNum);

   if (ObjFile != FALSE) {
      if (strcmp(saved_cur_dir, curDir) != 0 && DirInSymPath(".")) {
         UpdateSymInfo();
      }
   } else {
      if (strcmp(saved_cur_dir, curSymDir) != 0 && DirInSymPath(".")) {
         UpdateSymInfo();
      }
   }

   if (short_name) {
      sprintf(gszMsgBox, "Current file is '%s'.", rest);
   } else {
      if (*gzipped_fname == '\0') {
         sprintf(gszMsgBox, "Current file is '%s'.", file_name);
      } else {
         sprintf(gszMsgBox, "Current file is '%s'.", gzipped_fname);
      }
   }
   Msg(gszMsgBox);
   RestoreStatusStrings();

   RedrawTitleWindow();
   justDupped = FALSE;
   SetDefaultCursor(mainWindow);
   ShowCursor();

   CleanUpCmds();

   UpdateAllPinnedMenus();

   if (!navigatingBackAndForth) CommitNavigate();
   return foundGoodStateObject;
}

void DumpPatFill(FP, Fill, CellSize, BBox, Indent, EndOfLine)
   FILE *FP;
   int Fill, CellSize, Indent, EndOfLine;
   struct BBRec BBox;
{
   int ltx, lty, rbx, rby;
   struct BBRec bbox;

   bbox.ltx = BBox.ltx-1; bbox.lty = BBox.lty-1;
   bbox.rbx = BBox.rbx+1; bbox.rby = BBox.rby+1;

   ltx = ((bbox.ltx % CellSize) == 0) ? bbox.ltx :
         ((bbox.ltx > 0) ? ((int)(bbox.ltx / CellSize))*CellSize :
         ((int)(bbox.ltx / CellSize)-1)*CellSize);
   lty = ((bbox.lty % CellSize) == 0) ? bbox.lty :
         ((bbox.lty > 0) ? ((int)(bbox.lty / CellSize))*CellSize :
         ((int)(bbox.lty / CellSize)-1)*CellSize);
   rbx = ((bbox.rbx % CellSize) == 0) ? bbox.rbx :
         ((bbox.rbx > 0) ? ((int)(bbox.rbx / CellSize)+1)*CellSize :
         ((int)(bbox.rbx / CellSize))*CellSize);
   rby = ((bbox.rby % CellSize) == 0) ? bbox.rby :
         ((bbox.rby > 0) ? ((int)(bbox.rby / CellSize)+1)*CellSize :
         ((int)(bbox.rby / CellSize))*CellSize);

   if (fprintf(FP, "%spat%1d %1d %1d %1d %1d %1d TGPF%s",
         GetIndentString(Indent), Fill, CellSize, ltx, lty, rbx-ltx, rby-lty,
         (EndOfLine?"\n":"")) == EOF) {
      writeFileFailed = TRUE;
   }
}

void DumpSymOutline(FP, ObjPtr)
   FILE *FP;
   register struct ObjRec *ObjPtr;
{
   int ltx, lty, rbx, rby;

   ltx = ObjPtr->obbox.ltx - QUARTER_INCH + 1;
   lty = ObjPtr->obbox.lty - QUARTER_INCH + 1;
   rbx = ObjPtr->obbox.rbx + QUARTER_INCH - 1;
   rby = ObjPtr->obbox.rby + QUARTER_INCH - 1;

   fprintf(FP, "%s\n", gPsCmd[PS_GSAVE]);
   fprintf(FP, "   0 %s\n", gPsCmd[PS_SETGRAY]);
   fprintf(FP, "   [4 4] 0 %s\n", gPsCmd[PS_SETDASH]);
   fprintf(FP, "   %s\n   %1d %1d %s ",
         gPsCmd[PS_NEWPATH], ltx, lty, gPsCmd[PS_MOVETO]);
   fprintf(FP, "%1d %1d %s ", rbx, lty, gPsCmd[PS_LINETO]);
   fprintf(FP, "%1d %1d %s ", rbx, rby, gPsCmd[PS_LINETO]);
   fprintf(FP, "%1d %1d %s\n", ltx, rby, gPsCmd[PS_LINETO]);
   fprintf(FP, "   %s %s\n", gPsCmd[PS_CLOSEPATH], gPsCmd[PS_STROKE]);
   fprintf(FP, "%s\n", gPsCmd[PS_GRESTORE]);
}

static int	printingFirstPageNum=1;
static int	printingPageNum=1, printingPageRow=1, printingPageCol=1;
static int	printingLastPageNum=0;
static int	dumpOnePageInTileMode=FALSE;
static int	dumpOnePageInStackMode=FALSE;

static FILE	*dumpFP=NULL;
static char	tmpFile[MAXSTRING+1];
static int	llxTotal=0, llyTotal=0, urxTotal=0, uryTotal=0;
static int	totalBBoxValid=FALSE;

static int	msgAboutTiledPageScalingSeen=FALSE;

static
void DumpAttrs(FP, AttrPtr)
   FILE *FP;
   register struct AttrRec *AttrPtr;
{
   for ( ; AttrPtr != NULL; AttrPtr=AttrPtr->prev) {
      if (!AttrPtr->shown) {
         continue;
      }
      if (!AttrPtr->nameshown && *AttrPtr->attr_name.s == '!' &&
            strcmp(AttrPtr->attr_name.s, "!PAGE_NUM=") == 0) {
         if (pageLayoutMode == PAGE_STACK) {
            MiniLineInfo *pMiniLine=AttrPtr->obj->detail.t->minilines.first;
            char *c_ptr;

            if (pMiniLine != NULL &&
                  pMiniLine->first_block == pMiniLine->last_block &&
                  (strstr(pMiniLine->first_block->seg->dyn_str.s,
                  "!(STACKED_PAGE_NUM)") != NULL ||
                  strstr(pMiniLine->first_block->seg->dyn_str.s,
                  "!(STACKED_NUM_PAGES)") != NULL)) {
               char *psz_saved=NULL, *c_ptr1=NULL;
               int page_num_str_len=strlen("!(STACKED_PAGE_NUM)");
               int num_pages_str_len=strlen("!(STACKED_NUM_PAGES)");

               psz_saved = UtilStrDup(pMiniLine->first_block->seg->dyn_str.s);
               if (psz_saved == NULL) FailAllocMessage();
               while ((c_ptr=strstr(pMiniLine->first_block->seg->dyn_str.s,
                     "!(STACKED_PAGE_NUM)")) != NULL) {
                  c_ptr1 = &c_ptr[page_num_str_len];
                  sprintf(c_ptr, "%1d", printingPageNum);
                  c_ptr = &c_ptr[strlen(c_ptr)];
                  while (*c_ptr1 != '\0') *c_ptr++ = *c_ptr1++;
                  *c_ptr = '\0';
               }
               while ((c_ptr=strstr(pMiniLine->first_block->seg->dyn_str.s,
                     "!(STACKED_NUM_PAGES)")) != NULL) {
                  c_ptr1 = &c_ptr[num_pages_str_len];
                  sprintf(c_ptr, "%1d", lastPageNum);
                  c_ptr = &c_ptr[strlen(c_ptr)];
                  while (*c_ptr1 != '\0') *c_ptr++ = *c_ptr1++;
                  *c_ptr = '\0';
               }
               DumpTextObj(FP, AttrPtr->obj);
               DynStrSet(&pMiniLine->first_block->seg->dyn_str, psz_saved);
               UtilFree(psz_saved);
            } else {
               DumpTextObj(FP, AttrPtr->obj);
            }
         } else {
            /* pageLayoutMode == PAGE_TILE */
            MiniLineInfo *pMiniLine=AttrPtr->obj->detail.t->minilines.first;
            char *c_ptr=NULL;

            if (pMiniLine != NULL) {
               char *psz_saved=NULL, *c_ptr1=NULL, *c_ptr2=NULL;

               psz_saved = UtilStrDup(pMiniLine->first_block->seg->dyn_str.s);
               if (psz_saved == NULL) FailAllocMessage();
               for (c_ptr=pMiniLine->first_block->seg->dyn_str.s;
                     *c_ptr != '\0'; c_ptr++) {
                  int len=strlen("!(TILED_PAGE_ROW)");

                  if (*c_ptr == '!' &&
                        (strncmp(c_ptr, "!(TILED_PAGE_ROW)", len) == 0 ||
                        strncmp(c_ptr, "!(TILED_PAGE_COL)", len) == 0)) {
                     if (strncmp(c_ptr, "!(TILED_PAGE_ROW)", len) == 0) {
                        sprintf(c_ptr, "%1d", printingPageRow);
                     } else {
                        sprintf(c_ptr, "%1d", printingPageCol);
                     }
                     c_ptr1 = &c_ptr[len];
                     c_ptr = c_ptr2 = &c_ptr[strlen(c_ptr)];
                     while (*c_ptr1 != '\0') *c_ptr2++ = *c_ptr1++;
                     *c_ptr2 = '\0';
                     c_ptr--;
                  }
               }
               DumpTextObj(FP, AttrPtr->obj);
               DynStrSet(&pMiniLine->first_block->seg->dyn_str, psz_saved);
               UtilFree(psz_saved);
            } else {
               DumpTextObj(FP, AttrPtr->obj);
            }
         }
      } else {
         DumpTextObj(FP, AttrPtr->obj);
      }
   }
}

static
void DumpAnObj(FP, ObjPtr)
   FILE *FP;
   register struct ObjRec *ObjPtr;
{
   switch (ObjPtr->type) {
   case OBJ_POLY:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpPolyObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_BOX:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpBoxObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_OVAL:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpOvalObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_TEXT:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpTextObj(FP, ObjPtr);
      }
      break;
   case OBJ_POLYGON:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpPolygonObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_ARC:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpArcObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_RCBOX:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpRCBoxObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_XBM:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpXBmObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_XPM:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpXPmObj(FP, ObjPtr);
         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_SYM:
   case OBJ_ICON:
   case OBJ_GROUP:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         struct ObjRec *obj_ptr=ObjPtr->detail.r->last;

         for ( ; obj_ptr != NULL; obj_ptr = obj_ptr->prev) {
            obj_ptr->tmp_parent = ObjPtr;
            DumpAnObj(FP, obj_ptr);
         }
         DumpAttrs(FP, ObjPtr->lattr);
         if (ObjPtr->type == OBJ_SYM) DumpSymOutline(FP, ObjPtr);
      }
      break;
   case OBJ_PIN:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         struct ObjRec *obj_ptr=GetPinObj(ObjPtr);

         obj_ptr->tmp_parent = ObjPtr;
         DumpAnObj(FP, obj_ptr);

         DumpAttrs(FP, ObjPtr->lattr);
      }
      break;
   }
}

static int	llxPage=0, llyPage=0, urxPage=0, uryPage=0;
static int	minimalEPS=INVALID;

#define ALL_BBOX 0
#define PAGE_BBOX 1

int DumpBBox(fp, page_only, page_bbox)
   FILE *fp;
   int page_only;
   struct BBRec *page_bbox;
{
   register struct ObjRec *obj_ptr;
   int ltx=0, lty=0, rbx=0, rby=0, rc=TRUE;
   double llx1=0, lly1=0, urx1=0, ury1=0;

   if ((obj_ptr = topObj) == NULL) {
      ltx = lty = rbx = rby = 0;
      if (!(pageLayoutMode == PAGE_STACK &&
            printingFirstPageNum != printingLastPageNum)) {
         strcpy(gszMsgBox, "Warning:  The PostScript bounding box is empty!");
         if (PRTGIF) {
            fprintf(stderr, "%s\n", gszMsgBox);
         } else {
            Msg(gszMsgBox);
         }
      }
      rc = FALSE;
   } else if (pageLayoutMode == PAGE_STACK || page_bbox == NULL) {
      ltx = obj_ptr->bbox.ltx; lty = obj_ptr->bbox.lty;
      rbx = obj_ptr->bbox.rbx; rby = obj_ptr->bbox.rby;

      for (obj_ptr=topObj->next; obj_ptr != NULL; obj_ptr=obj_ptr->next) {
         if (obj_ptr->bbox.ltx < ltx) ltx = obj_ptr->bbox.ltx;
         if (obj_ptr->bbox.lty < lty) lty = obj_ptr->bbox.lty;
         if (obj_ptr->bbox.rbx > rbx) rbx = obj_ptr->bbox.rbx;
         if (obj_ptr->bbox.rby > rby) rby = obj_ptr->bbox.rby;
      }
   } else {
      int found=FALSE;

      for (obj_ptr=botObj; obj_ptr != NULL; obj_ptr=obj_ptr->prev) {
         if (Inside(obj_ptr->bbox, *page_bbox) ||
               BBoxIntersect(obj_ptr->bbox, *page_bbox)) {
            if (found) {
               if (obj_ptr->bbox.ltx < ltx) ltx = obj_ptr->bbox.ltx;
               if (obj_ptr->bbox.lty < lty) lty = obj_ptr->bbox.lty;
               if (obj_ptr->bbox.rbx > rbx) rbx = obj_ptr->bbox.rbx;
               if (obj_ptr->bbox.rby > rby) rby = obj_ptr->bbox.rby;
            } else {
               found = TRUE;
               ltx = obj_ptr->bbox.ltx; lty = obj_ptr->bbox.lty;
               rbx = obj_ptr->bbox.rbx; rby = obj_ptr->bbox.rby;
            }
         }
      }
      if (!found) {
         rc = FALSE;
         ltx = lty = rbx = rby = 0;
      } else if (page_bbox != NULL) {
         struct BBRec bbox, bbox1;

         bbox.ltx = ltx; bbox.lty = lty; bbox.rbx = rbx; bbox.rby = rby;
         if (IntersectRect(bbox, *page_bbox, &bbox1)) {
            ltx = bbox1.ltx-page_bbox->ltx; lty = bbox1.lty-page_bbox->lty;
            rbx = bbox1.rbx-page_bbox->ltx; rby = bbox1.rby-page_bbox->lty;
         }
      }
   }
   switch (pageStyle) {
   case PORTRAIT:
      llx1 = (double)(1.0*ltx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
            psXOff[pageStyle]*psDotsPerInch);
      lly1 = (double)(-1.0*rby*psDotsPerInch/PIX_PER_INCH*printMag/100 +
            psYOff[pageStyle]*psDotsPerInch);
      urx1 = (double)(1.0*rbx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
            psXOff[pageStyle]*psDotsPerInch);
      ury1 = (double)(-1.0*lty*psDotsPerInch/PIX_PER_INCH*printMag/100 +
            psYOff[pageStyle]*psDotsPerInch);
      break;
   case LANDSCAPE:
      llx1 = (double)(1.0*lty*psDotsPerInch/PIX_PER_INCH*printMag/100 -
            psYOff[pageStyle]*psDotsPerInch);
      lly1 = (double)(1.0*ltx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
            psXOff[pageStyle]*psDotsPerInch);
      urx1 = (double)(1.0*rby*psDotsPerInch/PIX_PER_INCH*printMag/100 -
            psYOff[pageStyle]*psDotsPerInch);
      ury1 = (double)(1.0*rbx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
            psXOff[pageStyle]*psDotsPerInch);
      break;
   default:
      fprintf(stderr, "Unrecognizable page style '%1d'\n", pageStyle);
      break;
   }

   if (rc) {
      llxPage = (llx1 >= 0.0) ? ((int)llx1)-1 : (-((int)(-llx1)))-1;
      llyPage = (lly1 >= 0.0) ? ((int)lly1)-1 : (-((int)(-lly1)))-1;
      urxPage = (urx1 >= 0.0) ? ((int)urx1)+1 : (-((int)(0.999-urx1)))+1;
      uryPage = (ury1 >= 0.0) ? ((int)ury1)+1 : (-((int)(0.999-ury1)))+1;
   } else {
      llxPage = llyPage = urxPage = uryPage = 0;
   }
   if (!page_only || !minimalEPS) {
      fprintf(fp,"%%%%%sBoundingBox: %1d %1d %1d %1d\n",
            (page_only ? "Page" : ""), llxPage, llyPage, urxPage, uryPage);
   }
   return rc;
}

void ModifyOutputFileName(FileName)
   char *FileName;
{
   register int i;
   int len;
   char s[MAXPATHLENGTH+1];

   if (*outputDir == '\0') return;

   strcpy(s, FileName);
   len = strlen(s);
   for (i=len-1; i >= 0 && s[i] != DIR_SEP; i--) ;
   if (i >= 0) {
      sprintf(FileName, "%s%c%s", outputDir, DIR_SEP, &s[i+1]);
   } else {
      sprintf(FileName, "%s%c%s", outputDir, DIR_SEP, s);
   }
}

static
void DumpTextObjInAscii(FP, ObjPtr)
   FILE *FP;
   register struct ObjRec *ObjPtr;
{
   struct TextRec *text_ptr=ObjPtr->detail.t;
   MiniLinesInfo *minilines=(&text_ptr->minilines);

   if (text_ptr->pen == NONEPAT) return;
   DumpMiniLinesInAscii(FP, minilines, &totalBBoxValid);
   if (fprintf(FP, "\n") == EOF) writeFileFailed = TRUE;
}

static
void DumpAttrsInAscii(FP, AttrPtr)
   FILE *FP;
   register struct AttrRec *AttrPtr;
{
   for ( ; AttrPtr != NULL; AttrPtr=AttrPtr->prev) {
      if (!AttrPtr->shown) {
         continue;
      }
      if (!AttrPtr->nameshown && *AttrPtr->attr_name.s == '!' &&
            strcmp(AttrPtr->attr_name.s, "!PAGE_NUM=") == 0) {
         if (pageLayoutMode == PAGE_STACK) {
            MiniLineInfo *pMiniLine=AttrPtr->obj->detail.t->minilines.first;
            char *c_ptr;

            if (pMiniLine != NULL &&
                  pMiniLine->first_block == pMiniLine->last_block &&
                  (strstr(pMiniLine->first_block->seg->dyn_str.s,
                  "!(STACKED_PAGE_NUM)") != NULL ||
                  strstr(pMiniLine->first_block->seg->dyn_str.s,
                  "!(STACKED_NUM_PAGES)") != NULL)) {
               char *psz_saved=NULL, *c_ptr1;
               int page_num_str_len=strlen("!(STACKED_PAGE_NUM)");
               int num_pages_str_len=strlen("!(STACKED_NUM_PAGES)");

               psz_saved = UtilStrDup(pMiniLine->first_block->seg->dyn_str.s);
               if (psz_saved == NULL) FailAllocMessage();
               while ((c_ptr=strstr(pMiniLine->first_block->seg->dyn_str.s,
                     "!(STACKED_PAGE_NUM)")) != NULL) {
                  c_ptr1 = &c_ptr[page_num_str_len];
                  sprintf(c_ptr, "%1d", printingPageNum);
                  c_ptr = &c_ptr[strlen(c_ptr)];
                  while (*c_ptr1 != '\0') *c_ptr++ = *c_ptr1++;
                  *c_ptr = '\0';
               }
               while ((c_ptr=strstr(pMiniLine->first_block->seg->dyn_str.s,
                     "!(STACKED_NUM_PAGES)")) != NULL) {
                  c_ptr1 = &c_ptr[num_pages_str_len];
                  sprintf(c_ptr, "%1d", lastPageNum);
                  c_ptr = &c_ptr[strlen(c_ptr)];
                  while (*c_ptr1 != '\0') *c_ptr++ = *c_ptr1++;
                  *c_ptr = '\0';
               }
               DumpTextObjInAscii(FP, AttrPtr->obj);
               DynStrSet(&pMiniLine->first_block->seg->dyn_str, psz_saved);
               UtilFree(psz_saved);
            } else {
               DumpTextObjInAscii(FP, AttrPtr->obj);
            }
         } else {
            MiniLineInfo *pMiniLine=AttrPtr->obj->detail.t->minilines.first;
            char *c_ptr;

            if (pMiniLine != NULL) {
               char *psz_saved, *c_ptr1, *c_ptr2;

               psz_saved = UtilStrDup(pMiniLine->first_block->seg->dyn_str.s);
               if (psz_saved == NULL) FailAllocMessage();
               for (c_ptr=pMiniLine->first_block->seg->dyn_str.s;
                     *c_ptr != '\0'; c_ptr++) {
                  int len=strlen("!(TILED_PAGE_ROW)");

                  if (*c_ptr == '!' &&
                        (strncmp(c_ptr, "!(TILED_PAGE_ROW)", len) == 0 ||
                        strncmp(c_ptr, "!(TILED_PAGE_COL)", len) == 0)) {
                     if (strncmp(c_ptr, "!(TILED_PAGE_ROW)", len) == 0) {
                        sprintf(c_ptr, "%1d", printingPageRow);
                     } else {
                        sprintf(c_ptr, "%1d", printingPageCol);
                     }
                     c_ptr1 = &c_ptr[len];
                     c_ptr = c_ptr2 = &c_ptr[strlen(c_ptr)];
                     while (*c_ptr1 != '\0') *c_ptr2++ = *c_ptr1++;
                     *c_ptr2 = '\0';
                     c_ptr--;
                  }
               }
               DumpTextObjInAscii(FP, AttrPtr->obj);
               DynStrSet(&pMiniLine->first_block->seg->dyn_str, psz_saved);
               UtilFree(psz_saved);
            } else {
               DumpTextObjInAscii(FP, AttrPtr->obj);
            }
         }
      } else {
         DumpTextObjInAscii(FP, AttrPtr->obj);
      }
   }
}

static
void DumpAnObjInAscii(FP, ObjPtr)
   FILE *FP;
   register struct ObjRec *ObjPtr;
{
   switch (ObjPtr->type) {
   case OBJ_POLY:
   case OBJ_BOX:
   case OBJ_OVAL:
   case OBJ_POLYGON:
   case OBJ_ARC:
   case OBJ_RCBOX:
   case OBJ_XBM:
   case OBJ_XPM:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpAttrsInAscii(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_TEXT:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         DumpTextObjInAscii(FP, ObjPtr);
      }
      break;
   case OBJ_SYM:
   case OBJ_ICON:
   case OBJ_GROUP:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         struct ObjRec *obj_ptr=ObjPtr->detail.r->last;

         for ( ; obj_ptr != NULL; obj_ptr = obj_ptr->prev) {
            obj_ptr->tmp_parent = ObjPtr;
            DumpAnObjInAscii(FP, obj_ptr);
         }
         DumpAttrsInAscii(FP, ObjPtr->lattr);
      }
      break;
   case OBJ_PIN:
      if (!colorLayers ||
            ObjPtr->tmp_parent!=NULL || ObjInVisibleLayer(ObjPtr)) {
         struct ObjRec *obj_ptr=GetPinObj(ObjPtr);

         obj_ptr->tmp_parent = ObjPtr;
         DumpAnObjInAscii(FP, obj_ptr);

         DumpAttrsInAscii(FP, ObjPtr->lattr);
      }
      break;
   }
}

#define WRITEBYTE(fp,byte) fputc((byte),(fp))

static
void WriteWord(fp, word)
   FILE *fp;
   unsigned short word;
{
   WRITEBYTE(fp, word&0xff);
   WRITEBYTE(fp, ((unsigned short)(word&0xff00))>>8);
}

static
void WriteDoubleWord(fp, dword)
   FILE *fp;
   unsigned long dword;
{
   WRITEBYTE(fp, dword&0xff);
   WRITEBYTE(fp, (dword&0xff00)>>8);
   WRITEBYTE(fp, (dword&0xff0000)>>16);
   WRITEBYTE(fp, (dword&0xff000000)>>24);
}

static char	psBopHook[MAXSTRING], psEopHook[MAXSTRING];
static int	psBopHookStatus=INVALID, psEopHookStatus=INVALID;
static int	colorBgInPrintingColorPS=INVALID;
static char	xbmToTiffCmd[MAXSTRING+1];
static char	psToPdfCmd[MAXSTRING+1];
static char	epsiExportExtension[MAXSTRING];
static int	overrideEPSIExportExtension=INVALID;
static char	pdfExportExtension[MAXSTRING];
static int	numberFileInPrintOnePage=INVALID;

void SetBopHook(buf)
   char *buf;
{
   strcpy(psBopHook, buf);
   psBopHookStatus = TRUE;
}

void SetEopHook(buf)
   char *buf;
{
   strcpy(psEopHook, buf);
   psEopHookStatus = TRUE;
}

static
void CleanUpDumpInitDefaults()
{
   *psBopHook = *psEopHook = '\0';
   psBopHookStatus = psEopHookStatus = INVALID;
   minimalEPS = INVALID;
   colorBgInPrintingColorPS = INVALID;
   *xbmToTiffCmd = '\0';
   *psToPdfCmd = '\0';
   *epsiExportExtension = '\0';
   *pdfExportExtension = '\0';
   overrideEPSIExportExtension = INVALID;
   numberFileInPrintOnePage = INVALID;
}

static
void GenDumpInitDefaults()
{
   static int init_tiffepsi=FALSE;
   char *c_ptr;

   if (minimalEPS == INVALID) {
      minimalEPS = !usePsAdobeString;
#ifdef _MINIMAL_EPS
      minimalEPS = (_MINIMAL_EPS != 0);
#endif /* _MINIMAL_EPS */
      if (!PRTGIF || cmdLineOpenDisplay) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"MinimalEPS")) != NULL) {
            minimalEPS = (UtilStrICmp(c_ptr, "true") == 0);
         }
      }
   }
   if (colorBgInPrintingColorPS == INVALID) {
      colorBgInPrintingColorPS = FALSE;
      if (!PRTGIF) {
         if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME,
               "ColorBgInPrintingColorPS")) != NULL &&
               UtilStrICmp(c_ptr, "true") == 0) {
            colorBgInPrintingColorPS = TRUE;
         }
      }
   }
   if (psBopHookStatus == INVALID) {
      *psBopHook = '\0';
      if (!PRTGIF) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PSBopHook")) != NULL) {
            strcpy(psBopHook, c_ptr);
            psBopHookStatus = TRUE;
         } else {
            psBopHookStatus = FALSE;
         }
      }
   }
   if (psEopHookStatus == INVALID) {
      *psEopHook = '\0';
      if (!PRTGIF) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PSEopHook")) != NULL) {
            strcpy(psEopHook, c_ptr);
            psEopHookStatus = TRUE;
         } else {
            psEopHookStatus = FALSE;
         }
      }
   }
   if (!init_tiffepsi) {
      init_tiffepsi = TRUE;
      if (PRTGIF) {
         generateTiffEPSI = cmdLineTiffEPSI;
         strcpy(xbmToTiffCmd, "xbmtopbm %s | pnmtotiff -none > %s");
      } else {
         if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "TiffEPSI")) != NULL) {
            if (UtilStrICmp(c_ptr, "true") == 0) {
               fprintf(stderr, "Obsoleted %s*%s used.  Ignored.\n\t%s %s %s.\n",
                     TOOL_NAME, "TiffEPSI",
                     "TiffEPSI export is now supported by", TOOL_NAME,
                     "directly as an export format");
            }
         }
         if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "XbmToTiff")) != NULL) {
            int count=0;

            strcpy(xbmToTiffCmd, c_ptr);
            for (c_ptr=strstr(xbmToTiffCmd,"%s"); c_ptr!=NULL;
                  c_ptr=strstr(++c_ptr,"%s")) {
               count++;
            }
            if (count != 1 && count != 2) {
               sprintf(gszMsgBox, "Invalid %s*%s: %s resource.\n\n'%s' used.",
                     TOOL_NAME, "XbmToTiff", xbmToTiffCmd,
                     "xbmtopbm %s | pnmtotiff -none > %s");
               MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
               strcpy(xbmToTiffCmd, "xbmtopbm %s | pnmtotiff -none > %s");
            }
         } else {
            strcpy(xbmToTiffCmd, "xbmtopbm %s | pnmtotiff -none > %s");
         }
      }
   }
   if (overrideEPSIExportExtension == INVALID) {
      *epsiExportExtension = '\0';
      overrideEPSIExportExtension = FALSE;
      if (!PRTGIF) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"EPSIExportExtension")) !=
               NULL) {
            strcpy(epsiExportExtension, c_ptr);
            if (strchr(epsiExportExtension, '.') != NULL ||
                  strchr(epsiExportExtension, ' ') != NULL) {
               fprintf(stderr, "Invalid %s*EPSIExportExtension: '%s'.\n",
                     TOOL_NAME, epsiExportExtension);
               *epsiExportExtension = '\0';
            }
         }
      }
   }
   strcpy(pdfExportExtension, "pdf");
   if (!PRTGIF || cmdLineOpenDisplay) {
      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PdfFileExtension")) !=
            NULL) {
         strcpy(pdfExportExtension, c_ptr);
         UtilTrimBlanks(pdfExportExtension);
         if (*pdfExportExtension == '\0') {
            fprintf(stderr, "Invalid %s*PdfFileExtension: '%s'.  'pdf' used.\n",
                  TOOL_NAME, pdfExportExtension);
            strcpy(pdfExportExtension, "pdf");
         }
      }
   }
   strcpy(psToPdfCmd, "ps2pdf %s");
   if (!PRTGIF || cmdLineOpenDisplay) {
      if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "PsToPdf")) != NULL) {
         int count=0;

         strcpy(psToPdfCmd, c_ptr);
         for (c_ptr=strstr(psToPdfCmd,"%s"); c_ptr!=NULL;
               c_ptr=strstr(++c_ptr,"%s")) {
            count++;
         }
         if (count != 1) {
            sprintf(gszMsgBox, "Invalid %s*%s: %s resource.\n\n'%s' used.",
                  TOOL_NAME, "PsToPdf", psToPdfCmd, "ps2pdf %s");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            strcpy(psToPdfCmd, "ps2pdf %s");
         }
      }
   }
   if (numberFileInPrintOnePage == INVALID) {
      numberFileInPrintOnePage = FALSE;
      if (PRTGIF) {
         if (cmdLineOneFilePerPage) numberFileInPrintOnePage = TRUE;
      } else {
         if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME,
               "NumberFileInPrintOnePage")) != NULL &&
               UtilStrICmp(c_ptr, "true") == 0) {
            numberFileInPrintOnePage = TRUE;
         }
      }
   }
   if (dontUseShortHandPS == INVALID) {
      dontUseShortHandPS = FALSE;
      if (!PRTGIF || cmdLineOpenDisplay) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"DontUseShortHandPS")) !=
               NULL && UtilStrICmp(c_ptr, "true") == 0) {
            dontUseShortHandPS = TRUE;
         }
      }
   }
   if (dontCondense == INVALID) {
      dontCondense = TRUE;
      if (!PRTGIF || cmdLineOpenDisplay) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"DontCondensePSFile")) !=
               NULL && UtilStrICmp(c_ptr, "false") == 0) {
            dontCondense = FALSE;
         }
      }
   }
   if (stripSimpleComments == INVALID) {
      stripSimpleComments = FALSE;
      if (!PRTGIF || cmdLineOpenDisplay) {
         if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME,
               "StripCondensedPSComments")) != NULL &&
               UtilStrICmp(c_ptr, "false") != 0) {
            static int shown_once=FALSE;

            if (!shown_once) {
               sprintf(gszMsgBox, "Obsoleted %s.%s used.  Ignored.",
                     TOOL_NAME, "StripCondensedPSComments");
               MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
               shown_once = TRUE;
            }
         }
      }
   }
   if (psRegMarksInTiledPageMode == INVALID) {
      psRegMarksInTiledPageMode = FALSE;
      psRegMarksGray = 0.95;
      if (!PRTGIF || cmdLineOpenDisplay) {
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,
               "PsRegMarksInTiledPageMode")) != NULL &&
               UtilStrICmp(c_ptr, "true") == 0) {
            psRegMarksInTiledPageMode = TRUE;
         }
         if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PsRegMarksGray")) !=
               NULL) {
            psRegMarksGray = (float)atof(c_ptr);
            if (psRegMarksGray < 0.0 || psRegMarksGray > 1.0) {
               fprintf(stderr, "Invalid %s*PsRegMarksGray: '%s', %s.\n",
                     TOOL_NAME, c_ptr, "0.95 used");
               psRegMarksGray = 0.95;
            }
         }
      }
   }
}

static
void GenTiffFile(tmp_fname, epsi_fname, condense, condensed_already)
   char *tmp_fname, *epsi_fname;
   int condense, condensed_already;
{
   register int i;
   char tiff_fname[MAXPATHLENGTH+1], xbm_fname[MAXPATHLENGTH+1];
   char cmd[MAXPATHLENGTH+1], buf[1024], * c_ptr;
   int bytes_read, len, count, tiff_fd, epsi_fd=(-1);
   FILE *tmp_fp=NULL, *epsi_fp=NULL, *pfp=NULL;
   struct stat stat_buf;
   long tmp_bytes, tiff_bytes;

   sprintf(xbm_fname, "%s%c%s", curDir, DIR_SEP, curFileName);
   len = strlen(xbm_fname);
   for (i=len-1; xbm_fname[i] != '.'; i--) ;
   sprintf(&xbm_fname[i], ".%s", XBM_FILE_EXT);
   ModifyOutputFileName(xbm_fname);

   sprintf(tiff_fname, "%s%c%s", curDir, DIR_SEP, curFileName);
   len = strlen(tiff_fname);
   for (i=len-1; tiff_fname[i] != '.'; i--) ;
   sprintf(&tiff_fname[i], ".tif");
   ModifyOutputFileName(tiff_fname);

   count = 0;
   for (c_ptr=strstr(xbmToTiffCmd,"%s"); c_ptr!=NULL;
         c_ptr=strstr(++c_ptr,"%s")) {
      count++;
   }
   if (count == 1) {
      sprintf(cmd, xbmToTiffCmd, xbm_fname);
   } else {
      sprintf(cmd, xbmToTiffCmd, xbm_fname, tiff_fname);
   }
   unlink(tiff_fname);

   sprintf(gszMsgBox, "Executing '%s'...", cmd);
   SetStringStatus(gszMsgBox);
   XSync(mainDisplay, False);
   if ((pfp=(FILE*)popen(cmd, "r")) == NULL) {
      sprintf(gszMsgBox, "Fail to execute '%s'.\n\nEPSI file not generated!",
            cmd);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   while (fgets(buf, sizeof(buf), pfp) != NULL) {
      Msg(buf);
   }
   pclose(pfp);
   SetStringStatus("...Done");

   if (stat(tmp_fname, &stat_buf) != 0) {
      sprintf(gszMsgBox, "Fail to get file info for '%s'.\n\n%s!",
            tmp_fname, "EPSI file not generated!");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   tmp_bytes = (long)stat_buf.st_size;
   if (stat(tiff_fname, &stat_buf) != 0) {
      sprintf(gszMsgBox, "Fail to get file info for '%s'.\n\n%s!",
            tiff_fname, "EPSI file not generated!");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   tiff_bytes = (long)stat_buf.st_size;

   if ((tmp_fp=fopen(tmp_fname, "r")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for read.", tmp_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   if ((tiff_fd=open(tiff_fname, O_RDONLY)) == (-1)) {
      sprintf(gszMsgBox, "Cannot open '%s' for read.", tiff_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      fclose(tmp_fp);
      return;
   }
   if (PRTGIF && cmdLineStdOut) {
      epsi_fp = stdout;
   } else if ((epsi_fp=fopen(epsi_fname, "w")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for read.", tiff_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      fclose(tmp_fp);
      close(tiff_fd);
      unlink(tiff_fname);
      return;
   }
   WRITEBYTE(epsi_fp, 0xc5);
   WRITEBYTE(epsi_fp, 0xd0);
   WRITEBYTE(epsi_fp, 0xd3);
   WRITEBYTE(epsi_fp, 0xc6);
   WriteDoubleWord(epsi_fp, 0x0000001e);
   WriteDoubleWord(epsi_fp, tmp_bytes);
   WriteDoubleWord(epsi_fp, 0L);
   WriteDoubleWord(epsi_fp, 0L);
   WriteDoubleWord(epsi_fp, tmp_bytes+0x1e);
   WriteDoubleWord(epsi_fp, tiff_bytes);
   WriteWord(epsi_fp, 0xffff);

   if (!writeFileFailed) {
      CondenseAndCopyPostScriptFile(tmp_fp, epsi_fp, tmp_fname,
            (!condensed_already && condense && !dontCondense));
   }
   fclose(tmp_fp);
   if (!(PRTGIF && cmdLineStdOut)) fclose(epsi_fp);
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      if (PRTGIF && cmdLineStdOut) {
         sprintf(gszMsgBox, "Write file failed.\n\nFile system may be full.");
      } else {
         sprintf(gszMsgBox,
               "Fail to write to '%s'.\n\nFile system may be full.",
               epsi_fname);
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      close(tiff_fd);
      unlink(tiff_fname);
      return;
   }
   if (PRTGIF && cmdLineStdOut) {
      epsi_fd = 1;
   } else if ((epsi_fd=open(epsi_fname, O_WRONLY|O_APPEND)) == (-1)) {
      sprintf(gszMsgBox, "Cannot open '%s' for append.", epsi_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      close(tiff_fd);
      unlink(tiff_fname);
      return;
   }
   while ((bytes_read=read(tiff_fd, buf, sizeof(buf))) > 0) {
      if (write(epsi_fd, buf, bytes_read) <= 0) {
         writeFileFailed = TRUE;
         break;
      }
   }
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            epsi_fname);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   }
   if (!(PRTGIF && cmdLineStdOut)) close(epsi_fd);
   close(tiff_fd);
   unlink(tiff_fname);
}

static int psA4PaperSize=FALSE;

static
void InitNeedA4()
{
   static int nInitialized=FALSE;

   if (nInitialized) return;
   nInitialized = TRUE;
   if (cmdLineA4) {
      psA4PaperSize = TRUE;
   } else if (!PRTGIF) {
      char *c_ptr;

      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PSA4PaperSize")) != NULL &&
            UtilStrICmp(c_ptr, "true") == 0) {
         psA4PaperSize = TRUE;
      }
   }
}

static
int NeedA4()
{
   int width_to_match, height_to_match;

   InitNeedA4();
   if (!psA4PaperSize) return FALSE;
   width_to_match = (int)(825.0*PIX_PER_INCH/printMag);
   height_to_match = (int)(117.0*PIX_PER_INCH*10.0/printMag);
   return (onePageWidth == width_to_match && onePageHeight == height_to_match);
}

void SetOutputFileName(ps_file, file_ext, pn_short_name, ppsz_rest)
   char *ps_file, *file_ext, **ppsz_rest;
   int *pn_short_name;
{
   if (PRTGIF) {
      char *psz_file=(*ppsz_rest);

      if (cmdLineOneFilePerPage) {
         if (curPage != NULL && curPage->page_file_name != NULL &&
               *curPage->page_file_name != '\0') {
            sprintf(ps_file, "%s.%s", curPage->page_file_name, file_ext);
         } else {
            sprintf(ps_file, "%s-%1d.%s", psz_file, cmdLinePageNum, file_ext);
         }
      } else {
         sprintf(ps_file, "%s.%s", psz_file, file_ext);
      }
   } else {
      int len=0, use_page_num=FALSE;
      char *psz=NULL;

      if (dumpOneFilePerPage ||
            (dumpOnePageInStackMode && numberFileInPrintOnePage)) {
         if (curPage != NULL && curPage->page_file_name != NULL &&
               *curPage->page_file_name != '\0') {
            sprintf(ps_file, "%s%c%s.", curDir, DIR_SEP,
                  curPage->page_file_name);
         } else {
            sprintf(ps_file, "%s%c%s", curDir, DIR_SEP, curFileName);
            use_page_num = TRUE;
         }
      } else {
         sprintf(ps_file, "%s%c%s", curDir, DIR_SEP, curFileName);
      }
      len = strlen(ps_file);
      if (len > 3 && strcmp(&ps_file[len-3], ".gz") == 0) {
         ps_file[len-3] = '\0';
         len -= 3;
      }
      if ((psz=UtilStrRChr(ps_file, '.')) == NULL) {
         TgAssert(FALSE,
               "Cannot find '.' in curFileName in SetOutputFileName()", NULL);
      }
      if (use_page_num) {
         sprintf(psz, "-%1d.%s", curPageNum, file_ext);
      } else {
         sprintf(psz, ".%s", file_ext);
      }
   }
   ModifyOutputFileName(ps_file);

   if (!PRTGIF && pn_short_name != NULL && ppsz_rest != NULL) {
      if (((*pn_short_name)=IsPrefix(bootDir, ps_file, ppsz_rest))) {
         *ppsz_rest = (&(*ppsz_rest)[1]);
      }
   }
}

static
void DumpRegMark(fp, pstem)
   FILE *fp;
   float *pstem;
{
   float fval=(*pstem);

   fprintf(dumpFP, "   %s ", gPsCmd[PS_GSAVE]);
   fprintf(dumpFP, "%.3f 0 %s ", -fval, gPsCmd[PS_RMOVETO]);
   fprintf(dumpFP, "2 %.3f %s 0 %s ", fval, gPsCmd[PS_MUL], gPsCmd[PS_RLINETO]);
   fprintf(dumpFP, "%s %s\n", gPsCmd[PS_STROKE], gPsCmd[PS_GRESTORE]);
   fprintf(dumpFP, "   %s ", gPsCmd[PS_GSAVE]);
   fprintf(dumpFP, "0 %.3f %s ", -fval, gPsCmd[PS_RMOVETO]);
   fprintf(dumpFP, "0 2 %.3f %s %s ", fval, gPsCmd[PS_MUL], gPsCmd[PS_RLINETO]);
   fprintf(dumpFP, "%s %s\n", gPsCmd[PS_STROKE], gPsCmd[PS_GRESTORE]);
}

static
int DoGenDump(FileName)
   char *FileName;
{
   register struct ObjRec *obj_ptr;
   char cmd[MAXSTRING+1], tmp_str[MAXSTRING+1], ps_file[MAXSTRING+1];
   char *rest=NULL, loc_time[MAXSTRING+1], *file_ext=NULL, *output_type=NULL;
   int i, len, short_name=FALSE, export_cap=0;
   time_t tloc;
   struct DocFontRec *next_df;

   if (!curDirIsLocal) {
      MsgBox("Cannot print a remote file!", TOOL_NAME, INFO_MB);
      return FALSE;
   }
   GenDumpInitDefaults();

   if (botObj==NULL && ((pageLayoutMode==PAGE_STACK &&
         dumpOnePageInStackMode) || pageLayoutMode==PAGE_TILE)) {
      sprintf(gszMsgBox, "No objects to print.");
      if (PRTGIF) {
         fprintf(stderr, "%s\n", gszMsgBox);
      } else {
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
      return FALSE;
   }
   if (whereToPrint != PRINTER) {
      if (!PRTGIF && !curFileDefined) {
         sprintf(gszMsgBox, "No current file.\n\nCannot generate %s output!",
               GetExportName(whereToPrint, EXPORT_THIS));
         if (PRTGIF) {
            fprintf(stderr, "%s\n", gszMsgBox);
         } else {
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
         return FALSE;
      }
   }
   if (!PRTGIF && printingPageNum==printingFirstPageNum) {
      Msg("Generating print file ...");
   }
   if (whereToPrint >= MAXDEFWHERETOPRINT) {
      export_cap = GetExportFilterCapability(whereToPrint);
   }
   if (whereToPrint == XBM_FILE || whereToPrint == GIF_FILE ||
         whereToPrint == HTML_FILE) {
      if (topObj == NULL) {
         sprintf(gszMsgBox, "No objects to export.");
         if (PRTGIF) {
            fprintf(stderr, "%s\n", gszMsgBox);
         } else {
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
      } else {
         int saved_colordump=colorDump, watch_cursor=watchCursorOnMainWindow;

         if (!watch_cursor) {
            SetWatchCursor(drawWindow);
            SetWatchCursor(mainWindow);
         }
         if (whereToPrint == GIF_FILE || whereToPrint == HTML_FILE) {
            colorDump = TRUE;
         }
         DumpXBitmapFile(FALSE, dumpOneFilePerPage ||
               (dumpOnePageInStackMode && numberFileInPrintOnePage));
         colorDump = saved_colordump;
         if (!watch_cursor) {
            SetDefaultCursor(mainWindow);
            ShowCursor(); 
         }
      }
      return TRUE;
   }
   /*
    * At this point, whereToPrint can only be PostScript, ASCII Text,
    *      or external export type (if whereToPrint >= MAXDEFWHERETOPRINT).
    */
   if (printingPageNum == printingFirstPageNum) {
      sprintf(tmpFile, "%sTgifXXXXXX", TMP_DIR);
      mktemp(tmpFile);
      unlink(tmpFile);

      if ((dumpFP=fopen(tmpFile, "w")) == NULL) {
         sprintf(tmp_str, "Cannot create '%s', print aborted.", tmpFile);
         if (PRTGIF) {
            fprintf(stderr, "%s\n", tmp_str);
         } else {
            MsgBox(tmp_str, TOOL_NAME, INFO_MB);
         }
         return FALSE;
      }

      if (PRTGIF && !cmdLineStdOut) {
         fprintf(stderr, "Writing to '%s' ...\n", tmpFile);
      }
      writeFileFailed = FALSE;
      ResetGrayDetection();
   }
   if (whereToPrint < MAXDEFWHERETOPRINT &&
         printingPageNum == printingFirstPageNum &&
         whereToPrint != TEXT_FILE) {
      PSUseShortHand();
      if (dontUseShortHandPS) {
         PSDontUseShortHand();
      }
      if (usePsAdobeString) {
         switch (whereToPrint) {
         case PRINTER:
         case PS_FILE:
         case PDF_FILE:
            if (*adobeString == '\0') {
               if (fprintf(dumpFP, "%%!PS-Adobe-2.0\n") == EOF) {
                  writeFileFailed = TRUE;
               }
            } else if (fprintf(dumpFP,"%%!PS-Adobe-%s\n",adobeString) == EOF) {
               writeFileFailed = TRUE;
            }
            break;
         case LATEX_FIG:
         case EPSI_FILE:
            if (*adobeString == '\0') {
               if (fprintf(dumpFP, "%%!PS-Adobe-2.0 EPSF-1.2\n") == EOF) {
                  writeFileFailed = TRUE;
               }
            } else if (*epsfString == '\0') {
               if (fprintf(dumpFP, "%%!PS-Adobe-%s\n", adobeString) == EOF) {
                  writeFileFailed = TRUE;
               }
            } else {
               if (fprintf(dumpFP, "%%!PS-Adobe-%s EPSF-%s\n",
                     adobeString, epsfString) == EOF) {
                  writeFileFailed = TRUE;
               }
            }
            break;
         }
      } else {
         if (preDumpSetup) PSUsePSAdobe();
         if (fprintf(dumpFP, "%%!\n") == EOF) {
            writeFileFailed = TRUE;
         }
      }
      if (pageLayoutMode == PAGE_STACK &&
            printingFirstPageNum == printingLastPageNum) {
         DumpBBox(dumpFP, ALL_BBOX, NULL);
      } else if (fprintf(dumpFP,"%%%%BoundingBox: (atend)\n") == EOF) {
         writeFileFailed = TRUE;
      }
      if (PRTGIF && !cmdLineOpenDisplay) {
         if (fprintf(dumpFP, "%%%%Title: %s\n", FileName) == EOF) {
            writeFileFailed = TRUE;
         }
      } else if (curFileDefined) {
         int gzipped=FALSE;

         strcpy(tmp_str, curFileName);
         if (FileNameHasExtension(tmp_str, OBJ_FILE_TYPE, &gzipped, NULL) &&
               gzipped) {
            char *psz=UtilStrRChr(tmp_str, '.');

            if (psz != NULL) *psz = '\0';
         }
         len = strlen(tmp_str);
         for (i=len-1; tmp_str[i] != '.'; i--) ;
         tmp_str[i] = '\0';
         if (fprintf(dumpFP, "%%%%Title: %s\n", tmp_str) == EOF) {
            writeFileFailed = TRUE;
         }
      } else if (fprintf(dumpFP, "%%%%Title: [Unnamed]\n") == EOF) {
         writeFileFailed = TRUE;
      }
      time(&tloc);
      strcpy(loc_time, ctime(&tloc));
      loc_time[24] = '\0';
      if (fprintf(dumpFP, "%%%%CreationDate: %s\n", loc_time) == EOF) {
         writeFileFailed = TRUE;
      }
      SetFullVersionString();
      if (fprintf(dumpFP, "%%%%Creator: %s by %s\n",
            fullToolName,
            "William Chia-Wei Cheng (william@cs.UCLA.edu)") == EOF) {
         writeFileFailed = TRUE;
      }
      if (whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) {
         if (fprintf(dumpFP, "%%%%Orientation: %s\n",
               (pageStyle==PORTRAIT ? "Portrait" : "Landscape")) == EOF) {
            writeFileFailed = TRUE;
         }
      }
      if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
            !minimalEPS) {
         if (fprintf(dumpFP, "%%%%Pages: %1d\n", 
               pageLayoutMode == PAGE_STACK ? lastPageNum :
               paperCol * paperRow) == EOF) {
            writeFileFailed = TRUE;
         }
         if (fprintf(dumpFP, "%%%%DocumentFonts: (atend)\n") == EOF) {
            writeFileFailed = TRUE;
         }
         if (NeedA4()) {
            if (fprintf(dumpFP, "%%%%DocumentPaperSizes: a4\n") == EOF) {
               writeFileFailed = TRUE;
            }
         }
         if (fprintf(dumpFP, "%%%%EndComments\n") == EOF) {
            writeFileFailed = TRUE;
         }
      }
      if (whereToPrint == EPSI_FILE && !generateTiffEPSI) {
         SaveStatusStrings();
         SetStringStatus("Generating preview bitmap...");
         GenPreviewBitmap(dumpFP, llxPage, llyPage, urxPage, uryPage);
         RestoreStatusStrings();
      }

      if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
            !minimalEPS) {
         if (fprintf(dumpFP, "%%%%BeginProlog\n") == EOF) {
            writeFileFailed = TRUE;
         }
      }

      DumpPSMacro(dumpFP);

      PrepareEightBitFontInfo();
      DumpEightBitFontInfo(dumpFP);

      if (fprintf(dumpFP, "end\n\n") == EOF) writeFileFailed = TRUE;
      if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
            !minimalEPS) {
         if (fprintf(dumpFP, "%%%%EndProlog\n") == EOF) writeFileFailed = TRUE;
         if (NeedA4()) {
            if (fprintf(dumpFP, "%%%%BeginSetup\n") == EOF ||
                  fprintf(dumpFP, "%%%%PaperSize: a4\n") == EOF ||
                  fprintf(dumpFP, "%%%%BeginPaperSize: a4\n") == EOF ||
                  fprintf(dumpFP, "a4\n") == EOF ||
                  fprintf(dumpFP, "%%%%EndPaperSize\n") == EOF ||
                  fprintf(dumpFP, "%%%%EndSetup\n") == EOF) {
               writeFileFailed = TRUE;
            }
         }
      }
   }
   if (pageLayoutMode == PAGE_STACK || (pageLayoutMode == PAGE_TILE &&
         ((paperCol == 1 && paperRow == 1) ||
         ((whereToPrint == LATEX_FIG || whereToPrint == EPSI_FILE) &&
         !dumpOnePageInTileMode)))) {
      printingPageRow = printingPageCol = 1;
      if (whereToPrint >= MAXDEFWHERETOPRINT) {
          if (!ExportVectorFile(dumpFP, dumpOneFilePerPage ||
                (dumpOnePageInStackMode && numberFileInPrintOnePage),
                printingPageNum, printingFirstPageNum, printingPageRow,
                printingPageCol)) {
             return FALSE;
          }
      } else if (whereToPrint == TEXT_FILE) {
         if (printingPageNum != printingFirstPageNum) {
            fprintf(dumpFP, "\014\n");
         }
         for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev) {
            obj_ptr->tmp_parent = NULL;
            DumpAnObjInAscii(dumpFP, obj_ptr);
         }
      } else {
         if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
               !minimalEPS) {
            fprintf(dumpFP, "%%%%Page: %1d %1d\n\n",
                  printingPageNum-printingFirstPageNum+1,
                  printingPageNum-printingFirstPageNum+1);
         }
         if (!minimalEPS) {
            DumpBBox(dumpFP, PAGE_BBOX, NULL);
         }
         if (totalBBoxValid) {
            if (topObj != NULL) {
               if (llxPage < llxTotal) llxTotal = llxPage;
               if (llyPage < llyTotal) llyTotal = llyPage;
               if (urxPage > urxTotal) urxTotal = urxPage;
               if (uryPage > uryTotal) uryTotal = uryPage;
            }
         } else if (topObj != NULL) {
            totalBBoxValid = TRUE;
            llxTotal = llxPage; llyTotal = llyPage;
            urxTotal = urxPage; uryTotal = uryPage;
         }
         if (psBopHookStatus == TRUE && (whereToPrint == PRINTER ||
               whereToPrint == PS_FILE || whereToPrint == PDF_FILE)) {
            fprintf(dumpFP, "userdict /%s known { %s } if\n\n",
                  psBopHook, psBopHook);
         }
         fprintf(dumpFP, "tgifdict begin\n");
         fprintf(dumpFP, "/tgifsavedpage save def\n\n");
         fprintf(dumpFP, "1 %s\n", gPsCmd[PS_SETMITERLIMIT]);
         fprintf(dumpFP, "1 %s\n\n", gPsCmd[PS_SETLINEWIDTH]);
         fprintf(dumpFP, "0 %s\n\n", gPsCmd[PS_SETGRAY]);

         if (pageStyle == LANDSCAPE) {
            fprintf(dumpFP, "90 %s\n", gPsCmd[PS_ROTATE]);
         }
         fprintf(dumpFP, "%1d %s %s %1d %s %s %s\n", psDotsPerInch,
               psXOffStr[pageStyle], gPsCmd[PS_MUL], psDotsPerInch,
               psYOffStr[pageStyle], gPsCmd[PS_MUL], gPsCmd[PS_TRANSLATE]);

#ifndef METRIC_PIX_PER_INCH
         fprintf(dumpFP, "%1d %1d %s %.3f %s 100 %s %s %s %s\n\n",
               psDotsPerInch, PIX_PER_INCH, gPsCmd[PS_DIV], printMag,
               gPsCmd[PS_MUL], gPsCmd[PS_DIV], gPsCmd[PS_DUP],
               gPsCmd[PS_NEG], gPsCmd[PS_SCALE]);
#else /* METRIC_PIX_PER_INCH */
         fprintf(dumpFP, "%1d %1d %s %.3f %s 100 %s %s %s %s\n\n",
               psDotsPerInch, (gridSystem==ENGLISH_GRID ? PIX_PER_INCH :
               METRIC_PIX_PER_INCH), gPsCmd[PS_DIV], printMag,
               gPsCmd[PS_MUL], gPsCmd[PS_DIV], gPsCmd[PS_DUP],
               gPsCmd[PS_NEG], gPsCmd[PS_SCALE]);
#endif /* ~METRIC_PIX_PER_INCH */

         fprintf(dumpFP, "%s\n\n", gPsCmd[PS_GSAVE]);

         if (needsTgifOrigCTM) {
            fprintf(dumpFP, "/tgiforigctm matrix currentmatrix def\n\n");
         }
         if (colorDump && colorBgInPrintingColorPS &&
               (whereToPrint == PS_FILE || whereToPrint == PDF_FILE ||
               whereToPrint == PRINTER)) {
            fprintf(dumpFP, "%% Background\n");
            if (myBgColor.red==0 && myBgColor.green==0 && myBgColor.blue==0) {
               fprintf(dumpFP, "0 %s\n", gPsCmd[PS_SETGRAY]);
            } else if (myBgColor.red==maxRGB && myBgColor.green==maxRGB &&
                  myBgColor.blue==maxRGB) {
               fprintf(dumpFP, "1 %s\n", gPsCmd[PS_SETGRAY]);
            } else {
               fprintf(dumpFP, "%.3f %.3f %.3f %s\n",
                     ((float)myBgColor.red/maxRGB),
                     ((float)myBgColor.green/maxRGB),
                     ((float)myBgColor.blue/maxRGB), gPsCmd[PS_SETRGBCOLOR]);
            }
            fprintf(dumpFP, "%s\n", gPsCmd[PS_NEWPATH]);
            fprintf(dumpFP,
                  "   0 0 %s 0 %d %s %d %d %s %d 0 %s\n",
                  gPsCmd[PS_MOVETO], onePageHeight, gPsCmd[PS_LINETO],
                  onePageWidth, onePageHeight, gPsCmd[PS_LINETO],
                  onePageWidth, gPsCmd[PS_LINETO]);
            fprintf(dumpFP, "%s %s\n\n", gPsCmd[PS_CLOSEPATH], gPsCmd[PS_FILL]);
         }

         for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev) {
            obj_ptr->tmp_parent = NULL;
            DumpAnObj(dumpFP, obj_ptr);
         }
         fprintf(dumpFP, "%s\n", gPsCmd[PS_GRESTORE]);
         fprintf(dumpFP, "tgifsavedpage restore\n");
         fprintf(dumpFP, "end\n");

         switch (whereToPrint) {
         case PRINTER:
         case PS_FILE:
         case PDF_FILE:
            if (psEopHookStatus == TRUE) {
               fprintf(dumpFP,"userdict /%s known { %s } if\n",
                     psEopHook, psEopHook);
            }
            fprintf(dumpFP, "showpage\n\n");
            break;
         case EPSI_FILE:
         case LATEX_FIG:
            if (showPageInEPS) {
               fprintf(dumpFP, "showpage\n\n");
            }
            break;
         }
      }
   } else if (whereToPrint >= MAXDEFWHERETOPRINT) {
       if (!ExportVectorFile(dumpFP, dumpOneFilePerPage ||
             (dumpOnePageInStackMode && numberFileInPrintOnePage),
             printingPageNum, printingFirstPageNum, printingPageRow,
             printingPageCol)) {
          return FALSE;
       }
   } else if (whereToPrint == TEXT_FILE) {
      for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev) {
         obj_ptr->tmp_parent = NULL;
         DumpAnObjInAscii(dumpFP, obj_ptr);
      }
   } else {
      int row, col;
      float one=(float)atof("1.0"), f_rbx, f_rby, x_dist, y_dist;

      if (tiledPageScaling == one && !msgAboutTiledPageScalingSeen) {
         char msg1[MAXSTRING+1];

         msgAboutTiledPageScalingSeen = TRUE;
         sprintf(gszMsgBox, "%s %s*TiledPageScaling is set to 1.",
               "Output may get truncated because", TOOL_NAME);
         sprintf(msg1, "    Reducing the above setting is recommended.");
         if (PRTGIF) {
            fprintf(stderr, "%s\n%s\n", gszMsgBox, msg1);
         } else {
            TwoLineMsg(gszMsgBox, msg1);
         }
      }
      f_rbx = (float)(((float)PIX_PER_INCH) * 100.0 / printMag *
            psPageWidthInInch[pageStyle]);
      f_rby = (float)(((float)PIX_PER_INCH) * 100.0 / printMag *
            psPageHeightInInch[pageStyle]);
      x_dist = (float)(((float)PIX_PER_INCH) * psPageWidthInInch[pageStyle] *
            (1.0-tiledPageScaling) * 50.0 / printMag);
      y_dist = (float)(((float)PIX_PER_INCH) * psPageHeightInInch[pageStyle] *
            (1.0-tiledPageScaling) * 50.0 / printMag);

      for (row = 0; row < paperRow; row++) {
         printingPageRow = row+1;
         for (col = 0; col < paperCol; col++) {
            struct BBRec page_bbox;
            float real_ltx, real_lty;

            if (dumpOnePageInTileMode &&
                  row*paperCol+col+1 != printingPageNum) {
               continue;
            }
            printingPageCol = col+1;
            page_bbox.ltx = col * onePageWidth;
            page_bbox.lty = row * onePageHeight;
            page_bbox.rbx = page_bbox.ltx + onePageWidth;
            page_bbox.rby = page_bbox.lty + onePageHeight;

            if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
                  !minimalEPS) {
               if (dumpOnePageInTileMode) {
                  fprintf(dumpFP, "%%%%Page: 1 1\n\n");
               } else {
                  fprintf(dumpFP, "%%%%Page: %1d %1d\n\n",
                        col+1+row*paperCol, col+1+row*paperCol);
               }
            }
            if (DumpBBox(dumpFP, PAGE_BBOX, &page_bbox)) {
               if (totalBBoxValid) {
                  if (topObj != NULL) {
                     if (llxPage < llxTotal) llxTotal = llxPage;
                     if (llyPage < llyTotal) llyTotal = llyPage;
                     if (urxPage > urxTotal) urxTotal = urxPage;
                     if (uryPage > uryTotal) uryTotal = uryPage;
                  }
               } else if (topObj != NULL) {
                  totalBBoxValid = TRUE;
                  llxTotal = llxPage; llyTotal = llyPage;
                  urxTotal = urxPage; uryTotal = uryPage;
               }
            }
            if (psBopHookStatus == TRUE && (whereToPrint == PRINTER ||
                  whereToPrint == PS_FILE || whereToPrint == PDF_FILE)) {
               fprintf(dumpFP, "userdict /%s known { %s } if\n\n",
                     psBopHook, psBopHook);
            }
            fprintf(dumpFP, "tgifdict begin\n");
            fprintf(dumpFP, "/tgifsavedpage save def\n\n");
            fprintf(dumpFP, "1 %s\n", gPsCmd[PS_SETMITERLIMIT]);
            fprintf(dumpFP, "1 %s\n\n", gPsCmd[PS_SETLINEWIDTH]);
            fprintf(dumpFP, "0 %s\n\n", gPsCmd[PS_SETGRAY]);

            if (pageStyle == LANDSCAPE) {
               fprintf(dumpFP, "90 %s\n", gPsCmd[PS_ROTATE]);
            }
            fprintf(dumpFP, "%1d %s %s %1d %s %s %s\n",
                  psDotsPerInch, psXOffStr[pageStyle], gPsCmd[PS_MUL],
                  psDotsPerInch, psYOffStr[pageStyle], gPsCmd[PS_MUL],
                  gPsCmd[PS_TRANSLATE]);

#ifndef METRIC_PIX_PER_INCH
            fprintf(dumpFP, "%1d %1d %s %.3f %s 100 %s %s %s %s\n\n",
                  psDotsPerInch, PIX_PER_INCH, gPsCmd[PS_DIV], printMag,
                  gPsCmd[PS_MUL], gPsCmd[PS_DIV], gPsCmd[PS_DUP],
                  gPsCmd[PS_NEG], gPsCmd[PS_SCALE]);
#else /* METRIC_PIX_PER_INCH */
            fprintf(dumpFP, "%1d %1d %s %.3f %s 100 %s %s %s %s\n\n",
                  psDotsPerInch, (gridSystem==ENGLISH_GRID ? PIX_PER_INCH :
                  METRIC_PIX_PER_INCH), gPsCmd[PS_DIV], printMag,
                  gPsCmd[PS_MUL], gPsCmd[PS_DIV], gPsCmd[PS_DUP],
                  gPsCmd[PS_NEG], gPsCmd[PS_SCALE]);
#endif /* ~METRIC_PIX_PER_INCH */

            fprintf(dumpFP, "%s\n\n", gPsCmd[PS_GSAVE]);

            fprintf(dumpFP, "%% SETUP CLIP FOR PAGE\n\n");
            fprintf(dumpFP, "%.3f %.3f %s\n", x_dist, y_dist,
                  gPsCmd[PS_TRANSLATE]);
            fprintf(dumpFP, "%.3f %.3f %s\n", tiledPageScaling,
                  tiledPageScaling, gPsCmd[PS_SCALE]);

            fprintf(dumpFP, "-%1d -%1d %s\n", page_bbox.ltx,
                  page_bbox.lty, gPsCmd[PS_TRANSLATE]);
            real_ltx = (float)(((float)col) *
                  ((float)psPageWidthInInch[pageStyle]) *
                  ((float)PIX_PER_INCH) * 100.0 / printMag);
            real_lty = (float)(((float)row) *
                  ((float)psPageHeightInInch[pageStyle]) *
                  ((float)PIX_PER_INCH) * 100.0 / printMag);
            if (psRegMarksInTiledPageMode && tiledPageScaling < (float)1.0) {
               float stem=((float)(min(x_dist,y_dist)))*((float)0.5);

               fprintf(dumpFP, "%s\n", gPsCmd[PS_GSAVE]);
               fprintf(dumpFP, "   0.1 %s %.3f %s\n", gPsCmd[PS_SETLINEWIDTH],
                     psRegMarksGray, gPsCmd[PS_SETGRAY]);

               fprintf(dumpFP, "   %s %.3f %.3f %s\n", gPsCmd[PS_NEWPATH],
                     real_ltx, real_lty, gPsCmd[PS_MOVETO]);
               DumpRegMark(dumpFP, &stem);
               fprintf(dumpFP, "   %.3f 0 %s\n", f_rbx, gPsCmd[PS_RMOVETO]);
               DumpRegMark(dumpFP, &stem);
               fprintf(dumpFP, "   0 %.3f %s\n", f_rby, gPsCmd[PS_RMOVETO]);
               DumpRegMark(dumpFP, &stem);
               fprintf(dumpFP, "   -%.3f 0 %s\n", f_rbx, gPsCmd[PS_RMOVETO]);
               DumpRegMark(dumpFP, &stem);

               fprintf(dumpFP, "%s\n", gPsCmd[PS_GRESTORE]);
            }
            fprintf(dumpFP, "%s\n", gPsCmd[PS_NEWPATH]);
            fprintf(dumpFP, "   %.3f %.3f %s\n", real_ltx, real_lty,
                  gPsCmd[PS_MOVETO]);
            fprintf(dumpFP, "   %.3f 0 %s\n", f_rbx, gPsCmd[PS_RLINETO]);
            fprintf(dumpFP, "   0 %.3f %s\n", f_rby, gPsCmd[PS_RLINETO]);
            fprintf(dumpFP, "   -%.3f 0 %s\n", f_rbx, gPsCmd[PS_RLINETO]);
            fprintf(dumpFP, "%s %s %s\n\n", gPsCmd[PS_CLOSEPATH],
                  gPsCmd[PS_CLIP], gPsCmd[PS_NEWPATH]);

            if (needsTgifOrigCTM) {
               fprintf(dumpFP, "/tgiforigctm matrix currentmatrix def\n\n");
            }
            if (colorDump && colorBgInPrintingColorPS &&
                  (whereToPrint == PS_FILE || whereToPrint == PDF_FILE ||
                  whereToPrint == PRINTER)) {
               fprintf(dumpFP, "%% Background\n");
               if (myBgColor.red==0 && myBgColor.green==0 &&
                     myBgColor.blue==0) {
                  fprintf(dumpFP, "0 %s\n", gPsCmd[PS_SETGRAY]);
               } else if (myBgColor.red==maxRGB && myBgColor.green==maxRGB &&
                     myBgColor.blue==maxRGB) {
                  fprintf(dumpFP, "1 %s\n", gPsCmd[PS_SETGRAY]);
               } else {
                  fprintf(dumpFP, "%.3f %.3f %.3f %s\n",
                        ((float)myBgColor.red/maxRGB),
                        ((float)myBgColor.green/maxRGB),
                        ((float)myBgColor.blue/maxRGB), gPsCmd[PS_SETRGBCOLOR]);
               }
               fprintf(dumpFP, "%s\n", gPsCmd[PS_NEWPATH]);
               fprintf(dumpFP, "   %.3f %.3f %s\n", real_ltx, real_lty,
                     gPsCmd[PS_MOVETO]);
               fprintf(dumpFP, "   %.3f 0 %s\n", f_rbx, gPsCmd[PS_LINETO]);
               fprintf(dumpFP, "   0 %.3f %s\n", f_rby, gPsCmd[PS_LINETO]);
               fprintf(dumpFP, "   -%.3f 0 %s\n", f_rbx, gPsCmd[PS_LINETO]);
               fprintf(dumpFP, "%s %s %s\n\n", gPsCmd[PS_CLOSEPATH],
                     gPsCmd[PS_FILL], gPsCmd[PS_NEWPATH]);
            }
            for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev) {
               if (Inside(obj_ptr->bbox, page_bbox) ||
                     BBoxIntersect(obj_ptr->bbox, page_bbox)) {
                  obj_ptr->tmp_parent = NULL;
                  DumpAnObj(dumpFP, obj_ptr);
               }
            }
            fprintf(dumpFP, "%s\n", gPsCmd[PS_GRESTORE]);
            fprintf(dumpFP, "tgifsavedpage restore\n");
            fprintf(dumpFP, "end\n");

            switch (whereToPrint) {
            case PRINTER:
            case PS_FILE:
            case PDF_FILE:
               if (psEopHookStatus == TRUE) {
                  fprintf(dumpFP, "userdict /%s known { %s } if\n",
                        psEopHook, psEopHook);
               }
               fprintf(dumpFP, "showpage\n\n");
               break;
            case EPSI_FILE:
            case LATEX_FIG:
               if (showPageInEPS) {
                  fprintf(dumpFP, "showpage\n\n");
               }
               break;
            }
         }
      }
   }
   if (printingPageNum != printingLastPageNum) return TRUE;

   if (whereToPrint < MAXDEFWHERETOPRINT && whereToPrint != TEXT_FILE) {
      if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
            !minimalEPS) {
         if (fprintf(dumpFP, "%%%%Trailer\n") == EOF) {
            writeFileFailed = TRUE;
         }
      }
      if (pageLayoutMode!=PAGE_STACK ||
            printingFirstPageNum!=printingLastPageNum) {
         if (fprintf(dumpFP,"%%%%BoundingBox: %1d %1d %1d %1d\n",
               llxTotal, llyTotal, urxTotal, uryTotal) == EOF) {
            writeFileFailed = TRUE;
         }
      }
      if (fprintf(dumpFP, "%%MatchingCreationDate: %s\n", loc_time) == EOF) {
         writeFileFailed = TRUE;
      }
      if ((whereToPrint != LATEX_FIG && whereToPrint != EPSI_FILE) ||
            !minimalEPS) {
         if (fprintf(dumpFP, "%%%%DocumentFonts: ") == EOF) {
            writeFileFailed = TRUE;
         }
         for ( ; firstDocFont != NULL; firstDocFont = next_df) {
            next_df = firstDocFont->next;
            if (fprintf(dumpFP, "%s", firstDocFont->name) == EOF) {
               writeFileFailed = TRUE;
            }
            if (next_df != NULL) {
               if (fprintf(dumpFP, "\n%%%%+ ") == EOF) {
                  writeFileFailed = TRUE;
               }
            }
            free(firstDocFont);
         }
         if (fprintf(dumpFP, "\n") == EOF) writeFileFailed = TRUE;
         if (fprintf(dumpFP, "%%%%EOF\n") == EOF) writeFileFailed = TRUE;
      }
   }
   fclose(dumpFP);
   dumpFP = NULL;

   EndGrayDetection();

   if (whereToPrint < MAXDEFWHERETOPRINT && whereToPrint != TEXT_FILE &&
         !totalBBoxValid) {
      sprintf(gszMsgBox, "No objects to print.");
      if (PRTGIF) {
         fprintf(stderr, "%s\n", gszMsgBox);
      } else {
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
      return FALSE;
   }
   if (writeFileFailed) {
      writeFileFailed = FALSE;
      sprintf(gszMsgBox, "Fail to write to '%s'.\n\nFile system may be full.",
            tmpFile);
      if (PRTGIF) {
         fprintf(stderr, "%s\n", gszMsgBox);
      } else {
         MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      }
#ifdef KEEP_WHEN_PRINT
      if (whereToPrint != PRINTER) unlink(tmpFile);
#else
      unlink(tmpFile);
#endif
      return FALSE;
   }
   if (preDumpSetup) {
      unlink(tmpFile);
      return TRUE;
   }
   switch (whereToPrint) {
   case PRINTER:
#ifdef VMS
#define PRINT_TRAILER ""
#else
#define PRINT_TRAILER " 2>&1"
#endif
      if (PRTGIF) {
         if (strstr(printCommand, "%s") == NULL) {
            if (lastFile) {
               sprintf(cmd,"%s %s",printCommand,tmpFile);
            } else {
               sprintf(cmd,"%s -h %s",printCommand,tmpFile);
            }
         } else {
            sprintf(cmd, printCommand, tmpFile);
         }
         if (strcmp(printCommand, "cat") != 0) {
            strcat(cmd, PRINT_TRAILER);
         }
         fprintf(stderr, "%s\n", cmd);
      } else {
         if (strstr(printCommand, "%s") == NULL) {
            sprintf(cmd,"%s %s",printCommand,tmpFile);
         } else {
            sprintf(cmd, printCommand, tmpFile);
         }
         sprintf(gszMsgBox, "Printing with '%s' command.", printCommand);
         if (strcmp(printCommand, "cat") != 0) {
            strcat(cmd, PRINT_TRAILER);
         }
         Msg(gszMsgBox);
      }
      if (strcmp(printCommand, "cat") == 0) {
         /*
          * Well, you can't really "cat" to a printer (without piping to
          *       anything else), so you must mean "catting" to stdout!
          */
         FILE *pfp=NULL;

         sprintf(gszMsgBox, "Executing '%s'...", cmd);
         SetStringStatus(gszMsgBox);
         if (!PRTGIF) XSync(mainDisplay, False);
         if ((pfp=(FILE*)popen(cmd, "r")) == NULL) return FALSE;

         while (fgets(gszMsgBox, sizeof(gszMsgBox)-1, pfp) != NULL) {
            printf("%s", gszMsgBox);
         }
         pclose(pfp);
         SetStringStatus("...Done");
      } else if (!ExecuteCmd(cmd)) {
         if (PRTGIF) {
            fprintf(stderr, "Cannot execute '%s', print aborted.\n", cmd);
         } else {
            sprintf(gszMsgBox, "Cannot execute '%s', print aborted.", cmd);
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
         unlink(tmpFile);
         return FALSE;
      }
      if (PRTGIF) {
         fprintf(stderr, "'%s' printed.\n\n", tmpFile);
      } else {
         Msg("Print completed.");
      }
      break;
   case EPSI_FILE:
   case LATEX_FIG:
   case PS_FILE:
   case PDF_FILE:
      switch (whereToPrint) {
      case EPSI_FILE:
         if (*epsiExportExtension == '\0') {
            file_ext = EPSF_FILE_EXT;
         } else {
            file_ext = epsiExportExtension;
         }
         output_type = "EPSI";
         break;
      case LATEX_FIG:
         file_ext = EPSF_FILE_EXT;
         output_type = "EPS";
         break;
      case PS_FILE:
         file_ext = PS_FILE_EXT;
         output_type = "PostScript";
         break;
      case PDF_FILE:
         file_ext = PS_FILE_EXT;
         output_type = "PDF";
         break;
      }
      if (PRTGIF) {
         SetOutputFileName(ps_file, file_ext, NULL, &FileName);
      } else {
         if (!curFileDefined) {
            sprintf(gszMsgBox, "%s.\n\n%s %s %s!",
                  "No current file", "Cannot generate", output_type, "output");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            unlink(tmpFile);
            return FALSE;
         }
         SetOutputFileName(ps_file, file_ext, &short_name, &rest);
         if (short_name && *outputDir=='\0') {
            sprintf(gszMsgBox, "Printing into '%s' ...", rest);
         } else {
            sprintf(gszMsgBox, "Printing into '%s' ...", ps_file);
         }
         Msg(gszMsgBox);
      }
      if (whereToPrint == EPSI_FILE && generateTiffEPSI) {
         int saved_colordump=colorDump, consensed_already=FALSE, ok=TRUE;

         SaveStatusStrings();
         SetStringStatus("Generating TIFF preview bitmap...");

         colorDump = FALSE;
         DumpXBitmapFile(FALSE, FALSE);
         colorDump = saved_colordump;

         if (psUseShortHand && !dontCondense) {
            if (CondensePostScriptFile(tmpFile, sizeof(tmpFile))) {
               consensed_already = TRUE;
            } else {
               ok = FALSE;
            }
         }
         if (ok) {
            GenTiffFile(tmpFile, ps_file, psUseShortHand, consensed_already);
         }
         RestoreStatusStrings();
         unlink(tmpFile);
      } else {
         if (!CopyAFile(tmpFile, (PRTGIF && cmdLineStdOut) ? NULL : ps_file,
               psUseShortHand)) {
            sprintf(gszMsgBox, "%s file not generated.", output_type);
            if (PRTGIF) {
               fprintf(stderr, "%s\n", gszMsgBox);
            } else {
               Msg(gszMsgBox);
            }
            unlink(tmpFile);
            return FALSE;
         }
         if (whereToPrint == PDF_FILE) {
            sprintf(cmd, psToPdfCmd, ps_file);
            if (PRTGIF) {
               SetOutputFileName(ps_file, pdfExportExtension, NULL, &FileName);
            } else {
               SetOutputFileName(ps_file, pdfExportExtension, &short_name,
                     &rest);
               if (short_name && *outputDir=='\0') {
                  sprintf(gszMsgBox, "Converting into '%s' ...", rest);
               } else {
                  sprintf(gszMsgBox, "Converting into '%s' ...", ps_file);
               }
               Msg(gszMsgBox);
            }
            strcat(cmd, " > ");
            strcat(cmd, ps_file);
            if (!ExecuteCmd(cmd)) {
               sprintf(gszMsgBox, "Cannot execute '%s', %s file not generated.",
                     cmd, output_type);
               if (PRTGIF) {
                  fprintf(stderr, "%s\n", gszMsgBox);
               } else {
                  Msg(gszMsgBox);
               }
               unlink(tmpFile);
               return FALSE;
            }
         }
      }
      if (tmpFileMode != 0 && chmod(ps_file, tmpFileMode)) {
         if (PRTGIF) {
            fprintf(stderr, "Cannot chmod '%s' to 0%03o.\n", ps_file,
                  tmpFileMode);
         } else {
            if (short_name && *outputDir=='\0') {
               sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", rest,
                     tmpFileMode);
            } else {
               sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", ps_file,
                     tmpFileMode);
            }
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
      }
      if (PRTGIF) {
         if (!cmdLineStdOut) {
            fprintf(stderr, "%s file printed into '%s'.\n\n", output_type,
                  ps_file);
         }
      } else {
         if (short_name && *outputDir=='\0') {
            sprintf(gszMsgBox, "%s file printed into '%s'.", output_type,
                  rest);
         } else {
            sprintf(gszMsgBox, "%s file printed into '%s'.", output_type,
                  ps_file);
         }
         Msg(gszMsgBox);
      }
      break;
   case TEXT_FILE:
      if (PRTGIF) {
         SetOutputFileName(ps_file, TEXT_FILE_EXT, NULL, &FileName);
      } else {
         if (!curFileDefined) {
            sprintf(gszMsgBox, "%s.\n\n%s!",
                  "No current file", "Cannot generate ASCII text output");
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            unlink(tmpFile);
            return FALSE;
         }
         SetOutputFileName(ps_file, TEXT_FILE_EXT, &short_name, &rest);
         if (short_name && *outputDir=='\0') {
            sprintf(gszMsgBox, "Printing into '%s' ...", rest);
         } else {
            sprintf(gszMsgBox, "Printing into '%s' ...", ps_file);
         }
         Msg(gszMsgBox);
      }
      if (!CopyAFile(tmpFile, (PRTGIF && cmdLineStdOut) ? NULL : ps_file,
            FALSE)) {
         if (PRTGIF) {
            fprintf(stderr, "ASCII text output not generated.\n");
         } else {
            Msg("ASCII text output not generated.");
         }
         unlink(tmpFile);
         return FALSE;
      }
      if (tmpFileMode != 0 && chmod(ps_file, tmpFileMode)) {
         if (PRTGIF) {
            fprintf(stderr, "Cannot chmod '%s' to 0%03o.\n", ps_file,
                  tmpFileMode);
         } else {
            if (short_name && *outputDir=='\0') {
               sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", rest,
                     tmpFileMode);
            } else {
               sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", ps_file,
                     tmpFileMode);
            }
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
      }
      if (PRTGIF) {
         if (!cmdLineStdOut) {
            fprintf(stderr, "ASCII text file printed into '%s'.\n\n", ps_file);
         }
      } else {
         if (short_name && *outputDir=='\0') {
            sprintf(gszMsgBox, "ASCII text file printed into '%s'.", rest);
         } else {
            sprintf(gszMsgBox, "ASCII text file printed into '%s'.", ps_file);
         }
         Msg(gszMsgBox);
      }
      break;
   default:
      /*
       * whereToPrint >= MAXDEFWHERETOPRINT
       */
      if (PRTGIF) {
         SetOutputFileName(ps_file, GetExportExt(whereToPrint), NULL,
               &FileName);
      } else {
         if (!curFileDefined) {
            sprintf(gszMsgBox, "%s.\n\nCannot export %s file!",
                  "No current file", GetExportName(whereToPrint, EXPORT_THIS));
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            unlink(tmpFile);
            return FALSE;
         }
         SetOutputFileName(ps_file, GetExportExt(whereToPrint), &short_name,
               &rest);
         if (short_name && *outputDir=='\0') {
            sprintf(gszMsgBox, "Printing into '%s' ...", rest);
         } else {
            sprintf(gszMsgBox, "Printing into '%s' ...", ps_file);
         }
         Msg(gszMsgBox);
      }
      if (!CopyAFile(tmpFile, (PRTGIF && cmdLineStdOut) ? NULL : ps_file,
            FALSE)) {
         sprintf(gszMsgBox, "%s file not generated.",
               GetExportName(whereToPrint, EXPORT_THIS));
         if (PRTGIF) {
            fprintf(stderr, "%s\n", gszMsgBox);
         } else {
            Msg(gszMsgBox);
         }
         unlink(tmpFile);
         return FALSE;
      }
      if (tmpFileMode != 0 && chmod(ps_file, tmpFileMode)) {
         if (PRTGIF) {
            fprintf(stderr, "Cannot chmod '%s' to 0%03o.\n", ps_file,
                  tmpFileMode);
         } else {
            if (short_name && *outputDir=='\0') {
               sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", rest,
                     tmpFileMode);
            } else {
               sprintf(gszMsgBox, "Cannot chmod '%s' to 0%03o.", ps_file,
                     tmpFileMode);
            }
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
         }
      }
      if (PRTGIF) {
         if (!cmdLineStdOut) {
            fprintf(stderr, "%s file printed into '%s'.\n\n",
                  GetExportName(whereToPrint, EXPORT_THIS), ps_file);
         }
      } else {
         if (short_name && *outputDir=='\0') {
            sprintf(gszMsgBox, "%s file printed into '%s'.",
                  GetExportName(whereToPrint, EXPORT_THIS), rest);
         } else {
            sprintf(gszMsgBox, "%s file printed into '%s'.",
                  GetExportName(whereToPrint, EXPORT_THIS), ps_file);
         }
         Msg(gszMsgBox);
      }
      break;
   }
#ifdef KEEP_WHEN_PRINT
   if (whereToPrint != PRINTER) {
      unlink(tmpFile);
   }
#else
   unlink(tmpFile);
#endif
   return TRUE;
}

static
int GenDump(FileName)
   char *FileName;
{
   int rc=INVALID, saved_where_to_print=whereToPrint;

   xpmHasTooManyColorsForPrinting = FALSE;
   if (whereToPrint == TIFFEPSI_FILE) {
      whereToPrint = EPSI_FILE;
      generateTiffEPSI = TRUE;
   }
   SaveStatusStrings();
   rc = DoGenDump(FileName);
   RestoreStatusStrings();
   if (saved_where_to_print == TIFFEPSI_FILE) {
      whereToPrint = TIFFEPSI_FILE;
      generateTiffEPSI = FALSE;
   }
   if (!preDumpSetup && xpmHasTooManyColorsForPrinting) {
      sprintf(gszMsgBox, "%s %s.\n\n%s %s.",
            "XPM objects that have more than 255 colors",
            "have been skipped for printing/exporting",
            "You can use ReduceColors() from the ReduceNumberOfColors Submenu",
            "of the ImageProc Menu to reduce their number of colors");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   }
   return rc;
}

void Dump(FileName)
   char *FileName;
{
   int ok=TRUE;
   char dummy_str[MAXSTRING+1];
   struct PageRec *saved_cur_page;

   if ((whereToPrint == LATEX_FIG || whereToPrint == EPSI_FILE ||
         whereToPrint == TIFFEPSI_FILE) && pageLayoutMode == PAGE_TILE &&
         (paperCol != 1 || paperRow != 1)) {
      strcpy(dummy_str, "Cannot print in EPS format in Tiled page mode.");
      if (PRTGIF) {
         fprintf(stderr, "%s\n", dummy_str);
      } else {
         MsgBox("Cannot print in EPS format in Tiled page mode.",
               TOOL_NAME, INFO_MB);
      }
      return;
   }
   if (PRTGIF) {
      RemoveFileNameExtension(FileName);
      if (cmdLineOneFilePerPage ||
            (cmdLineHasPageNum && pageLayoutMode==PAGE_STACK &&
            (whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
            whereToPrint==PRINTER))) {
         totalBBoxValid = FALSE;
         printingFirstPageNum = printingPageNum = printingLastPageNum =
               curPageNum;
         dumpOnePageInStackMode = TRUE;
      }
   } else {
      HighLightReverse();
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);
   }
   totalBBoxValid = FALSE;
   if (whereToPrint == PRINTER || whereToPrint == PS_FILE ||
         whereToPrint == PDF_FILE || whereToPrint == TEXT_FILE ||
         (whereToPrint >= MAXDEFWHERETOPRINT &&
         (GetExportFilterCapability(whereToPrint) & TGEF_CAP_MULTI_PAGE) !=
         0)) {
      if (PRTGIF && (cmdLineOneFilePerPage || (cmdLineHasPageNum &&
            pageLayoutMode==PAGE_STACK && (whereToPrint==PS_FILE ||
            whereToPrint==PDF_FILE || whereToPrint==PRINTER)))) {
         if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
               whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
               whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
            ResetPSInfo();

            ok = GenDump(FileName);

            DoneResetPSInfo();
         }
         if (ok) GenDump(FileName);
      } else {
         XColor *saved_tgif_colors=tgifColors;

         if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

         if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
               whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
               whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
            ResetPSInfo();

            printingFirstPageNum = 1;
            printingPageNum = 1;
            printingLastPageNum = lastPageNum;
            saved_cur_page = curPage;
            SaveStatusStrings();
            for (curPage=firstPage; ok && curPage != NULL;
                  curPage=curPage->next, printingPageNum++) {
               topObj = curPage->top;
               botObj = curPage->bot;
               sprintf(dummy_str, "Preprocess page %1d of %1d...",
                     printingPageNum, lastPageNum);
               SetStringStatus(dummy_str);
               ok = GenDump(FileName);
            }
            RestoreStatusStrings();
            curPage = saved_cur_page;
            topObj = curPage->top;
            botObj = curPage->bot;

            DoneResetPSInfo();
         }
         printingFirstPageNum = 1;
         printingPageNum = 1;
         printingLastPageNum = lastPageNum;
         saved_cur_page = curPage;
         SaveStatusStrings();
         for (curPage=firstPage; ok && curPage != NULL;
               curPage=curPage->next, printingPageNum++) {
            topObj = curPage->top;
            botObj = curPage->bot;
            sprintf(dummy_str, "Generating page %1d of %1d...",
                  printingPageNum, lastPageNum);
            SetStringStatus(dummy_str);
            ok = GenDump(FileName);
         }
         RestoreStatusStrings();
         curPage = saved_cur_page;
         topObj = curPage->top;
         botObj = curPage->bot;

         if (printUsingRequestedColor) tgifColors = saved_tgif_colors;
      }
   } else {
      XColor *saved_tgif_colors=tgifColors;

      if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

      if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
            whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
            whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
         ResetPSInfo();

         printingFirstPageNum = 1;
         printingPageNum = 1;
         printingLastPageNum = 1;
         ok = GenDump(FileName);

         DoneResetPSInfo();
      }
      printingFirstPageNum = 1;
      printingPageNum = 1;
      printingLastPageNum = 1;
      if (ok) GenDump(FileName);

      if (printUsingRequestedColor) tgifColors = saved_tgif_colors;
   }
   if (PRTGIF) {
      if (cmdLineOneFilePerPage || (cmdLineHasPageNum &&
            pageLayoutMode==PAGE_STACK && (whereToPrint==PS_FILE ||
            whereToPrint==PDF_FILE || whereToPrint==PRINTER))) {
         dumpOnePageInStackMode = FALSE;
      }
   } else {
      SetDefaultCursor(mainWindow);
      ShowCursor();
      HighLightForward();
   }
}

void DumpOnePageInTileMode(row, col)
   int row, col;
{
   int ok=TRUE;
   XColor *saved_tgif_colors=tgifColors;

   if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   totalBBoxValid = FALSE;
   printingFirstPageNum = printingPageNum = printingLastPageNum =
         row*paperCol+col+1;
   dumpOnePageInTileMode = TRUE;
   if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
         whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
         whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
      ResetPSInfo();
      ok = GenDump("");
      DoneResetPSInfo();
   }
   if (ok) GenDump("");
   dumpOnePageInTileMode = FALSE;

   SetDefaultCursor(mainWindow);
   ShowCursor();

   if (printUsingRequestedColor) tgifColors = saved_tgif_colors;
}

void DumpOnePageInStackMode()
{
   int ok=TRUE;
   XColor *saved_tgif_colors=tgifColors;

   if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   totalBBoxValid = FALSE;
   printingFirstPageNum = printingPageNum = printingLastPageNum = curPageNum;
   dumpOnePageInStackMode = TRUE;
   if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
         whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
         whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
      ResetPSInfo();
      ok = GenDump("");
      DoneResetPSInfo();
   }
   if (ok) GenDump("");
   dumpOnePageInStackMode = FALSE;

   SetDefaultCursor(mainWindow);
   ShowCursor();

   if (printUsingRequestedColor) tgifColors = saved_tgif_colors;
}

void DumpOneFilePerPage()
{
   int ok=TRUE, saved_cur_page_num=curPageNum;
   XColor *saved_tgif_colors=tgifColors;

   if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   if (lastPageNum != 1) dumpOneFilePerPage = TRUE;
   dumpOnePageInStackMode = TRUE;
   for (curPageNum=1; ok && curPageNum <= lastPageNum; curPageNum++) {
      totalBBoxValid = FALSE;
      GotoPageNum(curPageNum);
      printingFirstPageNum = printingPageNum = printingLastPageNum = curPageNum;
      if (whereToPrint==LATEX_FIG || whereToPrint==PS_FILE ||
            whereToPrint==PDF_FILE || whereToPrint==EPSI_FILE ||
            whereToPrint==TIFFEPSI_FILE) {
         ResetPSInfo();
         ok = GenDump("");
         DoneResetPSInfo();
         if (ok) ok = GenDump("");
      } else {
         ok = GenDump("");
      }
   }
   dumpOnePageInStackMode = FALSE;
   dumpOneFilePerPage = FALSE;

   SetDefaultCursor(mainWindow);
   ShowCursor();

   if (printUsingRequestedColor) tgifColors = saved_tgif_colors;

   if (!ok && curPageNum <= lastPageNum) {
      if (curPageNum == lastPageNum) {
         sprintf(gszMsgBox, "Page %1d has not been generated.", lastPageNum);
      } else {
         sprintf(gszMsgBox, "Pages %1d through %1d have not been generated.",
               curPageNum, lastPageNum);
      }
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
   }
   if (saved_cur_page_num != curPageNum) GotoPageNum(saved_cur_page_num);
}

void PrintWithCommand(FileName)
   char *FileName;
{
   struct PageRec *saved_cur_page=NULL;
   int ok=TRUE;
   XColor *saved_tgif_colors=tgifColors;

   if (whereToPrint != PRINTER) {
      MsgBox("PrintWithCmd only works when output device is the printer.",
            TOOL_NAME, INFO_MB);
      return;
   }
   Dialog("Please enter print command name:",
         "( <CR>: accept, <ESC>: cancel )", printCommand);
   if (*printCommand == '\0') return;

   if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

   SetWatchCursor(drawWindow);
   SetWatchCursor(mainWindow);

   if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
         whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
         whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
      ResetPSInfo();

      printingPageNum = 1;
      printingLastPageNum = lastPageNum;
      saved_cur_page = curPage;
      totalBBoxValid = FALSE;
      SaveStatusStrings();
      for (curPage=firstPage; ok && curPage != NULL; curPage=curPage->next,
            printingPageNum++) {
         char dummy_str[MAXSTRING+1];

         topObj = curPage->top;
         botObj = curPage->bot;
         sprintf(dummy_str, "Preprocess page %1d of %1d...", printingPageNum,
               lastPageNum);
         SetStringStatus(dummy_str);
         ok = GenDump(FileName);
      }
      RestoreStatusStrings();
      curPage = saved_cur_page;
      topObj = curPage->top;
      botObj = curPage->bot;

      DoneResetPSInfo();
   }
   printingPageNum = 1;
   printingLastPageNum = lastPageNum;
   saved_cur_page = curPage;
   totalBBoxValid = FALSE;
   SaveStatusStrings();
   for (curPage=firstPage; ok && curPage != NULL; curPage=curPage->next,
         printingPageNum++) {
      char dummy_str[MAXSTRING+1];

      topObj = curPage->top;
      botObj = curPage->bot;
      sprintf(dummy_str, "Generating page %1d of %1d...", printingPageNum,
            lastPageNum);
      SetStringStatus(dummy_str);
      ok = GenDump(FileName);
   }
   RestoreStatusStrings();
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;

   SetDefaultCursor(mainWindow);
   ShowCursor();

   if (printUsingRequestedColor) tgifColors = saved_tgif_colors;
}

void PrintSelectedObjs()
{
   struct SelRec *top_sel_ptr, *bot_sel_ptr, *sel_ptr, *next_sel;
   struct ObjRec *saved_top_obj, *saved_bot_obj, *obj_ptr;

   if (topSel == NULL) {
      MsgBox("No objects selected!\n\nNothing printed!", TOOL_NAME, INFO_MB);
      return;
   }
   HighLightReverse();
   PushPageInfo();
   saved_top_obj = topObj;
   saved_bot_obj = botObj;

   JustDupSelObj(&top_sel_ptr, &bot_sel_ptr);
   curPage->top = topObj = top_sel_ptr->obj;
   curPage->bot = botObj = bot_sel_ptr->obj;
   for (sel_ptr=topSel, obj_ptr=topObj; obj_ptr!=NULL;
         sel_ptr=sel_ptr->next, obj_ptr=obj_ptr->next) {
      CopyObjId(sel_ptr->obj, obj_ptr);
      CopyObjLocks(sel_ptr->obj, obj_ptr);
   }

   if (pageLayoutMode == PAGE_STACK) {
      DumpOnePageInStackMode();
   } else {
      int ok=TRUE;
      XColor *saved_tgif_colors=tgifColors;

      if (printUsingRequestedColor) tgifColors = tgifRequestedColors;

      printingFirstPageNum = 1;
      printingPageNum = 1;
      printingLastPageNum = 1;
      totalBBoxValid = FALSE;
      if (whereToPrint==PRINTER || whereToPrint==LATEX_FIG ||
            whereToPrint==PS_FILE || whereToPrint==PDF_FILE ||
            whereToPrint==EPSI_FILE || whereToPrint==TIFFEPSI_FILE) {
         ResetPSInfo();
         ok = GenDump("");
         DoneResetPSInfo();
      }
      if (ok) GenDump("");

      if (printUsingRequestedColor) tgifColors = saved_tgif_colors;
   }
   DelAllObj();
   for (sel_ptr = top_sel_ptr; sel_ptr != NULL; sel_ptr = next_sel) {
      next_sel = sel_ptr->next;
      free(sel_ptr);
   }
   PopPageInfo();
   curPage->top = topObj = saved_top_obj;
   curPage->bot = botObj = saved_bot_obj;
   RedrawAnArea(botObj, selLtX-GRID_ABS_SIZE(1), selLtY-GRID_ABS_SIZE(1),
         selRbX+GRID_ABS_SIZE(1), selRbY+GRID_ABS_SIZE(1));
   HighLightForward();
}

void SetPrintReduction()
{
   float fval;
   char buf[MAXSTRING+1];

   Dialog("Please specify percent reduction (<100) or enlargement (>100):",
         "( <CR>: accept, <ESC>: cancel )", buf);
   if (*buf == '\0') return;

   if (sscanf(buf, "%f", &fval) != 1) {
      sprintf(gszMsgBox, "Invalid reduction '%s'.", buf);
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   printMag = fval;
   if (UpdPageStyle(pageStyle)) {
      UpdDrawWinBBox();
      AdjSplineVs();
      ClearAndRedrawDrawWindow();
   }
   RedrawScrollBars();
   RedrawRulers();
   RedrawTitleWindow();
   SetFileModified(TRUE);

   FormatFloat(&printMag, buf);
   if (printMag <= 100.0) {
      sprintf(gszMsgBox, "New reduction is %s%%.", buf);
   } else {
      sprintf(gszMsgBox, "New enlargement is %s%%.", buf);
   }
   Msg(gszMsgBox);
}

void NewProc()
{
   if (gstWBInfo.do_whiteboard) {
      XBell(mainDisplay, 0);
      if (MsgBox("Ok to clear WhiteBoard?", TOOL_NAME, YNC_MB) != MB_ID_YES) {
         return;
      }
   } else {
      while (!DirIsRemote(curDir) && fileModified) {
         XBell(mainDisplay, 0);
         switch (MsgBox("File modified, save file before clear? [ync](y)",
               TOOL_NAME, YNC_MB)) {
         case MB_ID_YES: SaveFile(); break;
         case MB_ID_NO: TieLooseEnds(); SetFileModified(FALSE); break;
         case MB_ID_CANCEL: return;
         }
      }
   }
   if (inHyperSpace) ToggleHyperSpace(FALSE);

   CleanUpCmds();
   CleanUpDrawingWindow();
   ClearFileInfo(TRUE);
   ClearAndRedrawDrawWindow();

   CheckFileAttrsInLoad();

   Msg("Editing no file.");
   if (gstWBInfo.do_whiteboard) {
      RecordWBClearAll();
   } else {
      objId = 0;
   }
   RedrawTitleWindow();
   DelAllPages();
   lastPageNum = 1;
   InitPage();
   ShowPage();
}

void OpenProc()
{
   char file_name[MAXPATHLENGTH+1];
   int do_not_save=FALSE, need_to_check_auto_exec=FALSE;

   if (gstWBInfo.do_whiteboard) {
      MsgBox("'Open' is disabled when running in the WhiteBoard mode!",
            TOOL_NAME, INFO_MB);
      return;
   }
   while (!DirIsRemote(curDir) && fileModified) {
      XBell(mainDisplay, 0);
      switch (MsgBox("File modified, save file before open? [ync](y)",
            TOOL_NAME, YNC_MB)) {
      case MB_ID_YES: SaveFile(); break;
      case MB_ID_NO: do_not_save = TRUE; SetFileModified(FALSE); break;
      case MB_ID_CANCEL: return;
      }
   }
   if (SelectFileName("Please select a file to OPEN ...",file_name) !=
         INVALID) {
      SetWatchCursor(drawWindow);
      SetWatchCursor(mainWindow);

      if (FileIsRemote(file_name)) {
         char remote_fname[MAXPATHLENGTH+1];
         char *page_spec=NULL;

         if (!FormNewFileName(curDir, file_name, NULL, remote_fname,
               &page_spec)) {
            sprintf(gszMsgBox, "Invalid remote file name '%s'.", file_name);
            MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
            if (do_not_save) SetFileModified(TRUE);
         } else {
            char *buf=NULL, *content_type=NULL, final_url[MAXPATHLENGTH+1];
            int rc, buf_sz=0, is_html=FALSE;

            *final_url = '\0';
            SaveStatusStrings();
            rc = LoadRemoteFileInMem(remote_fname, &buf, &content_type,
                  &buf_sz, &is_html, TRUE, final_url, sizeof(final_url));
            RestoreStatusStrings();
            if (rc && buf != NULL) {
               if (*final_url != '\0') {
                  UtilStrCpyN(remote_fname, sizeof(remote_fname), final_url);
               }
               if (LoadRemoteFileFromMem(remote_fname, buf, content_type,
                     buf_sz, is_html)) {
                  need_to_check_auto_exec = TRUE;
               }
            } else if (do_not_save) {
               SetFileModified(TRUE);
            }
            if (content_type != NULL) FreeRemoteBuf(content_type);
            if (buf != NULL) FreeRemoteBuf(buf);
            if (page_spec != NULL) {
               int new_page_num=(-1);

               need_to_check_auto_exec = FALSE;
               if (!GetPageNumFromPageSpec(page_spec, &new_page_num)) {
                  sprintf(gszMsgBox, "Invalid page specified for '%s'.",
                        file_name);
                  MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
               } else if (new_page_num != curPageNum) {
                  BeforeNavigate();
                  GotoPageNum(new_page_num);
                  ShowPage();
                  ClearAndRedrawDrawWindow();
                  RedrawTitleWindow();
                  RedrawRulers();
                  RedrawScrollBars();
                  justDupped = FALSE;
               }
            }
         }
         if (page_spec != NULL) free(page_spec);
      } else {
         int gzipped=FALSE, obj_file=FALSE;

         obj_file = FileNameHasExtension(file_name, OBJ_FILE_TYPE, &gzipped,
               NULL);
         if (!LoadFile(file_name, obj_file, obj_file && gzipped)) {
            if (do_not_save) {
               SetFileModified(TRUE);
            }
         } else {
            need_to_check_auto_exec = TRUE;
         }
      }
      SetDefaultCursor(mainWindow);
      ShowCursor();
   } else if (do_not_save) {
      SetFileModified(TRUE);
   }

   if (need_to_check_auto_exec) {
      struct AttrRec *exec_attr=FindFileAttrWithName("auto_exec=");

      if (exec_attr != NULL) {
         DoExecLoop(NULL, exec_attr);
      }
   }
}

void SetTemplate()
{
   char file_name[MAXPATHLENGTH+1], *rest=NULL;
   int short_name, rc;
   FILE *fp=NULL;
   struct AttrRec *attr_ptr=NULL;
   struct SelRec *saved_top_sel=NULL, *saved_bot_sel=NULL;

   if (gstWBInfo.do_whiteboard) {
      sprintf(gszMsgBox, "SetTemplate() is disabled in WhiteBoard.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   MakeQuiescent();

   if (importFromLibrary) {
      char name[MAXSTRING+1], path[MAXSTRING+1];

      if (SelectFromLibrary("Please select an object file to IMPORT...",
            OBJ_FILE_EXT, name, path) == INVALID) {
         return;
      }
      sprintf(file_name, "%s%c%s", path, DIR_SEP, name);
   } else {
      sprintf(gszMsgBox, "Please select a file to be used as template...");
      if (SelectFileNameToImport(gszMsgBox, OBJ_FILE_EXT, file_name) ==
            INVALID) {
         return;
      } else if (FileIsRemote(file_name)) {
         MsgBox("Using a remote template file is not supported.", TOOL_NAME,
               INFO_MB);
         return;
      }
   }
   if (FileIsRemote(file_name)) {
      sprintf(gszMsgBox, "Using a remote template file is not supported.");
      MsgBox(gszMsgBox, TOOL_NAME, INFO_MB);
      return;
   }
   if ((short_name=IsPrefix(bootDir, file_name, &rest))) ++rest;
   if ((fp=fopen(file_name, "r")) == NULL) {
      sprintf(gszMsgBox, "Cannot open '%s' for reading.\n\n%s?",
            short_name ? rest : file_name,
            "Would you still like to use it as a template file");
      if (MsgBox(gszMsgBox, TOOL_NAME, YNC_MB) != MB_ID_YES) {
         return;
      }
   } else {
      fclose(fp);
   }
   StartCompositeCmd();
   importingFile = TRUE;
   rc = ImportGivenFile(file_name, TRUE);
   importingFile = FALSE;
   if (rc == BAD) {
      EndCompositeCmd();
      return;
   } else if (rc == FALSE) {
      sprintf(gszMsgBox, "%s '%s' as a template file?",
            "Would you still like to use",
            short_name ? rest : file_name);
      if (MsgBox(gszMsgBox, TOOL_NAME, YNC_MB) != MB_ID_YES) {
         EndCompositeCmd();
         return;
      }
   }
   HighLightReverse();
   saved_top_sel = topSel;
   saved_bot_sel = botSel;
   topSel = botSel = NULL;
   UpdSelBBox();

   if ((attr_ptr=FindAttrWithName(tgifObj, "template=", NULL)) != NULL) {
      ReplaceAttrFirstValue(tgifObj, attr_ptr, short_name ? rest : file_name);
   } else {
      int x=tgifObj->obbox.ltx, y=tgifObj->obbox.rby;

      AddObj(NULL, topObj, tgifObj);
      UpdSelBBox();

      PrepareToReplaceAnObj(tgifObj);
      attr_ptr = AddAttrByNameAndValue(tgifObj, "template=",
            short_name ? rest : file_name);
      attr_ptr->shown = TRUE;
      attr_ptr->obj->color = colorIndex;
      MoveObj(attr_ptr->obj, x-attr_ptr->obj->x, y-attr_ptr->obj->y);
      UpdTextBBox(attr_ptr->obj);
      AdjObjBBox(tgifObj);

      recordCmdIncludeTgifObj = TRUE;
      RecordReplaceAnObj(tgifObj);
      recordCmdIncludeTgifObj = FALSE;

      RemoveAllSel();
      UnlinkObj(topObj);
      UpdSelBBox();

      SetFileModified(TRUE);
      justDupped = FALSE;
   }
   EndCompositeCmd();
   sprintf(gszMsgBox, "Template set to '%s'.", short_name ? rest : file_name);
   Msg(gszMsgBox);
   topSel = saved_top_sel;
   botSel = saved_bot_sel;
   UpdSelBBox();
   HighLightForward();
}

char * fileMenuStr[] =
      { "New             ^n",
        "Open            ^o",
        "Save            ^s",
        "SaveNew        ^#s",
        "Import          #p",
        "ImportXBitmap  ^#.",
        "ImportXPixmap  ^#,",
        "ImportEPSFile   #(",
        "ImportGIFFile     ",
        "ImportOtherFile   ",
        "EmbedEPSFile      ",
        "BrowseXBitmap     ",
        "BrowseXPixmap     ",
        "BrowseOther       ",
        "ChangeDomain    ^c",
        "Print           ^p",
        "PrintWithCmd    ^-",
        "PrintSelectedObjs ",
        "PrintOnePage      ",
        "SetExportPixelTrim",
        "InputPolyPts      ",
        "InputPolygonPts   ",
        "SetTemplate       ",
        "Solve           #s",
        "Simulate        #y",
        "Probe           #b",
        "Animate         ^z",
        "Escape          #x",
        "SaveSelectedAs  #~",
        "SaveSymInLibrary  ",
        "Quit            ^q",
        NULL
      };

int QuitProc()
{
   int do_not_save=FALSE;

   while (!DirIsRemote(curDir) && fileModified) {
      XBell(mainDisplay, 0);
      switch (MsgBox("File modified, save file before quit? [ync](y)",
            TOOL_NAME, YNC_MB)) {
      case MB_ID_YES: SaveFile(); break;
      case MB_ID_NO: do_not_save = TRUE; SetFileModified(FALSE); break;
      case MB_ID_CANCEL: return INVALID;
      }
   }
   if (AncesterModified()) {
      switch (MsgBox("Ancester file modified, still quitting? [ync](y)",
            TOOL_NAME, YNC_MB)) {
      case MB_ID_YES: return FILE_QUIT;
      case MB_ID_NO:
         if (do_not_save) SetFileModified(TRUE);
         return INVALID;
      case MB_ID_CANCEL:
         if (do_not_save) SetFileModified(TRUE);
         return INVALID;
      }
   }
   gnQuit = TRUE;
   return FILE_QUIT;
}

int SolveProc()
{
   if (!saveTmpOnReturn) return FILE_SOLVE;

   switch (SaveTmpFile("tmpmodel")) {
   case OBJ_FILE_TYPE: return FILE_SOLVE;
   case SYM_FILE_TYPE: return INVALID;
   case PIN_FILE_TYPE: return INVALID;
   case INVALID: return INVALID;
   }
   return INVALID;
}

int SimulateProc()
{
   if (!saveTmpOnReturn) return FILE_SIMULATE;

   switch (SaveTmpFile("tmpmodel")) {
   case OBJ_FILE_TYPE: return FILE_SIMULATE;
   case SYM_FILE_TYPE: return INVALID;
   case PIN_FILE_TYPE: return INVALID;
   case INVALID: return INVALID;
   }
   return INVALID;
}

int ProbeProc()
{
   if (!saveTmpOnReturn) return FILE_PROBE;

   switch (SaveTmpFile("tmpmodel")) {
   case OBJ_FILE_TYPE: return FILE_PROBE;
   case SYM_FILE_TYPE: return INVALID;
   case PIN_FILE_TYPE: return INVALID;
   case INVALID: return INVALID;
   }
   return INVALID;
}

int AnimateProc()
{
   if (!saveTmpOnReturn) return FILE_ANIMATE;

   switch (SaveTmpFile("tmpmodel")) {
   case OBJ_FILE_TYPE: return FILE_ANIMATE;
   case SYM_FILE_TYPE: return INVALID;
   case PIN_FILE_TYPE: return INVALID;
   case INVALID: return INVALID;
   }
   return INVALID;
}

int EscapeProc()
{
   return FILE_ESCAPE;
}

int RefreshFileMenu(menu)
   TgMenu *menu;
{
   int ok=TRUE;

   /* Print/ExportInColor */
   ok &= TgSetMenuItemCheckById(menu, CMDID_TOGGLEBWCOLORPS, colorDump);
   /* PrintOneFilePerPage */
   ok &= TgEnableMenuItemById(menu, CMDID_PRINTONEFILEPERPAGE,
         (pageLayoutMode==PAGE_STACK));

   return ok;
}

TgMenu *CreateFileMenu(parent_menu, x, y, menu_info)
   TgMenu *parent_menu;
   int x, y;
   TgMenuInfo *menu_info;
{
   TgMenu *menu=TgCreateMenuFromMenuInfo(parent_menu, x, y, menu_info);

   if (menu != NULL) {
      if (!RefreshFileMenu(menu)) {
         return TgDestroyMenu(menu, TRUE);
      }
      menu->refresh_proc = ((RefreshMenuFunc*)RefreshFileMenu);
   }
   return menu;
}

int FileMenu(X, Y, TrackMenuBar)
   int X, Y, TrackMenuBar;
{
   int rc=INVALID;
   TgMenu *menu=(fileMenuInfo.create_proc)(NULL, X, Y, &fileMenuInfo);

   activeMenu = MENU_FILE;
   if (menu != NULL) {
      rc = TgMenuLoop(menu);
      TgDestroyMenu(menu, TRUE);
   }
   return rc;
}

void CleanUpFiles()
{
   ClearBgColorInfo(FALSE);
   ClearBgPixmapInfo(FALSE);

   CleanUpDumpInitDefaults();
   ClearFileInfo(TRUE);
   if (usePaperSizeStoredInFile) ResetOnePageSize();
   fileModified = FALSE;
}
